Package: ameras
Title: Analyze Multiple Exposure Realizations in Association Studies
Version: 0.4.0
Authors@R: c(person(given = "Sander",
                      family = "Roberti",
                      role =  c("aut", "cre"),
                      email = "sander.roberti@nih.gov",
                      comment = c(ORCID = "0000-0002-6275-7442")),
              person(given = "William",
                      family = "Wheeler",
                      role = "aut",
                      email = "WheelerB@imsweb.com"),
               person(given = "Deukwoo",
                      family = "Kwon",
                      role = "aut",
                      email = "DKwon@uams.edu",
                      comment = c(ORCID = "0000-0001-5376-5320")),
               person(given = "Ruth",
                      family = "Pfeiffer",
                      role = "ctb",
                      email = "pfeiffer@mail.nih.gov",
                      comment = c(ORCID = "0000-0001-7791-2698")),
                person("NCI", role=c("cph","fnd")))
Depends: R (>= 4.1.0), stats, nimble
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), ggplot2, dplyr, tidyr,
        scales, patchwork
Description: Analyze association studies with multiple realizations of a noisy or uncertain exposure. These can be obtained from e.g. a two-dimensional Monte Carlo dosimetry system (Simon et al 2015 <doi:10.1667/RR13729.1>) to characterize exposure uncertainty.
    The implemented methods are regression calibration (Carroll et al. 2006 <doi:10.1201/9781420010138>), extended 
    regression calibration (Little et al. 2023 <doi:10.1038/s41598-023-42283-y>), Monte Carlo maximum 
    likelihood (Stayner et al. 2007 <doi:10.1667/RR0677.1>), frequentist model averaging (Kwon et al. 2023 <doi:10.1371/journal.pone.0290498>), 
    and Bayesian model averaging (Kwon et al. 2016 <doi:10.1002/sim.6635>). Supported model families are
    Gaussian, binomial, multinomial, Poisson, proportional hazards, and conditional logistic.
License: MIT + file LICENSE
Imports: Rcpp (>= 1.0.10), RcppEigen, coda, numDeriv, mvtnorm, methods,
        MCMCvis, tidyselect, lifecycle, graphics, grDevices
LinkingTo: Rcpp, RcppEigen
NeedsCompilation: yes
VignetteBuilder: knitr
Config/testthat/edition: 3
Author: Sander Roberti [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6275-7442>),
  William Wheeler [aut],
  Deukwoo Kwon [aut] (ORCID: <https://orcid.org/0000-0001-5376-5320>),
  Ruth Pfeiffer [ctb] (ORCID: <https://orcid.org/0000-0001-7791-2698>),
  NCI [cph, fnd]
Maintainer: Sander Roberti <sander.roberti@nih.gov>
URL: https://ameras.sanderroberti.com,
        https://github.com/sanderroberti/ameras
BugReports: https://github.com/sanderroberti/ameras/issues
Packaged: 2026-05-29 15:56:50 UTC; robertis2
Repository: CRAN
Date/Publication: 2026-05-29 23:30:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-30 02:49:59 UTC; unix
Archs: ameras.so.dSYM

useDynLib(ameras, .registration = TRUE)
importFrom(Rcpp, evalCpp)
import(RcppEigen)
import(methods)
importFrom(
  "stats",
  "cov",
  "var",
  "optim",
  "pchisq",
  "quantile",
  "rexp",
  "qnorm",
  "rnorm",
  "runif",
  "sd",
  "uniroot",
  "optimize",
  ".getXlevels",
  "model.frame",
  "model.matrix",
  "na.pass",
  "terms",
  "as.formula",
  "confint",
  "qqline",
  "qqnorm",
  "ppoints",
  "smooth.spline"
)
importFrom("mvtnorm", "rmvnorm")
importFrom(
  "nimble",
  "nimbleCode",
  "nimbleFunction",
  "getNimbleOption",
  "nf_preProcessMemberDataObject",
  "nimbleModel",
  "compileNimble",
  "configureMCMC",
  "buildMCMC",
  "runMCMC"
)
importFrom("coda", "HPDinterval", "as.mcmc")
importFrom("numDeriv", "hessian")
importFrom("MCMCvis", "MCMCsummary", "MCMCtrace")
importFrom("utils", "head")
importFrom("graphics", "abline", "panel.smooth", "par", "text", "lines")
importFrom("grDevices", "dev.interactive")
export(
  ameras,
  transform1,
  transform1.inv,
  transform1.jacobian,
  traceplot,
  ecdfplot,
  dose_lrt,
  Rhat,
  included_realizations,
  convergence,
  summary_table
)
S3method(coef, amerasfit)
S3method(summary, amerasfit)
S3method(print, summary.amerasfit)
S3method(traceplot, amerasfit)
S3method(print, amerasfit)
S3method(confint, amerasfit)
S3method(dose_lrt, amerasfit)
S3method(included_realizations, amerasfit)
S3method(convergence, amerasfit)
S3method(Rhat, amerasfit)
S3method(residuals, amerasfit)
S3method(plot, amerasfit)
S3method(summary_table, amerasfit)
S3method(vcov, amerasfit)

# ameras <a href="https://ameras.sanderroberti.com"><img src="man/figures/logo.png" align="right" height="139" alt="ameras website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/sanderroberti/ameras/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sanderroberti/ameras/actions/workflows/R-CMD-check.yaml)
[![](https://cranlogs.r-pkg.org/badges/grand-total/ameras)](https://CRAN.R-project.org/package=ameras)
[![Codecov test
coverage](https://codecov.io/gh/sanderroberti/ameras/graph/badge.svg)](https://app.codecov.io/gh/sanderroberti/ameras)
<!-- badges: end -->

The goal of ameras is to provide a user-friendly interface to analyze
association studies with multiple realizations of a noisy exposure using
a variety of methods. ameras supports continuous, count, binary,
multinomial, and right-censored time-to-event outcomes. For binary
outcomes, the nested case-control design is also accommodated. Besides
the common exponential relative risk model $RR=\exp(\beta D)$ for the
exposure-outcome association with noisy exposure $D$, linear excess
relative risk $RR=1+\beta D$ and linear-exponential excess relative risk
models $RR=1+\beta_1 D \exp(\beta_2 D)$ can be used.

## Installation

To install from CRAN:

``` r
install.packages("ameras")
```

To install the development version from GitHub:

``` r
# install.packages("pak") # If pak is not already installed
pak::pak("sanderroberti/ameras")
```

## Example

This is a basic example which shows you how to fit a simple logistic
regression model. First, we visualize the dose uncertainty. In the left
panel of the plot below, the empirical cumulative distribution function
(ECDF) is plotted for each dose realization. In other words, each curve
shows one distribution of dose across individuals. The spread within
individual curves reflects the dose range across individuals, while the
spread between curves reflects between-realization variation on the
cohort level.

In the right panel, ECDFs are plotted for each individual, showing
distributions within individuals. A wide spread within individual curves
is indicative of large within-individual variation, while the spread
between curves reflects between-individual variation.

``` r
library(ameras)
#> Loading required package: nimble
#> nimble version 1.4.2 is loaded.
#> For more information on NIMBLE and a User Manual,
#> please visit https://R-nimble.org.
#> 
#> Attaching package: 'nimble'
#> The following object is masked from 'package:stats':
#> 
#>     simulate
#> The following object is masked from 'package:base':
#> 
#>     declare
data(data, package="ameras")
ecdfplot(data, paste0("V", 1:10), show.mean = FALSE)
```

<img src="man/figures/README-unnamed-chunk-2-1.png" alt="" width="100%" />

Next, we apply all available methods to the data:

``` r
set.seed(12345) # For reproducibility
fit <- ameras(Y.binomial~dose(V1:V10), data, family="binomial", 
              methods=c("RC","ERC","MCML", "FMA", "BMA"))
#> Note: BMA may require extensive computation time
#> Fitting RC
#> Fitting ERC
#> Fitting MCML
#> Fitting FMA
#> Fitting BMA
#> Defining model
#> Building model
#> Setting data and initial values
#> Running calculate on model
#>   [Note] Any error reports that follow may simply reflect missing values in model variables.
#> Checking model sizes and dimensions
#>   [Note] This model is not fully initialized. This is not an error.
#>          To see which variables are not initialized, use model$initializeInfo().
#>          For more information on model initialization, see help(modelInitialization).
#> Compiling
#>   [Note] This may take a minute.
#>   [Note] Use 'showCompilerOutput = TRUE' to see C++ compilation details.
#> Compiling
#>   [Note] This may take a minute.
#>   [Note] Use 'showCompilerOutput = TRUE' to see C++ compilation details.
#> running chain 1...
#> |-------------|-------------|-------------|-------------|
#> |-------------------------------------------------------|
#> running chain 2...
#> |-------------|-------------|-------------|-------------|
#> |-------------------------------------------------------|
summary(fit)
#> Call:
#> ameras(formula = Y.binomial ~ dose(V1:V10), data = data, family = "binomial", 
#>     methods = c("RC", "ERC", "MCML", "FMA", "BMA"))
#> 
#> Rows: 3000
#> 
#> Total CPU runtime: 51.2 seconds
#> 
#> CPU runtime in seconds by method:
#> 
#>  Method    Fit  CI  Total
#>      RC  0.028 0.0  0.028
#>     ERC  7.951 0.0  7.951
#>    MCML  0.110 0.0  0.110
#>     FMA  0.311 0.0  0.311
#>     BMA 42.780 0.0 42.780
#> 
#> Summary of coefficients by method:
#> 
#>  Method        Term Estimate      SE Rhat   n.eff
#>      RC (Intercept)  -0.8847 0.07378   NA      NA
#>      RC        dose   0.8020 0.13751   NA      NA
#>     ERC (Intercept)  -0.8849 0.07477   NA      NA
#>     ERC        dose   0.8214 0.14304   NA      NA
#>    MCML (Intercept)  -0.8758 0.07323   NA      NA
#>    MCML        dose   0.7910 0.13644   NA      NA
#>     FMA (Intercept)  -0.8760 0.07320   NA      NA
#>     FMA        dose   0.7917 0.13656   NA      NA
#>     BMA (Intercept)  -0.8736 0.07174 1.00 1024.00
#>     BMA        dose   0.7911 0.13323 1.00  999.00
#> 
#> Note: confidence intervals not yet computed. Use confint() to add them.
```

Finally, we add confidence intervals to the `fit` object:

``` r
fit <- confint(fit, type=c("wald.orig","percentile"))
#> RC confidence intervals:
#> 
#>               lower   upper
#> (Intercept) -1.0293 -0.7401
#> dose         0.5324  1.0715
#> 
#> ERC confidence intervals:
#> 
#>               lower   upper
#> (Intercept) -1.0314 -0.7384
#> dose         0.5411  1.1018
#> 
#> MCML confidence intervals:
#> 
#>               lower   upper
#> (Intercept) -1.0193 -0.7323
#> dose         0.5236  1.0584
#> 
#> FMA confidence intervals:
#> 
#>               lower   upper
#> (Intercept) -1.0199 -0.7327
#> dose         0.5241  1.0594
#> 
#> BMA confidence intervals:
#> 
#>              lower  upper
#> (Intercept) -1.019 -0.735
#> dose         0.552  1.080
summary(fit)
#> Call:
#> ameras(formula = Y.binomial ~ dose(V1:V10), data = data, family = "binomial", 
#>     methods = c("RC", "ERC", "MCML", "FMA", "BMA"))
#> 
#> Rows: 3000
#> 
#> Total CPU runtime: 51.2 seconds
#> 
#> CPU runtime in seconds by method:
#> 
#>  Method    Fit    CI  Total
#>      RC  0.028 0.001  0.029
#>     ERC  7.951 0.000  7.951
#>    MCML  0.110 0.000  0.110
#>     FMA  0.311 0.004  0.315
#>     BMA 42.780 0.001 42.781
#> 
#> Summary of coefficients by method:
#> 
#>  Method        Term Estimate      SE CI.lower CI.upper Rhat   n.eff
#>      RC (Intercept)  -0.8847 0.07378  -1.0293  -0.7401   NA      NA
#>      RC        dose   0.8020 0.13751   0.5324   1.0715   NA      NA
#>     ERC (Intercept)  -0.8849 0.07477  -1.0314  -0.7384   NA      NA
#>     ERC        dose   0.8214 0.14304   0.5411   1.1018   NA      NA
#>    MCML (Intercept)  -0.8758 0.07323  -1.0193  -0.7323   NA      NA
#>    MCML        dose   0.7910 0.13644   0.5236   1.0584   NA      NA
#>     FMA (Intercept)  -0.8760 0.07320  -1.0199  -0.7327   NA      NA
#>     FMA        dose   0.7917 0.13656   0.5241   1.0594   NA      NA
#>     BMA (Intercept)  -0.8736 0.07174  -1.0185  -0.7350 1.00 1024.00
#>     BMA        dose   0.7911 0.13323   0.5520   1.0798 1.00  999.00
```

For visual model diagnostics using residuals, use `plot()`. For example,
for regression calibration:

``` r
plot(fit, methods="RC")
```

<img src="man/figures/README-unnamed-chunk-5-1.png" alt="" width="100%" /><img src="man/figures/README-unnamed-chunk-5-2.png" alt="" width="100%" />

See the vignettes for additional details on model fitting, confidence
intervals, transformations, standard analyses with one dose realization,
manual FMA through multiple RC fits, and parallel FMA.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(ameras)
library(ggplot2)
data(data, package="ameras")

## ----modelfit.exp-------------------------------------------------------------
fit.ameras.exp <- ameras(Y.binomial~dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                         data=data, family="binomial", methods="RC")
summary(fit.ameras.exp)

## ----modelfit.err-------------------------------------------------------------
fit.ameras.err <- ameras(Y.binomial~dose(V1:V10, deg=2, model="ERR")+X1+X2, 
                         data=data, family="binomial", methods="RC")
summary(fit.ameras.err)

## ----modelfit.linexp----------------------------------------------------------
fit.ameras.linexp <- ameras(Y.binomial~dose(V1:V10, model="LINEXP")+X1+X2, 
                         data=data, family="binomial", methods="RC")
summary(fit.ameras.linexp)

## ----comparison, fig.width=7, fig.height=6------------------------------------
ggplot(data.frame(x=c(0, 5)), aes(x))+
  theme_light(base_size=15)+
  xlab("Exposure")+
  ylab("Relative risk")+
  labs(col="Model", lty="Model") +
  theme(legend.position = "inside", 
        legend.position.inside = c(.2,.85),
        legend.box.background = element_rect(color = "black", fill = "white", linewidth = 1))+
  stat_function(aes(col="Linear-quadratic ERR", lty="Linear-quadratic ERR" ),fun=function(x){
    1+fit.ameras.err$RC$coefficients["dose"]*x + fit.ameras.err$RC$coefficients["dose_squared"]*x^2
  }, linewidth=1.2) + 
  stat_function(aes(col="Exponential", lty="Exponential"),fun=function(x){
    exp(fit.ameras.exp$RC$coefficients["dose"]*x + fit.ameras.exp$RC$coefficients["dose_squared"]*x^2)
  }, linewidth=1.2) +
  stat_function(aes(col="Linear-exponential", lty="Linear-exponential"),fun=function(x){
    1+fit.ameras.linexp$RC$coefficients["dose_linear"]*x * exp(fit.ameras.linexp$RC$coefficients["dose_exponential"]*x)
  }, linewidth=1.2)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(ameras)


## ----fits1, eval = identical(Sys.getenv("NOT_CRAN"), "true")------------------
data(data, package="ameras")
dosevars <- paste0("V", 1:10)
fit.ameras <- ameras(Y.binomial~dose(V1:V10, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods=c("RC"))

fit.ameras.waldorig <- confint(fit.ameras, type="wald.orig")
fit.ameras.waldtransformed <- confint(fit.ameras, type="wald.transformed")
fit.ameras.proflik <- confint(fit.ameras, type="proflik", parm="all")

summary(fit.ameras.waldorig)
summary(fit.ameras.waldtransformed)
summary(fit.ameras.proflik)

## ----fits2, eval = identical(Sys.getenv("NOT_CRAN"), "true")------------------
set.seed(123)
fit.ameras2 <- ameras(Y.binomial~dose(V1:V10, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods=c("FMA"))

fit.ameras.hpd <- confint(fit.ameras2, type="hpd")
fit.ameras.percentile <- confint(fit.ameras2, type="percentile")

summary(fit.ameras.hpd)
summary(fit.ameras.percentile)

## ----include = FALSE----------------------------------------------------------
  knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
  )

## ----setup--------------------------------------------------------------------
  library(ameras)

## -----------------------------------------------------------------------------
data(data, package="ameras")
head(data)

## ----modelfit.linreg, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(12345)
fit.ameras.linreg <- ameras(Y.gaussian~dose(V1:V10)+X1+X2, data=data, 
                            family="gaussian", niter.BMA=5000, nburnin.BMA=1000,
                            methods=c("RC", "ERC", "MCML", "FMA", "BMA"))

## ----eval = identical(Sys.getenv("NOT_CRAN"), "true")-------------------------
str(fit.ameras.linreg)

## ----eval = identical(Sys.getenv("NOT_CRAN"), "true")-------------------------
fit.ameras.linreg$RC

## ----eval = identical(Sys.getenv("NOT_CRAN"), "true")-------------------------
summary(fit.ameras.linreg)

## ----eval = identical(Sys.getenv("NOT_CRAN"), "true")-------------------------
coef(fit.ameras.linreg)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
traceplot(fit.ameras.linreg)

## ----modelfit.logreg, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(33521)
fit.ameras.logreg <- ameras(Y.binomial~dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                            data=data, family="binomial", 
                            methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                            niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.logreg)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.logreg)

## ----modelfit.logreg.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(3521216)
fit.ameras.logreg.lin <- ameras(Y.binomial~dose(V1:V10, deg=1, model="EXP")+X1+X2,  
                                data=data, family="binomial", 
                                methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.logreg.lin)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.logreg.lin)

## ----modelfit.poisson, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(332101)
fit.ameras.poisson <- ameras(Y.poisson~dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                             data=data, family="poisson", 
                             methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                             niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.poisson)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.poisson)

## ----modelfit.poisson.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(24252)
fit.ameras.poisson.lin <- ameras(Y.poisson~dose(V1:V10, deg=1, model="EXP")+X1+X2, 
                                 data=data, family="poisson", 
                                 methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                 niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.poisson.lin)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.poisson.lin)

## ----modelfit.prophaz, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(332120000)
fit.ameras.prophaz <- ameras(Surv(time, status)~
                               dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                             data=data, family="prophaz", 
                             methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                             niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.prophaz)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.prophaz)

## ----eval = identical(Sys.getenv("NOT_CRAN"), "true")-------------------------
fit.ameras.prophaz$BMA$prophaz.timepoints

## ----modelfit.prophaz.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(24978252)
fit.ameras.prophaz.lin <- ameras(Surv(time, status)~
                                   dose(V1:V10, deg=1, model="EXP")+X1+X2, 
                                 data=data, family="prophaz", 
                                 methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                 niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.prophaz.lin)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.prophaz.lin)

## ----modelfit.multinomial, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(33)
fit.ameras.multinomial <- ameras(Y.multinomial~
                                   dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                                 data=data, family="multinomial",
                                 methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                 niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.multinomial)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.multinomial)

## ----modelfit.multinomial.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(44)
fit.ameras.multinomial.lin <- ameras(Y.multinomial~
                                       dose(V1:V10, deg=1, model="EXP")+X1+X2, 
                                     data=data, family="multinomial",
                                     methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                     niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.multinomial.lin)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.multinomial.lin)

## ----modelfit.clogit, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(3301)
fit.ameras.clogit <- ameras(Y.clogit~dose(V1:V10, deg=2, model="EXP")+X1+X2+
                              strata(setnr), data=data, family="clogit", 
                            methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                            niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.clogit)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.clogit)

## ----modelfit.clogit.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
set.seed(4401)
fit.ameras.clogit.lin <- ameras(Y.clogit~dose(V1:V10, deg=2, model="EXP")+X1+X2+
                                  strata(setnr), data=data, family="clogit", 
                                methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                niter.BMA = 5000, nburnin.BMA = 1000)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
summary(fit.ameras.clogit.lin)

## ----fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")----
coef(fit.ameras.clogit.lin)

## ----include = FALSE----------------------------------------------------------
  knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
  )

## ----setup--------------------------------------------------------------------
  library(ameras)

## -----------------------------------------------------------------------------
transform1 <- function(params, index.t=1:length(params), lowlimit=rep(0,length(index.t)), 
                       boundcheck=FALSE, boundtol=1e-3, ...){
  
  if(length(index.t)!=length(lowlimit)) 
    stop("Length mismatch between index.t and lowlimit")
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  params[index.t] <- exp(params[index.t]) + lowlimit
  if(boundcheck){
    if(any(params[index.t]-lowlimit < boundtol)) 
      warning(paste0("WARNING: one or multiple parameter estimates within ", boundtol, " of 
        lower bounds. Try different bounds or starting values."))
  }
  return(params)
}

transform1.jacobian <- function(params, index.t=1:length(params), ...){ 
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  grad <- rep(1, length(params))
  grad[index.t] <- exp(params[index.t])
  if(length(params)>1){
    return(diag(grad))
  } else{
    return(matrix(grad))
  }
}

## -----------------------------------------------------------------------------
transform.sigmoid <- function(params, index.t=1:length(params), a=rep(0,length(index.t)), 
                              b=rep(1,length(index.t)), boundcheck=FALSE, boundtol=1e-3, ...){
  
  if(length(index.t)!=length(a) | length(index.t) != length(b)) 
    stop("Length mismatch between index.t, a, and b")
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  
  params[index.t] <- a + (b-a) * 1/(1+exp(-1*params[index.t]))
  if(boundcheck){
    if(any( (params[index.t]-a < boundtol) | (b-params[index.t] < boundtol))) 
      warning(paste0("WARNING: one or multiple parameter estimates within ", boundtol, 
      " of bounds. Try different bounds or starting values."))
  }
  return(params)
}

## -----------------------------------------------------------------------------
transform.sigmoid.jacobian <- function(params, index.t=1:length(params), 
                                       a=rep(0,length(index.t)), b=rep(1,length(index.t)), ...){ 
  if(length(index.t)!=length(a) | length(index.t) != length(b)) 
    stop("Length mismatch between index.t, a, and b")
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  grad <- rep(1, length(params))
  grad[index.t] <- (b-a)*exp(-1*params[index.t])/(1+exp(-1*params[index.t]))^2
  if(length(params)>1){
    return(diag(grad))
  } else{
    return(matrix(grad))
  }
}

## ----modelfit.sigmoid---------------------------------------------------------
data(data, package="ameras")
fit.ameras.sigmoid <- ameras(Y.binomial~dose(V1:V10, deg=2, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods="RC", transform=transform.sigmoid, 
                            transform.jacobian=transform.sigmoid.jacobian, index.t=4:5)
summary(fit.ameras.sigmoid)

## ----modelfit.transform1------------------------------------------------------
fit.ameras.transform1 <- ameras(Y.binomial~dose(V1:V10, deg=2, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods="RC")
summary(fit.ameras.transform1)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
)

## ----setup--------------------------------------------------------------------
library(ameras)

## ----data---------------------------------------------------------------------
data(data, package = "ameras")

## Use one exposure realization as the single observed dose.
data$D <- data$V1
head(data[c("D", "Y.gaussian", "Y.binomial", "Y.poisson", "X1", "X2")])

## ----helper-------------------------------------------------------------------
compare_coefficients <- function(ameras_fit, standard_fit) {
  ameras_terms <- c("(Intercept)", "X1", "X2", "dose")
  standard_terms <- c("(Intercept)", "X1", "X2", "D")

  out <- data.frame(
    term = ameras_terms,
    ameras = unname(ameras_fit$RC$coefficients[ameras_terms]),
    standard = unname(stats::coef(standard_fit)[standard_terms])
  )

  out$ameras <- round(out$ameras, 4)
  out$standard <- round(out$standard, 4)
  out
}

## ----gaussian-----------------------------------------------------------------
fit_ameras_gaussian <- ameras(Y.gaussian ~ dose(D) + X1 + X2,
                              data = data,
                              family = "gaussian")

fit_lm <- lm(Y.gaussian ~ D + X1 + X2, data = data)

compare_coefficients(fit_ameras_gaussian, fit_lm)

## ----gaussian-summary---------------------------------------------------------
fit_ameras_gaussian <- confint(fit_ameras_gaussian,
                               type = "wald.orig",
                               print = FALSE)
summary(fit_ameras_gaussian)

## ----binomial-----------------------------------------------------------------
fit_ameras_binomial <- ameras(Y.binomial ~ dose(D, model = "EXP") + X1 + X2,
                              data = data,
                              family = "binomial")

fit_glm_binomial <- glm(Y.binomial ~ D + X1 + X2,
                        data = data,
                        family = binomial())

compare_coefficients(fit_ameras_binomial, fit_glm_binomial)

## ----poisson------------------------------------------------------------------
fit_ameras_poisson <- ameras(Y.poisson ~ dose(D, model = "EXP") + X1 + X2,
                             data = data,
                             family = "poisson")

fit_glm_poisson <- glm(Y.poisson ~ D + X1 + X2,
                       data = data,
                       family = poisson())

compare_coefficients(fit_ameras_poisson, fit_glm_poisson)

## ----multiple-realizations----------------------------------------------------
fit_ameras_rc <- ameras(Y.binomial ~ dose(V1:V10, model = "EXP") + X1 + X2,
                        data = data,
                        family = "binomial")
summary(fit_ameras_rc)

## ----multiple-methods, eval = FALSE-------------------------------------------
# fit_ameras_all <- ameras(Y.binomial ~ dose(V1:V10, model = "EXP") + X1 + X2,
#                          data = data,
#                          family = "binomial",
#                          methods = c("RC", "ERC", "MCML", "FMA", "BMA"))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
)

## ----setup--------------------------------------------------------------------
library(ameras)

## ----data---------------------------------------------------------------------
data(data, package = "ameras")
dosevars <- paste0("V", 1:10)

## ----fit-helper---------------------------------------------------------------
passes_fma_screen <- function(rc) {
  hessian <- rc$optim$hessian

  if (is.null(hessian) || rc$optim$convergence != 0) {
    return(FALSE)
  }

  isTRUE(det(hessian) != 0 &&
    rcond(hessian) > .Machine$double.eps &&
    all(eigen(hessian)$values > 0))
}

fit_rc_realization <- function(dosevar, data) {
  formula <- stats::as.formula(
    paste0("Y.gaussian ~ dose(", dosevar, ") + X1 + X2")
  )

  fit <- ameras(formula, data = data, family = "gaussian")
  rc <- fit$RC

  list(
    dosevar = dosevar,
    coefficients = rc$coefficients,
    vcov = rc$vcov,
    loglik = rc$loglik,
    AIC = -2 * rc$loglik + 2 * length(rc$coefficients),
    include = passes_fma_screen(rc)
  )
}

## ----fit-realizations---------------------------------------------------------
rc_summaries <- lapply(dosevars, fit_rc_realization, data = data)

## ----save-fits, eval = FALSE--------------------------------------------------
# dir.create("fma-fits", showWarnings = FALSE)
# 
# for (dosevar in dosevars) {
#   fit_summary <- fit_rc_realization(dosevar, data = data)
#   saveRDS(fit_summary, file = file.path("fma-fits", paste0(dosevar, ".rds")))
# }
# 
# fit_files <- file.path("fma-fits", paste0(dosevars, ".rds"))
# rc_summaries <- lapply(fit_files, readRDS)

## ----assemble-helper----------------------------------------------------------
assemble_manual_fma <- function(rc_summaries, MFMA = 100000) {
  n_total <- length(rc_summaries)
  valid <- vapply(rc_summaries, function(x) isTRUE(x$include), logical(1))
  n_screen_excluded <- sum(!valid)

  if (!any(valid)) {
    stop("No realization-specific fits are available for FMA.")
  }

  original_indices <- which(valid)
  rc_summaries <- rc_summaries[valid]

  AIC <- vapply(rc_summaries, `[[`, numeric(1), "AIC")
  weights <- exp(-0.5 * (AIC - min(AIC)))
  weights <- weights / sum(weights)

  allocated_samples <- round(weights * MFMA)
  keep <- allocated_samples > 0
  n_weight_excluded <- sum(!keep)

  if (!any(keep)) {
    stop("No realizations received FMA samples. Increase MFMA.")
  }

  rc_summaries <- rc_summaries[keep]
  original_indices <- original_indices[keep]
  weights <- weights[keep]
  allocated_samples <- allocated_samples[keep]
  names(weights) <- vapply(rc_summaries, `[[`, character(1), "dosevar")
  diagnostics <- data.frame(
    stage = c(
      "total realization-specific fits",
      "excluded by convergence/Hessian screen",
      "excluded after zero-sample allocation",
      "included in manual FMA"
    ),
    realizations = c(
      n_total,
      n_screen_excluded,
      n_weight_excluded,
      length(original_indices)
    ),
    row.names = NULL
  )

  samples <- do.call(
    "rbind",
    Map(
      function(fit, n) {
        mvtnorm::rmvnorm(n = n, mean = fit$coefficients, sigma = fit$vcov)
      },
      rc_summaries,
      allocated_samples
    )
  )

  samples <- as.data.frame(samples)
  names(samples) <- names(rc_summaries[[1]]$coefficients)

  coefficients <- colMeans(samples)
  se <- apply(samples, 2, stats::sd)
  percentile_CI <- t(
    apply(samples, 2, stats::quantile, probs = c(0.025, 0.975))
  )
  colnames(percentile_CI) <- c("lower", "upper")

  list(
    coefficients = coefficients,
    SE = se,
    percentile_CI = percentile_CI,
    vcov = stats::var(samples),
    weights = weights,
    samples = samples,
    included.realizations = original_indices,
    included.samples = nrow(samples),
    diagnostics = diagnostics
  )
}

## ----manual-fma---------------------------------------------------------------
set.seed(100)
manual_fma <- assemble_manual_fma(rc_summaries, MFMA = 100000)
manual_fma$diagnostics

manual_summary <- data.frame(
  term = names(manual_fma$coefficients),
  estimate = unname(manual_fma$coefficients),
  SE = unname(manual_fma$SE),
  percentile_lower = manual_fma$percentile_CI[, "lower"],
  percentile_upper = manual_fma$percentile_CI[, "upper"],
  row.names = NULL
)

manual_summary[-1] <- round(manual_summary[-1], 4)
manual_summary

## ----manual-weights-----------------------------------------------------------
round(manual_fma$weights, 4)
manual_fma$included.realizations
manual_fma$included.samples

## ----builtin-fma--------------------------------------------------------------
set.seed(100)
fit_builtin_fma <- suppressWarnings(
  ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
         data = data,
         family = "gaussian",
         methods = "FMA",
         MFMA = 100000)
)

comparison <- data.frame(
  term = names(manual_fma$coefficients),
  manual = unname(manual_fma$coefficients),
  builtin = unname(fit_builtin_fma$FMA$coefficients[names(manual_fma$coefficients)]),
  row.names = NULL
)

comparison[-1] <- round(comparison[-1], 4)
comparison

## ----compare-weights----------------------------------------------------------
weight_comparison <- merge(
  data.frame(dose = names(manual_fma$weights),
             manual = unname(manual_fma$weights)),
  data.frame(dose = names(fit_builtin_fma$FMA$weights),
             builtin = unname(fit_builtin_fma$FMA$weights)),
  by = "dose",
  all = TRUE
)

weight_comparison[-1] <- round(weight_comparison[-1], 4)
weight_comparison

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
)

## ----setup--------------------------------------------------------------------
library(ameras)
data(data, package = "ameras")

## ----contrast-fit-------------------------------------------------------------
fit_contrast <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ M1) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_contrast)

## ----subgroup-fit-------------------------------------------------------------
fit_subgroup <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ 0 + M1) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_subgroup)

## ----coding-equivalence-------------------------------------------------------
contrast_coef <- fit_contrast$RC$coefficients
subgroup_coef <- fit_subgroup$RC$coefficients

data.frame(
  term = c("reference group", "non-reference group"),
  from_contrast = c(
    contrast_coef["dose"],
    contrast_coef["dose"] + contrast_coef["dose:M1"]
  ),
  from_subgroup = c(
    subgroup_coef["dose[M1=0]"],
    subgroup_coef["dose[M1=1]"]
  ),
  row.names = NULL
)

## ----factor-setup-------------------------------------------------------------
data$M_factor <- factor(
  ifelse(data$M1 == 0, "unexposed modifier group", "modifier group"),
  levels = c("unexposed modifier group", "modifier group")
)

## ----factor-contrast----------------------------------------------------------
fit_factor_contrast <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ M_factor) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_factor_contrast)

## ----factor-subgroup----------------------------------------------------------
fit_factor_subgroup <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ 0 + M_factor) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_factor_subgroup)

## ----three-level-factor-------------------------------------------------------
data$M3 <- factor(
  rep(c("low", "middle", "high"), length.out = nrow(data)),
  levels = c("low", "middle", "high")
)

fit_three_level <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ 0 + M3) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_three_level)

## ----subgroup-ci, eval = FALSE------------------------------------------------
# fit_subgroup <- confint(
#   fit_subgroup,
#   type = "proflik",
#   parm = "dose"
# )
# 
# summary(fit_subgroup)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
)

future_available <- requireNamespace("future", quietly = TRUE) &&
  requireNamespace("future.apply", quietly = TRUE)
run_parallel_examples <- future_available

## ----setup--------------------------------------------------------------------
library(ameras)

## ----data---------------------------------------------------------------------
data(data, package = "ameras")

## ----sequential-plan, eval = FALSE--------------------------------------------
# future::plan(future::sequential)
# 
# fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
#               data = data,
#               family = "gaussian",
#               methods = "FMA")

## ----multisession-plan, eval = FALSE------------------------------------------
# future::plan(future::multisession, workers = 2)
# 
# fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
#               data = data,
#               family = "gaussian",
#               methods = "FMA")
# 
# future::plan(future::sequential)

## ----chunk-size, eval = FALSE-------------------------------------------------
# future::plan(future::multisession, workers = 2)
# 
# fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
#               data = data,
#               family = "gaussian",
#               methods = "FMA",
#               future.chunk.size.FMA = 2)
# 
# future::plan(future::sequential)

## ----compare-plans, eval = run_parallel_examples------------------------------
old_plan <- future::plan()
on.exit(future::plan(old_plan), add = TRUE)

future::plan(future::sequential)
set.seed(2024)
sequential_elapsed <- system.time(
  sequential_fit <- suppressWarnings(
    ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
           data = data,
           family = "gaussian",
           methods = "FMA",
           MFMA = 10000)
  )
)[["elapsed"]]

future::plan(future::multisession, workers = 2)
set.seed(2024)
parallel_elapsed <- system.time(
  parallel_fit <- suppressWarnings(
    ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
           data = data,
           family = "gaussian",
           methods = "FMA",
           MFMA = 10000,
           future.chunk.size.FMA = 2)
  )
)[["elapsed"]]

future::plan(old_plan)

## ----compare-results, eval = run_parallel_examples----------------------------
coefficient_comparison <- data.frame(
  term = names(sequential_fit$FMA$coefficients),
  sequential = unname(sequential_fit$FMA$coefficients),
  parallel = unname(parallel_fit$FMA$coefficients),
  row.names = NULL
)

coefficient_comparison[-1] <- round(coefficient_comparison[-1], 4)
coefficient_comparison

data.frame(
  plan = c("sequential", "multisession"),
  elapsed_seconds = round(c(sequential_elapsed, parallel_elapsed), 2)
)

## ----reproducibility, eval = FALSE--------------------------------------------
# future::plan(future::multisession, workers = 2)
# 
# set.seed(2024)
# fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
#               data = data,
#               family = "gaussian",
#               methods = "FMA")
# 
# future::plan(future::sequential)

ameras_main <- function(
  family = "gaussian",
  methods = "RC",
  dosevars,
  data,
  deg = 1,
  doseRRmod = "ERR",
  transform = NULL,
  transform.jacobian = NULL,
  Y = NULL,
  M = NULL,
  X = NULL,
  offset = NULL,
  inpar = NULL,
  entry = NULL,
  exit = NULL,
  status = NULL,
  setnr = NULL,
  unweightedFMA = FALSE,
  loglim = 1e-30,
  MFMA = 100000,
  prophaz.numints.BMA = 10,
  ERRprior.BMA = "doubleexponential",
  nburnin.BMA = 5000,
  niter.BMA = 20000,
  nchains.BMA = 2,
  thin.BMA = 10,
  included.replicates.BMA = 1:length(dosevars),
  optim.method = "Nelder-Mead",
  control = NULL,
  ...
) {
  if (is.null(control)) {
    control <- list(reltol = 1e-10)
  }

  if (!is.null(transform) & is.null(transform.jacobian)) {
    stop("transform.jacobian is required when using a transformation")
  }

  # Family in ("gaussian", "poisson", "prophaz", "binomial")
  # gaussian requires Y and can use X and M
  # binomial requires Y and can use X and M
  # poisson requires Y and can use offset, X and M
  # prophaz requires status and exit and can use entry, X and M

  # Method in ("MCML", "RC", "ERC", "FMA", "BMA")
  # If both FMA and BMA are to be run, run FMA first to determine the included replicates and skip that step in BMA
  if ("FMA" %in% methods & "BMA" %in% methods) {
    methods[methods %in% c("FMA", "BMA")] <- c("FMA", "BMA")
  }

  # If both RC and ERC are to be run, run RC first to determine initial values for ERC (not implemented yet)
  if ("RC" %in% methods & "ERC" %in% methods) {
    methods[methods %in% c("RC", "ERC")] <- c("RC", "ERC")
  }

  # CI in ("wald.orig","wald.transformed", "proflik") for method in ("RC", "ERC", "MCML")
  # CI in ("percentile", "hpd") for method in ("FMA", "BMA")

  # Input checks
  if (family != "gaussian" & is.null(doseRRmod)) {
    stop("doseRRmod is required for all families other than Gaussian")
  }

  if (family == "prophaz") {
    if (is.null(exit)) {
      stop("exit is required for family=prophaz")
    }
    if (is.null(status)) stop("status is required for family=prophaz")
  }

  # For the Gaussian family, sigma needs to be >0 and so use standard reparametrization if another is not defined
  if (family == "gaussian" & is.null(transform)) {
    transform <- function(params, boundcheck = FALSE, ...) {
      return(transform1(
        params = params,
        index.t = 2 + length(X) + length(M) * deg + deg,
        boundcheck = boundcheck,
        ...
      ))
    }
    transform.jacobian <- function(params, ...) {
      return(transform1.jacobian(
        params = params,
        index.t = 2 + length(X) + length(M) * deg + deg,
        ...
      ))
    }
  }

  # For linear ERR models, if no transformation is specified, use transform1 with lower limits -1/max(dose) for linear dose-response and (0,-1/max(dose^2)) for linear and linear-quadratic parameters, respectively. If effect modifiers M are specified, no transformation is used for those parameters. If negative RRs are evaluated, the program will produce an error and the user should specify a different transformation or bounds
  if (family != "gaussian") {
    if (doseRRmod == "ERR" & is.null(transform)) {
      if (deg == 1) {
        lwlmt <- -1 / max(data[, dosevars])
      } else {
        lwlmt <- c(0, -1 / max(data[, dosevars]^2))
      }
      if (family != "multinomial") {
        transform <- function(params, boundcheck = FALSE, ...) {
          return(transform1(
            params = params,
            index.t = (1 *
              (!(family %in% c("prophaz", "clogit"))) +
              length(X) +
              1):(1 *
              (!(family %in% c("prophaz", "clogit"))) +
              length(X) +
              deg),
            lowlimit = lwlmt,
            boundcheck = boundcheck,
            ...
          ))
        }
        transform.jacobian <- function(params, ...) {
          return(transform1.jacobian(
            params = params,
            index.t = (1 *
              (!(family %in% c("prophaz", "clogit"))) +
              length(X) +
              1):(1 *
              (!(family %in% c("prophaz", "clogit"))) +
              length(X) +
              deg),
            lowlimit = lwlmt,
            ...
          ))
        }
      } else {
        lwlmt <- rep(lwlmt, length(levels(data[, Y])) - 1)
        indx <- do.call(
          "c",
          lapply(0:(length(levels(data[, Y])) - 2), function(xx) {
            xx *
              (1 + length(X) + length(M) * deg + deg) +
              ((1 + length(X) + 1):(1 + length(X) + deg))
          })
        )

        transform <- function(params, boundcheck = FALSE, ...) {
          return(transform1(
            params = params,
            index.t = indx,
            lowlimit = lwlmt,
            boundcheck = boundcheck,
            ...
          ))
        }
        transform.jacobian <- function(params, ...) {
          return(transform1.jacobian(
            params = params,
            index.t = indx,
            lowlimit = lwlmt,
            ...
          ))
        }
      }
    } else if (doseRRmod == "LINEXP" & is.null(transform)) {
      lwlmt <- 0 # Lower bound of 0 for beta1, no bound for beta2

      if (family != "multinomial") {
        transform <- function(params, boundcheck = FALSE, ...) {
          return(transform1(
            params = params,
            index.t = (1 *
              (!(family %in% c("prophaz", "clogit"))) +
              length(X) +
              1),
            lowlimit = lwlmt,
            boundcheck = boundcheck,
            ...
          ))
        }
        transform.jacobian <- function(params, ...) {
          return(transform1.jacobian(
            params = params,
            index.t = (1 *
              (!(family %in% c("prophaz", "clogit"))) +
              length(X) +
              1),
            lowlimit = lwlmt,
            ...
          ))
        }
      } else {
        lwlmt <- rep(lwlmt, length(levels(data[, Y])) - 1)
        indx <- do.call(
          "c",
          lapply(0:(length(levels(data[, Y])) - 2), function(xx) {
            xx * (1 + length(X) + length(M) * deg + deg) + (1 + length(X) + 1)
          })
        )

        transform <- function(params, boundcheck = FALSE, ...) {
          return(transform1(
            params = params,
            index.t = indx,
            lowlimit = lwlmt,
            boundcheck = boundcheck,
            ...
          ))
        }
        transform.jacobian <- function(params, ...) {
          return(transform1.jacobian(
            params = params,
            index.t = indx,
            lowlimit = lwlmt,
            ...
          ))
        }
      }
    }
  }

  results <- list(
    transform = transform,
    transform.jacobian = transform.jacobian
  )
  for (method in methods) {
    if (method == "MCML") {
      message("Fitting MCML")
      fit <- ameras.mcml(
        family = family,
        dosevars = dosevars,
        data = data,
        deg = deg,
        transform = transform,
        transform.jacobian = transform.jacobian,
        Y = Y,
        M = M,
        X = X,
        offset = offset,
        inpar = inpar,
        entry = entry,
        exit = exit,
        status = status,
        setnr = setnr,
        doseRRmod = doseRRmod,
        loglim = loglim,
        control = control,
        optim.method = optim.method,
        ...
      )
      results <- c(results, list(MCML = fit))
    } else if (method == "RC") {
      message("Fitting RC")
      fit <- ameras.rc(
        family = family,
        dosevars = dosevars,
        data = data,
        deg = deg,
        ERC = FALSE,
        transform = transform,
        transform.jacobian = transform.jacobian,
        Y = Y,
        M = M,
        X = X,
        offset = offset,
        inpar = inpar,
        entry = entry,
        exit = exit,
        status = status,
        setnr = setnr,
        doseRRmod = doseRRmod,
        loglim = loglim,
        control = control,
        optim.method = optim.method,
        ...
      )
      results <- c(results, list(RC = fit))
    } else if (method == "ERC") {
      message("Fitting ERC")
      fit <- ameras.rc(
        family = family,
        dosevars = dosevars,
        data = data,
        deg = deg,
        ERC = TRUE,
        transform = transform,
        transform.jacobian = transform.jacobian,
        Y = Y,
        M = M,
        X = X,
        offset = offset,
        inpar = inpar,
        entry = entry,
        exit = exit,
        status = status,
        setnr = setnr,
        doseRRmod = doseRRmod,
        loglim = loglim,
        control = control,
        optim.method = optim.method,
        ...
      )
      results <- c(results, list(ERC = fit))
    } else if (method == "FMA") {
      message("Fitting FMA")
      fit <- ameras.fma(
        family = family,
        dosevars = dosevars,
        data = data,
        deg = deg,
        transform = transform,
        transform.jacobian = transform.jacobian,
        Y = Y,
        M = M,
        X = X,
        offset = offset,
        inpar = inpar,
        entry = entry,
        exit = exit,
        status = status,
        setnr = setnr,
        doseRRmod = doseRRmod,
        unweighted = unweightedFMA,
        MFMA = MFMA,
        control = control,
        ...
      )
      results <- c(results, list(FMA = fit))
    } else if (method == "BMA") {
      message("Fitting BMA")
      if (!is.null(included.replicates.BMA)) {
        increps <- included.replicates.BMA
      } else if (!is.null(results$FMA$included.replicates)) {
        increps <- results$FMA$included.replicates
      } else {
        increps <- NULL
      }
      fit <- ameras.bma(
        family = family,
        dosevars = dosevars,
        data = data,
        deg = deg,
        transform = transform,
        Y = Y,
        M = M,
        X = X,
        offset = offset,
        inpar = inpar,
        entry = entry,
        exit = exit,
        status = status,
        setnr = setnr,
        doseRRmod = doseRRmod,
        ERRprior = ERRprior.BMA,
        prophaz_numints = prophaz.numints.BMA,
        nburnin = nburnin.BMA,
        niter = niter.BMA,
        nchains = nchains.BMA,
        thin = thin.BMA,
        control = control,
        included.replicates = increps,
        optim.method = optim.method,
        ...
      )
      results <- c(results, list(BMA = fit))
    }
  }

  return(results)
}

ameras <- function(
  formula = NULL,
  data,
  family = "gaussian",
  methods = "RC",
  # Old arguments kept temporarily for deprecation
  Y = NULL,
  dosevars = NULL,
  doseRRmod = NULL,
  deg = NULL,
  M = NULL,
  X = NULL,
  offset = NULL,
  entry = NULL,
  exit = NULL,
  setnr = NULL,
  CI = NULL,
  params.profCI = NULL,
  maxit.profCI = NULL,
  tol.profCI = NULL,
  # Current arguments
  transform = NULL,
  transform.jacobian = NULL,
  inpar = NULL,
  loglim = 1e-30,
  MFMA = 100000,
  prophaz.numints.BMA = 10,
  ERRprior.BMA = "doubleexponential",
  nburnin.BMA = 5000,
  niter.BMA = 20000,
  nchains.BMA = 2,
  thin.BMA = 10,
  included.replicates.BMA = NULL,
  optim.method = "Nelder-Mead",
  control = NULL,
  keep.data = TRUE,
  ...
) {
  old_interface <- is.null(formula) || !inherits(formula, "formula")

  # Check for errors
  check_df(data)
  check_family(family)
  methods <- check_methods(methods)

  if ("BMA" %in% methods) {
    message("Note: BMA may require extensive computation time")
  }

  # Detect which interface is being used and resolve parsed arguments
  if (!old_interface) {
    # New formula interface
    # Error if old arguments are also supplied
    old_supplied <- Filter(
      Negate(is.null),
      list(
        Y = Y,
        dosevars = dosevars,
        doseRRmod = doseRRmod,
        deg = deg,
        M = M,
        X = X,
        offset = offset,
        entry = entry,
        exit = exit,
        setnr = setnr
      )
    )
    if (length(old_supplied)) {
      stop(
        "Old-style arguments cannot be combined with the formula interface: ",
        paste(names(old_supplied), collapse = ", "),
        ".\n",
        "Please use the formula interface only. See ?ameras for details."
      )
    }

    # Warn for CI arguments moved to confint
    if (!is.null(CI)) {
      lifecycle::deprecate_warn(
        when = "0.2.0",
        what = "ameras(CI)",
        details = paste0(
          "CI is deprecated in ameras(). ",
          "Use confint() after fitting instead, e.g.:\n",
          "  fit <- ameras(...)\n",
          "  fit <- confint(fit, type='wald.orig')"
        )
      )
    }

    parsed <- parse_ameras_formula(formula, data, family)
  } else {
    # Old direct argument interface
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "ameras(Y)",
      details = paste0(
        "The direct argument interface is deprecated and will be removed ",
        "in the next release. ",
        "Please use the formula interface instead:\n",
        "  ameras(Y ~ dose(D1:D10, model='ERR') + X1, ",
        "data=data, family='binomial')\n",
        "See ?ameras for details."
      )
    )

    if (!is.null(CI)) {
      lifecycle::deprecate_warn(
        when = "0.2.0",
        what = "ameras(CI)",
        details = paste0(
          "CI is deprecated in ameras(). ",
          "Use confint() after fitting instead, e.g.:\n",
          "  fit <- ameras(...)\n",
          "  fit <- confint(fit, method='wald.orig')"
        )
      )
    }

    # Validate old-style arguments
    if (is.null(dosevars)) {
      stop("dosevars is required")
    }
    if (is.null(Y)) {
      stop("Y is required")
    }

    # Construct parsed list from direct arguments
    # matching the structure returned by parse_ameras_formula
    if (family == "prophaz") {
      parsed_Y <- NULL
      parsed_status <- Y
    } else {
      parsed_Y <- Y
      parsed_status <- NULL
    }

    parsed <- list(
      Y = parsed_Y,
      status = parsed_status,
      entry = entry,
      exit = exit,
      dosevars = dosevars,
      doseRRmod = doseRRmod %||% if (family != "gaussian") "ERR" else NULL,
      deg = deg %||% 1,
      M = M,
      X = X,
      X_formula = if (!is.null(X)) {
        as.formula(paste("~", paste(X, collapse = "+")))
      } else {
        NULL
      },
      offset = offset,
      setnr = setnr
    )
  }

  if (!is.null(parsed$X_formula)) {
    X_matrix <- model.matrix(parsed$X_formula, data = data)[, -1, drop = FALSE]
    X_colnames <- colnames(X_matrix)

    # Add expanded columns to data, avoiding name conflicts
    existing <- intersect(X_colnames, colnames(data))
    new_cols <- setdiff(X_colnames, colnames(data))

    if (length(new_cols)) {
      data[, new_cols] <- X_matrix[, new_cols, drop = FALSE]
    }

    X <- X_colnames
  } else {
    X <- NULL
  }

  if (family == "clogit") {
    Y <- NULL
    status <- parsed$Y
  } else if (family == "prophaz") {
    Y <- NULL
    status <- parsed$status
  } else {
    Y <- parsed$Y
    status <- NULL
  }

  # Checks that depend on parsed formula
  if (family != "gaussian") {
    check_doseRRmod(parsed$doseRRmod)
  }

  check_Y(Y %||% status, data, family)
  check_D(parsed$dosevars, data, methods)
  check_M(parsed$M, data)
  check_X(X, data)

  if (family == "poisson") {
    check_offset(parsed$offset, data)
  }
  if (family == "prophaz") {
    check_entry_exit(parsed$entry, parsed$exit, data)
  }
  if (family == "clogit") {
    check_setnr(parsed$setnr, data)
  }

  deg <- check_deg(parsed$deg)

  if (family != "multinomial") {
    inpar <- check_inpar(inpar, family, parsed$M, parsed$X, deg)
  } else {
    inpar <- check_inpar(
      inpar,
      family,
      parsed$M,
      parsed$X,
      deg,
      multinom_levels = length(levels(data[, parsed$Y]))
    )
  }

  if (!is.null(parsed$doseRRmod) && parsed$doseRRmod == "LINEXP") {
    deg <- 2
  }

  M <- getVarNumbers(parsed$M, data)
  X <- getVarNumbers(X, data)

  # Add mean dose for RC and ERC to the data
  data$rcdose_ameras <- rowMeans(data[, parsed$dosevars, drop = FALSE])

  model_list <- list(
    data = if (keep.data) data else NULL,
    keep.data = keep.data,
    family = family,
    dosevars = parsed$dosevars,
    Y = Y,
    M = M,
    X_formula = parsed$X_formula,
    X = X,
    offset = parsed$offset,
    entry = parsed$entry,
    exit = parsed$exit,
    status = status,
    setnr = parsed$setnr,
    deg = deg,
    doseRRmod = parsed$doseRRmod,
    loglim = loglim,
    optim.method = optim.method,
    control = control
  )

  result <- ameras_main(
    family = family,
    methods = methods,
    dosevars = parsed$dosevars,
    data = data,
    deg = deg,
    doseRRmod = parsed$doseRRmod,
    transform = transform,
    transform.jacobian = transform.jacobian,
    setnr = parsed$setnr,
    Y = Y,
    M = M,
    X = X,
    offset = parsed$offset,
    inpar = inpar,
    entry = parsed$entry,
    exit = parsed$exit,
    status = status,
    loglim = loglim,
    MFMA = MFMA,
    prophaz.numints.BMA = prophaz.numints.BMA,
    ERRprior.BMA = ERRprior.BMA,
    nburnin.BMA = nburnin.BMA,
    niter.BMA = niter.BMA,
    nchains.BMA = nchains.BMA,
    thin.BMA = thin.BMA,
    included.replicates.BMA = included.replicates.BMA %||%
      seq_along(parsed$dosevars),
    control = control,
    optim.method = optim.method,
    ...
  )
  ret <- c(
    list(
      call = match.call(),
      formula = formula,
      num.rows = nrow(data),
      num.replicates = length(parsed$dosevars),
      transform = result$transform,
      transform.jacobian = result$transform.jacobian,
      other.args = list(...),
      model = model_list,
      CI.computed = FALSE
    ),
    result[setdiff(names(result), c("transform", "transform.jacobian"))]
  )

  ret <- new_amerasfit(ret)

  # Compute CIs for old interface backwards compatibility
  if (old_interface && !is.null(CI)) {
    ret <- confint(
      ret,
      type = CI,
      parm = params.profCI %||% "dose",
      maxit.profCI = maxit.profCI %||% 20,
      tol.profCI = tol.profCI %||% 1e-2
    )
  }

  ret
}

set.seed(123)
data("data", package = "ameras")
dosevars <- paste0("V", 1:10)


# Snapshot test
binomial_combos <- Filter(
  function(x) x$family == "binomial",
  snapshot_combinations
)

for (combo in binomial_combos) {
  label <- sprintf("%s_%s_deg%d", combo$family, combo$doseRRmod, combo$deg)

  test_that(paste("snapshot:", label), {
    fit <- fit_combination(
      family = combo$family,
      Y = combo$Y,
      doseRRmod = combo$doseRRmod,
      deg = combo$deg,
      X = "X1",
      M = NULL,
      data = data,
    )
    fit <- confint(fit, type = c("wald.orig", "percentile"))
    expect_snapshot(fit$RC$coefficients)
    expect_snapshot(fit$RC$sd)
    expect_snapshot(fit$RC$CI)
  })
}


# Test all non-RC methods with snapshot for a basic model
for (method in setdiff(all_methods, "RC")) {
  test_that(paste("binomial snapshot:", method), {
    if (method %in% c("ERC", "MCML", "BMA")) {
      skip_on_cran()
    }

    fit <- fit_combination(
      family = "binomial",
      Y = "Y.binomial",
      doseRRmod = "EXP",
      deg = 2,
      X = NULL,
      M = NULL,
      data = data,
      methods = method,
      niter.BMA = 1000,
      nburnin.BMA = 500
    )
    fit <- confint(fit, type = c("wald.orig", "percentile"))
    expect_snapshot(fit[[method]]$coefficients)
    expect_snapshot(fit[[method]]$sd)
    expect_snapshot(fit[[method]]$CI)
  })
}


# Basic no-error check for RC and all combinations of doseRRmod, deg, and lengths of X and M

for (combo in binomial_combos) {
  for (cov_combo in covariate_combinations) {
    X_label <- if (is.null(cov_combo$X)) {
      "NULL"
    } else {
      paste(cov_combo$X, collapse = "-")
    }
    M_label <- if (is.null(cov_combo$M)) {
      "NULL"
    } else {
      paste(cov_combo$M, collapse = "-")
    }
    label <- sprintf(
      "%s_%s_deg%d_X%s_M%s",
      combo$family,
      combo$doseRRmod,
      combo$deg,
      X_label,
      M_label
    )

    test_that(label, {
      expect_no_error({
        if (length(cov_combo$M) > 0) {
          fit_combination(
            family = combo$family,
            Y = combo$Y,
            deg = combo$deg,
            doseRRmod = combo$doseRRmod,
            X = cov_combo$X,
            M = cov_combo$M,
            data = data,
            methods = "RC",
            CI = "wald.orig",
            transform = transform1,
            transform.jacobian = transform1.jacobian,
            index.t = (length(cov_combo$X) + 2):(length(cov_combo$X) +
              length(cov_combo$M) * combo$deg +
              combo$deg +
              1)
          )
        } else {
          fit_combination(
            family = combo$family,
            Y = combo$Y,
            deg = combo$deg,
            doseRRmod = combo$doseRRmod,
            X = cov_combo$X,
            M = cov_combo$M,
            data = data,
            methods = "RC",
            CI = "wald.orig"
          )
        }
      })
    })
  }
}

# Test RC with all 45 combinations of doseRRmod, deg and lengths of X and M
#
# test_that("ameras binomial ERR deg=1 methods=RC X=c(X1,X2) M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1, methods="RC", X=c("X1","X2"), M=c("M1", "M2"), doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=4:6))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=c(X1,X2) M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=c("X1","X2"), M=c("M1", "M2"), doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=4:9))
# })
#
# test_that("ameras binomial ERR deg=1 methods=RC X=X1 M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=c("X1"), M=c("M1", "M2"), doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=3:5))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=X1 M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=c("X1"), M=c("M1", "M2"), doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=3:8))
# })
#
# test_that("ameras binomial ERR deg=1 methods=RC X=NULL M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=NULL, M=c("M1", "M2"), doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=2:4))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=NULL M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=NULL, M=c("M1", "M2"), doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=2:7))
# })
#
#
#
#
#
#
# test_that("ameras binomial ERR deg=1 methods=RC X=c(X1,X2) M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=c("X1","X2"), M="M1", doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=4:5))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=c(X1,X2) M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=c("X1","X2"), M="M1", doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=4:7))
# })
#
# test_that("ameras binomial ERR deg=1 methods=RC X=X1 M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=c("X1"), M="M1", doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=3:4))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=X1 M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=c("X1"), M="M1", doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=3:6))
# })
#
# test_that("ameras binomial ERR deg=1 methods=RC X=NULL M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=NULL, M="M1", doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=2:3))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=NULL M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=NULL, M="M1", doseRRmod="ERR", transform=transform1, transform.jacobian=transform1.jacobian, index.t=2:5))
# })
#
#
#
#
#
#
# test_that("ameras binomial ERR deg=1 methods=RC X=c(X1,X2) M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=c("X1","X2"), M=NULL, doseRRmod="ERR"))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=c(X1,X2) M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=c("X1","X2"), M=NULL, doseRRmod="ERR"))
# })
#
# test_that("ameras binomial ERR deg=1 methods=RC X=X1 M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=c("X1"), M=NULL, doseRRmod="ERR"))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=X1 M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=2, methods="RC", X=c("X1"), M=NULL, doseRRmod="ERR"))
# })
#
# test_that("ameras binomial ERR deg=1 methods=RC X=NULL M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial",  dosevars=dosevars, deg=1, methods="RC", X=NULL, M=NULL, doseRRmod="ERR"))
# })
#
# test_that("ameras binomial ERR deg=2 methods=RC X=NULL M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2, methods="RC", X=NULL, M=NULL, doseRRmod="ERR"))
# })
#
#
#
#
#
# #------------------------
#
#
# test_that("ameras binomial LINEXP   methods=RC X=c(X1,X2) M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,   methods="RC", X=c("X1","X2"), M=c("M1", "M2"), doseRRmod="LINEXP", transform=transform1, transform.jacobian=transform1.jacobian, index.t=4:9))
# })
#
# test_that("ameras binomial LINEXP   methods=RC X=X1 M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,  methods="RC", X=c("X1"), M=c("M1", "M2"), doseRRmod="LINEXP", transform=transform1, transform.jacobian=transform1.jacobian, index.t=3:8))
# })
#
#
# test_that("ameras binomial LINEXP   methods=RC X=NULL M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,  methods="RC", X=NULL, M=c("M1", "M2"), doseRRmod="LINEXP", transform=transform1, transform.jacobian=transform1.jacobian, index.t=2:7))
# })
#
#
#
#
#
#
# test_that("ameras binomial LINEXP   methods=RC X=c(X1,X2) M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,   methods="RC", X=c("X1","X2"), M="M1", doseRRmod="LINEXP", transform=transform1, transform.jacobian=transform1.jacobian, index.t=4:7))
# })
#
#
# test_that("ameras binomial LINEXP   methods=RC X=X1 M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,   methods="RC", X=c("X1"), M="M1", doseRRmod="LINEXP", transform=transform1, transform.jacobian=transform1.jacobian, index.t=3:6))
# })
#
#
# test_that("ameras binomial LINEXP   methods=RC X=NULL M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_warning(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,  methods="RC", X=NULL, M="M1", doseRRmod="LINEXP", transform=transform1, transform.jacobian=transform1.jacobian, index.t=2:5))
# })
#
#
#
#
#
#
#
# test_that("ameras binomial LINEXP   methods=RC X=c(X1,X2) M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,   methods="RC", X=c("X1","X2"), M=NULL, doseRRmod="LINEXP"))
# })
#
#
# test_that("ameras binomial LINEXP   methods=RC X=X1 M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,  methods="RC", X=c("X1"), M=NULL, doseRRmod="LINEXP"))
# })
#
# test_that("ameras binomial LINEXP   methods=RC X=NULL M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars,   methods="RC", X=NULL, M=NULL, doseRRmod="LINEXP"))
# })
#
#
# # ------------------------
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=c(X1,X2) M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,  methods="RC", X=c("X1","X2"), M=c("M1", "M2"), doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=c(X1,X2) M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2,  methods="RC", X=c("X1","X2"), M=c("M1", "M2"), doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=X1 M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1, methods="RC", X=c("X1"), M=c("M1", "M2"), doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=X1 M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2,  methods="RC", X=c("X1"), M=c("M1", "M2"), doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=NULL M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,  methods="RC", X=NULL, M=c("M1", "M2"), doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=NULL M=c(M1,M2)", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2, methods="RC", X=NULL, M=c("M1", "M2"), doseRRmod="EXP"))
# })
#
#
#
#
#
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=c(X1,X2) M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,  methods="RC", X=c("X1","X2"), M="M1", doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=c(X1,X2) M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2, methods="RC", X=c("X1","X2"), M="M1", doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=X1 M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,methods="RC", X=c("X1"), M="M1", doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=X1 M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2,  methods="RC", X=c("X1"), M="M1", doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=NULL M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,  methods="RC", X=NULL, M="M1", doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=NULL M=M1", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2,  methods="RC", X=NULL, M="M1", doseRRmod="EXP"))
# })
#
#
#
#
#
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=c(X1,X2) M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,  methods="RC", X=c("X1","X2"), M=NULL, doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=c(X1,X2) M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2,  methods="RC", X=c("X1","X2"), M=NULL, doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=X1 M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,  methods="RC", X=c("X1"), M=NULL, doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=X1 M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2,  methods="RC", X=c("X1"), M=NULL, doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=1 methods=RC X=NULL M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=1,  methods="RC", X=NULL, M=NULL, doseRRmod="EXP"))
# })
#
# test_that("ameras binomial  EXP deg=2 methods=RC X=NULL M=NULL", {
#   data("data", package="ameras")
#   dosevars <- paste0("V",1:10)
#   expect_no_error(ameras(data=data, family="binomial", Y="Y.binomial", dosevars=dosevars, deg=2,  methods="RC", X=NULL, M=NULL, doseRRmod="EXP"))
# })
#
#
#

ameras.mcml <- function(
  family,
  dosevars,
  data,
  deg,
  transform = NULL,
  transform.jacobian = NULL,
  Y = NULL,
  M = NULL,
  X = NULL,
  offset = NULL,
  inpar = NULL,
  entry = NULL,
  exit = NULL,
  status = NULL,
  setnr = setnr,
  doseRRmod = NULL,
  loglim = 1e-30,
  optim.method = "Nelder-Mead",
  control = list(reltol = 1e-10),
  ...
) {
  # if(length(CI)==0) stop("No CI method specified")
  # CI <- CI[CI %in% c("wald.orig","wald.transformed", "proflik")]
  # if(length(CI)>1) stop("Provide one type of CI for MCML: one of wald.orig, wald.transformed, and proflik")
  # if(length(CI)==0) stop("Incorrect CI method specified, should be one of wald.orig, wald.transformed, and proflik")
  # if(CI=="proflik" & !(params.profCI %in% c("dose","all"))) stop("Incorrect choice of parameters for profile likelihood CI supplied, should be either all or dose")
  # if(CI=="wald.transformed" & is.null(transform)) stop("No transformation specified, specify transformation or choose a different CI type")
  #
  # if(CI=="proflik") message("Note: computation times for profile likelihood intervals for MCML may be extensive with large datasets or complex models")

  if (family == "gaussian") {
    if (is.null(Y)) {
      stop("Y is required for family=gaussian")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 2 + length(X) + length(M) * deg + deg)
    }

    loglik.mcml <- function(params, ...) {
      logliks <- loglik.gaussian(
        params,
        D = dosevars,
        X = X,
        Y = Y,
        M = M,
        data = data,
        deg = deg,
        ERC = FALSE,
        loglim = loglim,
        transform = transform,
        ...
      )
      return(
        -1 *
          log(mean(exp(pmax(
            pmin(-1 * logliks - max(-1 * logliks), 7e1),
            -7e1
          )))) -
          max(-1 * logliks)
      ) # with explim=7e1
      #return(mean(logliks)) # mean of log likelihoods
    }

    parnames <- c(
      "(Intercept)",
      names(data[, X, drop = FALSE]),
      c("dose", "dose_squared")[1:deg]
    )
    if (!is.null(M)) {
      parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
      if (deg == 2) {
        parnames <- c(
          parnames,
          paste0("dose_squared:", names(data[, M, drop = FALSE]))
        )
      }
    }
    parnames <- c(parnames, "sigma")
  } else if (family == "binomial") {
    if (is.null(Y)) {
      stop("Y is required for family=binomial")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
    }

    loglik.mcml <- function(params, ...) {
      logliks <- loglik.binomial(
        params,
        D = dosevars,
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = FALSE,
        loglim = loglim,
        transform = transform,
        ...
      )
      #return(log(mean(exp(logliks-mean(logliks))))+mean(logliks)) # log of mean of likelihoods
      return(
        -1 *
          log(mean(exp(pmax(
            pmin(-1 * logliks - max(-1 * logliks), 7e1),
            -7e1
          )))) -
          max(-1 * logliks)
      ) # with explim=7e1
      #return(mean(logliks)) # mean of log likelihoods
    }

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "poisson") {
    if (is.null(Y)) {
      stop("Y is required for family=poisson")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
    }

    loglik.mcml <- function(params, ...) {
      logliks <- loglik.poisson(
        params,
        D = dosevars,
        X = X,
        Y = Y,
        offset = offset,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        loglim = loglim,
        transform = transform,
        ...
      )
      #return(log(mean(exp(logliks-mean(logliks))))+mean(logliks)) # log of mean of likelihoods
      return(
        -1 *
          log(mean(exp(pmax(
            pmin(-1 * logliks - max(-1 * logliks), 7e1),
            -7e1
          )))) -
          max(-1 * logliks)
      ) # with explim=7e1
      #return(mean(logliks)) # mean of log likelihoods
    }

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "clogit") {
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=clogit")
    }
    if (is.null(status)) {
      stop("status is required for family=clogit")
    }

    designmat <- t(model.matrix(~ as.factor(data[, setnr]) - 1))

    if (is.null(inpar)) {
      inpar <- rep(0, length(X) + length(M) * deg + deg)
    }

    loglik.mcml <- function(params, ...) {
      logliks <- loglik.clogit(
        params,
        D = dosevars,
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        designmat = designmat,
        data = data,
        deg = deg,
        ERC = FALSE,
        loglim = loglim,
        transform = transform,
        ...
      )
      return(
        -1 *
          log(mean(exp(pmax(
            pmin(-1 * logliks - max(-1 * logliks), 7e1),
            -7e1
          )))) -
          max(-1 * logliks)
      ) # with explim=7e1
      #return(mean(logliks)) # mean of log likelihoods
    }

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "prophaz") {
    if (is.null(exit)) {
      stop("exit is required for family=prophaz")
    }
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=prophaz")
    }
    if (is.null(status)) {
      stop("status is required for family=prophaz")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, length(X) + length(M) * deg + deg)
    }

    loglik.mcml <- function(params, ...) {
      logliks <- loglik.prophaz(
        params,
        D = dosevars,
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        entry = entry,
        exit = exit,
        data = data,
        deg = deg,
        ERC = FALSE,
        loglim = loglim,
        transform = transform,
        ...
      )
      return(
        -1 *
          log(mean(exp(pmax(
            pmin(-1 * logliks - max(-1 * logliks), 7e1),
            -7e1
          )))) -
          max(-1 * logliks)
      ) # with explim=7e1
      #return(mean(logliks)) # mean of log likelihoods
    }

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "multinomial") {
    if (is.null(Y)) {
      stop("Y is required for family=multinomial")
    }

    if (is.null(inpar)) {
      inpar <- rep(
        0,
        (length(unique(data[, Y])) - 1) *
          (1 + length(X) + length(M) * deg + deg)
      )
    }

    loglik.mcml <- function(params, ...) {
      logliks <- loglik.multinomial(
        params,
        D = dosevars,
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = FALSE,
        loglim = loglim,
        transform = transform,
        ...
      )
      #return(log(mean(exp(logliks-mean(logliks))))+mean(logliks)) # log of mean of likelihoods
      return(
        -1 *
          log(mean(exp(pmax(
            pmin(-1 * logliks - max(-1 * logliks), 7e1),
            -7e1
          )))) -
          max(-1 * logliks)
      ) # with explim=7e1
      #return(mean(logliks)) # mean of log likelihoods
    }

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }

    mylv <- levels(data[, Y])

    mylv <- mylv[-length(mylv)]

    parnames <- do.call(
      "c",
      lapply(mylv, function(lv) paste0("(", lv, ")_", parnames))
    )
  }
  t0 <- proc.time()
  if (length(parnames) == 1) {
    # Optimize 1-dimensional model: use optimize instead of optim
    fit0 <- optimize(f = loglik.mcml, lower = -20, upper = 5, ...)
    fit <- list(
      par = fit0$minimum,
      value = fit0$objective,
      convergence = 0,
      hessian = numDeriv::hessian(func = loglik.mcml, x = fit0$minimum, ...)
    )
  } else {
    fit <- optim(
      inpar,
      loglik.mcml,
      method = optim.method,
      control = control,
      ...
    )
    if (optim.method == "Nelder-Mead") {
      count0 <- fit$counts
      fit <- optim(
        fit$par,
        loglik.mcml,
        method = "BFGS",
        control = control,
        ...
      )
      fit$counts <- replace(count0, is.na(count0), 0) +
        replace(fit$counts, is.na(fit$counts), 0)
    }
    fit$hessian <- numDeriv::hessian(func = loglik.mcml, x = fit$par, ...)
  }

  if (!is.null(transform) & !is.null(transform.jacobian)) {
    if (is.function(transform) & is.function(transform.jacobian)) {
      if ("boundcheck" %in% names(formals(transform))) {
        coefs <- transform(fit$par, boundcheck = TRUE, ...)
      } else {
        coefs <- transform(fit$par, ...)
      }
      if (
        det(fit$hessian) != 0 &
          rcond(fit$hessian) > .Machine$double.eps &
          all(eigen(fit$hessian)$values > 0)
      ) {
        jac <- transform.jacobian(fit$par, ...)
        vcov <- jac %*% solve(fit$hessian) %*% t(jac)
      } else {
        warning(
          "WARNING: Hessian was not invertible or inverse was not positive definite, variance matrix could not be obtained"
        )
        vcov <- matrix(NA, ncol = length(parnames), nrow = length(parnames))
      }
    } else {
      stop("transform and transform.jacobian should be functions")
    }
  } else {
    coefs <- fit$par
    if (
      det(fit$hessian) != 0 &
        rcond(fit$hessian) > .Machine$double.eps &
        all(eigen(fit$hessian)$values > 0)
    ) {
      vcov <- solve(fit$hessian)
    } else {
      warning(
        "WARNING: Hessian was not invertible or inverse was not positive definite, variance matrix could not be obtained"
      )
      vcov <- matrix(NA, ncol = length(parnames), nrow = length(parnames))
    }
  }

  names(coefs) <- parnames
  rownames(vcov) <- colnames(vcov) <- parnames

  t1 <- proc.time()
  timedif <- t1 - t0
  runtime <- paste(
    round(as.numeric(as.difftime(timedif["elapsed"], units = "secs")), 1),
    "seconds"
  )

  out <- list(
    coefficients = coefs,
    sd = sqrt(diag(vcov)),
    vcov = vcov,
    optim = list(
      par = fit$par,
      hessian = fit$hessian,
      convergence = fit$convergence,
      counts = fit$counts
    ),
    loglik = -1 * fit$value,
    runtime = runtime
  )
  return(out)
}

ameras.rc <- function(
  family,
  dosevars,
  data,
  deg,
  ERC = FALSE,
  transform = NULL,
  transform.jacobian = NULL,
  Y = NULL,
  M = NULL,
  X = NULL,
  offset = NULL,
  inpar = NULL,
  entry = NULL,
  exit = NULL,
  status = NULL,
  setnr = NULL,
  doseRRmod = NULL,
  loglim = 1e-30,
  optim.method = "Nelder-Mead",
  control = list(reltol = 1e-10),
  ...
) {
  if (ERC & family != "poisson") {
    Kmat <- cov(t(data[, dosevars]))
  } else {
    Kmat <- NULL
  }

  t0 <- proc.time()
  if (family == "gaussian") {
    if (is.null(Y)) {
      stop("Y is required for family=gaussian")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 2 + length(X) + length(M) * deg + deg)
    }

    fit <- optim(
      inpar,
      loglik.gaussian,
      D = "rcdose_ameras",
      X = X,
      Y = Y,
      M = M,
      data = data,
      deg = deg,
      ERC = ERC,
      Kmat = Kmat,
      loglim = loglim,
      transform = transform,
      method = optim.method,
      control = control,
      ...
    )
    if (optim.method == "Nelder-Mead") {
      count0 <- fit$counts
      fit <- optim(
        fit$par,
        loglik.gaussian,
        D = "rcdose_ameras",
        X = X,
        Y = Y,
        M = M,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        method = "BFGS",
        control = control,
        ...
      )
      fit$counts <- replace(count0, is.na(count0), 0) +
        replace(fit$counts, is.na(fit$counts), 0)
    }
    fit$hessian <- numDeriv::hessian(
      func = loglik.gaussian,
      x = fit$par,
      D = "rcdose_ameras",
      X = X,
      Y = Y,
      M = M,
      data = data,
      deg = deg,
      ERC = ERC,
      Kmat = Kmat,
      loglim = loglim,
      transform = transform,
      ...
    )

    parnames <- c(
      "(Intercept)",
      names(data[, X, drop = FALSE]),
      c("dose", "dose_squared")[1:deg]
    )
    if (!is.null(M)) {
      parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))

      if (deg == 2) {
        parnames <- c(
          parnames,
          paste0("dose_squared:", names(data[, M, drop = FALSE]))
        )
      }
    }
    parnames <- c(parnames, "sigma")
  } else if (family == "binomial") {
    if (is.null(Y)) {
      stop("Y is required for family=binomial")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
    }

    fit <- optim(
      inpar,
      loglik.binomial,
      D = "rcdose_ameras",
      X = X,
      Y = Y,
      M = M,
      doseRRmod = doseRRmod,
      data = data,
      deg = deg,
      ERC = ERC,
      Kmat = Kmat,
      loglim = loglim,
      transform = transform,
      method = optim.method,
      control = control,
      ...
    )
    if (optim.method == "Nelder-Mead") {
      count0 <- fit$counts
      fit <- optim(
        fit$par,
        loglik.binomial,
        D = "rcdose_ameras",
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        method = "BFGS",
        control = control,
        ...
      )
      fit$counts <- replace(count0, is.na(count0), 0) +
        replace(fit$counts, is.na(fit$counts), 0)
    }

    fit$hessian <- numDeriv::hessian(
      func = loglik.binomial,
      x = fit$par,
      D = "rcdose_ameras",
      X = X,
      Y = Y,
      M = M,
      doseRRmod = doseRRmod,
      data = data,
      deg = deg,
      ERC = ERC,
      Kmat = Kmat,
      loglim = loglim,
      transform = transform,
      ...
    )

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "poisson") {
    if (is.null(Y)) {
      stop("Y is required for family=poisson")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
    }
    if (ERC) {
      fit <- optim(
        inpar,
        loglik.poisson.erc,
        D = dosevars,
        X = X,
        Y = Y,
        offset = offset,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        loglim = loglim,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )
    } else {
      fit <- optim(
        inpar,
        loglik.poisson,
        D = "rcdose_ameras",
        X = X,
        Y = Y,
        offset = offset,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        loglim = loglim,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )
    }
    if (optim.method == "Nelder-Mead") {
      count0 <- fit$counts
      if (ERC) {
        fit <- optim(
          fit$par,
          loglik.poisson.erc,
          D = dosevars,
          X = X,
          Y = Y,
          offset = offset,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          loglim = loglim,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
      } else {
        fit <- optim(
          fit$par,
          loglik.poisson,
          D = "rcdose_ameras",
          X = X,
          Y = Y,
          offset = offset,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          loglim = loglim,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
      }
      fit$counts <- replace(count0, is.na(count0), 0) +
        replace(fit$counts, is.na(fit$counts), 0)
    }

    if (ERC) {
      fit$hessian <- numDeriv::hessian(
        func = loglik.poisson.erc,
        x = fit$par,
        D = dosevars,
        X = X,
        Y = Y,
        offset = offset,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        loglim = loglim,
        transform = transform,
        ...
      )
    } else {
      fit$hessian <- numDeriv::hessian(
        func = loglik.poisson,
        x = fit$par,
        D = "rcdose_ameras",
        X = X,
        Y = Y,
        offset = offset,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        loglim = loglim,
        transform = transform,
        ...
      )
    }
    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family %in% c("clogit")) {
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=prophaz")
    }
    if (is.null(status)) {
      stop("status is required for family=prophaz")
    }

    designmat <- t(model.matrix(~ as.factor(data[, setnr]) - 1))

    if (is.null(inpar)) {
      inpar <- rep(0, length(X) + length(M) * deg + deg)
    }

    if (length(X) + length(M) * deg + deg == 1) {
      # Optimize 1-dimensional model: use optimize instead of optim

      fit0 <- optimize(
        f = loglik.clogit,
        lower = -20,
        upper = 5,
        D = "rcdose_ameras",
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        designmat = designmat,
        entry = entry,
        exit = exit,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        ...
      )

      fit <- list(
        par = fit0$minimum,
        value = fit0$objective,
        convergence = 0,
        hessian = numDeriv::hessian(
          func = loglik.clogit,
          x = fit0$minimum,
          D = "rcdose_ameras",
          status = status,
          X = X,
          M = M,
          doseRRmod = doseRRmod,
          designmat = designmat,
          entry = entry,
          exit = exit,
          data = data,
          deg = deg,
          ERC = ERC,
          Kmat = Kmat,
          loglim = loglim,
          transform = transform,
          ...
        )
      )
    } else {
      fit <- optim(
        inpar,
        loglik.clogit,
        D = "rcdose_ameras",
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        designmat = designmat,
        entry = entry,
        exit = exit,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )

      if (optim.method == "Nelder-Mead") {
        count0 <- fit$counts
        fit <- optim(
          fit$par,
          loglik.clogit,
          D = "rcdose_ameras",
          status = status,
          X = X,
          M = M,
          doseRRmod = doseRRmod,
          designmat = designmat,
          entry = entry,
          exit = exit,
          data = data,
          deg = deg,
          ERC = ERC,
          Kmat = Kmat,
          loglim = loglim,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
        fit$counts <- replace(count0, is.na(count0), 0) +
          replace(fit$counts, is.na(fit$counts), 0)
      }
      fit$hessian <- numDeriv::hessian(
        func = loglik.clogit,
        x = fit$par,
        D = "rcdose_ameras",
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        designmat = designmat,
        entry = entry,
        exit = exit,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        ...
      )
    }

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "prophaz") {
    if (is.null(exit)) {
      stop("exit is required for family=prophaz")
    }
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=prophaz")
    }
    if (is.null(status)) {
      stop("status is required for family=prophaz")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, length(X) + length(M) * deg + deg)
    }

    if (length(X) + length(M) * deg + deg == 1) {
      # Optimize 1-dimensional model: use optimize instead of optim
      fit0 <- optimize(
        f = loglik.prophaz,
        lower = -20,
        upper = 5,
        D = "rcdose_ameras",
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        entry = entry,
        exit = exit,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        ...
      )
      fit <- list(
        par = fit0$minimum,
        value = fit0$objective,
        convergence = 0,
        hessian = numDeriv::hessian(
          func = loglik.prophaz,
          x = fit0$minimum,
          D = "rcdose_ameras",
          status = status,
          X = X,
          M = M,
          doseRRmod = doseRRmod,
          entry = entry,
          exit = exit,
          data = data,
          deg = deg,
          ERC = ERC,
          Kmat = Kmat,
          loglim = loglim,
          transform = transform,
          ...
        )
      )
    } else {
      fit <- optim(
        inpar,
        loglik.prophaz,
        D = "rcdose_ameras",
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        entry = entry,
        exit = exit,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )

      if (optim.method == "Nelder-Mead") {
        count0 <- fit$counts
        fit <- optim(
          fit$par,
          loglik.prophaz,
          D = "rcdose_ameras",
          status = status,
          X = X,
          M = M,
          doseRRmod = doseRRmod,
          entry = entry,
          exit = exit,
          data = data,
          deg = deg,
          ERC = ERC,
          Kmat = Kmat,
          loglim = loglim,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
        fit$counts <- replace(count0, is.na(count0), 0) +
          replace(fit$counts, is.na(fit$counts), 0)
      }
      fit$hessian <- numDeriv::hessian(
        func = loglik.prophaz,
        x = fit$par,
        D = "rcdose_ameras",
        status = status,
        X = X,
        M = M,
        doseRRmod = doseRRmod,
        entry = entry,
        exit = exit,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        ...
      )
    }

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "multinomial") {
    if (is.null(Y)) {
      stop("Y is required for family=multinomial")
    }

    if (is.null(inpar)) {
      inpar <- rep(
        0,
        (length(unique(data[, Y])) - 1) *
          (1 + length(X) + length(M) * deg + deg)
      )
    }

    fit <- optim(
      inpar,
      loglik.multinomial,
      D = "rcdose_ameras",
      X = X,
      Y = Y,
      M = M,
      doseRRmod = doseRRmod,
      data = data,
      deg = deg,
      ERC = ERC,
      Kmat = Kmat,
      loglim = loglim,
      transform = transform,
      method = optim.method,
      control = control,
      ...
    )
    if (optim.method == "Nelder-Mead") {
      count0 <- fit$counts
      fit <- optim(
        fit$par,
        loglik.multinomial,
        D = "rcdose_ameras",
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = ERC,
        Kmat = Kmat,
        loglim = loglim,
        transform = transform,
        method = "BFGS",
        control = control,
        ...
      )
      fit$counts <- replace(count0, is.na(count0), 0) +
        replace(fit$counts, is.na(fit$counts), 0)
    }
    fit$hessian <- numDeriv::hessian(
      func = loglik.multinomial,
      x = fit$par,
      D = "rcdose_ameras",
      X = X,
      Y = Y,
      M = M,
      doseRRmod = doseRRmod,
      data = data,
      deg = deg,
      ERC = ERC,
      Kmat = Kmat,
      loglim = loglim,
      transform = transform,
      ...
    )

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }

    mylv <- levels(data[, Y])

    mylv <- mylv[-length(mylv)]

    parnames <- do.call(
      "c",
      lapply(mylv, function(lv) paste0("(", lv, ")_", parnames))
    )
  }

  if (!is.null(transform) & !is.null(transform.jacobian)) {
    if (is.function(transform) & is.function(transform.jacobian)) {
      if ("boundcheck" %in% names(formals(transform))) {
        coefs <- transform(fit$par, boundcheck = TRUE, ...)
      } else {
        coefs <- transform(fit$par, ...)
      }

      if (
        det(fit$hessian) != 0 &
          rcond(fit$hessian) > .Machine$double.eps &
          all(eigen(fit$hessian)$values > 0)
      ) {
        jac <- transform.jacobian(fit$par, ...)
        vcov <- jac %*% solve(fit$hessian) %*% t(jac)
      } else {
        warning(
          "WARNING: Hessian was not invertible or inverse was not positive definite, variance matrix could not be obtained"
        )
        vcov <- matrix(NA, ncol = length(parnames), nrow = length(parnames))
      }
    } else {
      stop("transform and transform.jacobian should be functions")
    }
  } else {
    coefs <- fit$par
    if (
      det(fit$hessian) != 0 &
        rcond(fit$hessian) > .Machine$double.eps &
        all(eigen(fit$hessian)$values > 0)
    ) {
      vcov <- solve(fit$hessian)
    } else {
      warning(
        "WARNING: Hessian was not invertible or inverse was not positive definite, variance matrix could not be obtained"
      )
      vcov <- matrix(NA, ncol = length(parnames), nrow = length(parnames))
    }
  }

  names(coefs) <- parnames
  rownames(vcov) <- colnames(vcov) <- parnames

  t1 <- proc.time()
  timedif <- t1 - t0
  runtime <- paste(
    round(as.numeric(as.difftime(timedif["elapsed"], units = "secs")), 1),
    "seconds"
  )

  out <- list(
    coefficients = coefs,
    sd = sqrt(diag(vcov)),
    vcov = vcov,
    optim = list(
      par = fit$par,
      hessian = fit$hessian,
      convergence = fit$convergence,
      counts = fit$counts
    ),
    loglik = -1 * fit$value,
    runtime = runtime,
    ERC = ERC
  )

  return(out)
}

loglik.binomial <- function(
  params,
  Y,
  D,
  M = NULL,
  X = NULL,
  data,
  doseRRmod,
  ERC = FALSE,
  Kmat = NULL,
  deg = 1,
  loglim = 1e-30,
  transform = NULL,
  ...
) {
  # params = c(a0, a1, ..., ap, b1, b2, (bm1), (bm2))

  if (length(params) != deg * length(M) + deg + length(X) + 1) {
    stop("Parameter vector length mismatch")
  }
  if (ERC & length(D) > 1) {
    stop(
      "ERC only works with one supplied dose (i.e., the mean across replicates)"
    )
  }
  # doseRRmod = ERR or EXP
  if (!is.null(transform)) {
    if (is.function(transform)) {
      params <- transform(params = params, ...)
    } else {
      stop("transform should be a function")
    }
  }

  a0 <- params[1]

  if (!is.null(X)) {
    a <- params[2:(length(X) + 1)]
    Xlinpred <- c(as.matrix(data[, X]) %*% a)
  } else {
    Xlinpred <- 0
  }

  b1 <- params[length(X) + 2]

  if (deg == 2) {
    b2 <- params[length(X) + 3]
  } else if (deg == 1) {
    b2 <- NULL
  }

  if (!is.null(M)) {
    bm1 <- params[(2 + deg + length(X)):(length(M) + deg + length(X) + 1)]
    if (deg == 2) {
      bm2 <- params[
        (length(M) + 2 + deg + length(X)):(2 * length(M) + deg + length(X) + 1)
      ]
    } else {
      bm2 <- NULL
    }
  } else {
    bm1 <- bm2 <- NULL
  }

  if (deg == 2) {
    betavec <- c(b1, b2, bm1, bm2)
  } else {
    betavec <- c(b1, bm1)
  }

  A <- exp(pmin(a0 + Xlinpred, 7e1)) *
    exposureRR(
      params = betavec,
      D = D,
      M = M,
      data = data,
      doseRRmod = doseRRmod,
      deg = deg
    )

  if (any(A < 0)) {
    stop("RR <0, please check supplied transformation")
  }
  p <- A / (1 + A)
  p <- pmin(pmax(p, 1e-10), 1 - 1e-10)

  if (length(D) > 1) {
    ls <- colSums(log(p^data[, Y] * (1 - p)^(1 - data[, Y])))
    ls <- as.numeric(ls)
  } else {
    ls <- sum(log(p^data[, Y] * (1 - p)^(1 - data[, Y])))
  }
  if (ERC) {
    derivs <- dRRdD(
      params = betavec,
      D = D,
      M = M,
      data = data,
      doseRRmod = doseRRmod,
      deg = deg
    )
    dpdk <- exp(pmin(a0 + Xlinpred, 7e1)) * derivs$first / (1 + A)^2
    dpdk2 <- exp(pmin(a0 + Xlinpred, 7e1)) *
      derivs$second /
      (1 + A)^2 -
      2 * exp(pmin(a0 + Xlinpred, 7e1))^2 * derivs$first^2 / (1 + A)^3

    mymat <- tcrossprod((-1)^(1 - data[, Y])) *
      tcrossprod(dpdk) /
      tcrossprod(p^data[, Y] * (1 - p)^(1 - data[, Y]))

    diag(mymat) <- (-1)^(1 - data[, Y]) *
      dpdk2 /
      (p^data[, Y] * (1 - p)^(1 - data[, Y]))

    return(-1 * (ls + log(max(1 + .5 * sum(mymat * Kmat), loglim))))
  } else {
    return(-1 * ls)
  }
}


loglik.poisson <- function(
  params,
  Y,
  D,
  X = NULL,
  offset = NULL,
  M = NULL,
  doseRRmod,
  data,
  deg = 1,
  loglim = 1e-30,
  transform = NULL,
  ...
) {
  # C++ version
  # params = c(a0, a1, ..., ap, b1, b2, (bm1), (bm2))

  if (length(params) != deg * length(M) + deg + length(X) + 1) {
    stop("Parameter vector length mismatch")
  }

  if (!is.null(transform)) {
    if (is.function(transform)) {
      params <- transform(params = params, ...)
    } else {
      stop("transform should be a function")
    }
  }

  if (is.null(offset)) {
    offset <- 1
  } else {
    offset <- data[, offset]
  }

  a0 <- params[1]

  if (!is.null(X)) {
    a <- params[2:(length(X) + 1)]
    Xlinpred <- c(as.matrix(data[, X]) %*% a)
  } else {
    Xlinpred <- 0
  }

  b1 <- params[length(X) + 2]

  if (deg == 2) {
    b2 <- params[length(X) + 3]
  } else if (deg == 1) {
    b2 <- 0
  }

  if (!is.null(M)) {
    bm1 <- params[(2 + deg + length(X)):(length(M) + deg + length(X) + 1)]
    if (deg == 2) {
      bm2 <- params[
        (length(M) + 2 + deg + length(X)):(2 * length(M) + deg + length(X) + 1)
      ]
    } else {
      bm2 <- NULL
    }
  } else {
    bm1 <- bm2 <- NULL
  }

  if (deg == 2) {
    betavec <- c(b1, b2, bm1, bm2)
  } else {
    betavec <- c(b1, bm1)
  }

  mus <- pmax(
    exp(pmin(a0 + Xlinpred, 7e1)) *
      pmax(
        exposureRR(
          params = betavec,
          D = D,
          M = M,
          data = data,
          doseRRmod = doseRRmod,
          deg = deg
        ),
        loglim
      ) *
      offset,
    loglim
  )
  if (any(mus < 0)) {
    stop("RR <0, please check supplied transformation")
  }

  if (length(D) > 1) {
    #ls <- colSums(data[,Y]*log(mus)-mus-lfactorial(data[,Y]))
    ls <- colSums(
      data[, Y] *
        log(mus) -
        mus -
        ifelse(data[, Y] > 0, data[, Y] * log(data[, Y]) - data[, Y], 0)
    ) # Stirling's approximation like Mark
    ls <- as.numeric(ls)
  } else {
    #ls <- sum(data[,Y]*log(mus)-mus-lfactorial(data[,Y]))
    ls <- sum(
      data[, Y] *
        log(mus) -
        mus -
        ifelse(data[, Y] > 0, data[, Y] * log(data[, Y]) - data[, Y], 0)
    ) # Stirling's approximation like Mark
  }

  return(-1 * ls)
}


loglik.poisson.erc <- function(
  params,
  Y,
  D,
  X = NULL,
  offset = NULL,
  M = NULL,
  doseRRmod,
  data,
  deg = 1,
  loglim = 1e-30,
  transform = NULL,
  ...
) {
  # C++ version
  # params = c(a0, a1, ..., ap, b1, b2, (bm1), (bm2))

  if (length(params) != deg * length(M) + deg + length(X) + 1) {
    stop("Parameter vector length mismatch")
  }
  #if(ERC & is.null(Kmat)) stop("Kmat necessary for ERC")
  #if(ERC & length(D)>1) stop("ERC only works with one supplied dose (i.e., the mean across replicates)")

  dosemat <- as.matrix(data[, D])

  if (!is.null(transform)) {
    if (is.function(transform)) {
      params <- transform(params = params, ...)
    } else {
      stop("transform should be a function")
    }
  }

  if (is.null(offset)) {
    offset <- 1
  } else {
    offset <- data[, offset]
  }

  a0 <- params[1]

  if (!is.null(X)) {
    a <- params[2:(length(X) + 1)]
    Xlinpred <- c(as.matrix(data[, X]) %*% a)
  } else {
    Xlinpred <- 0
  }

  b1 <- params[length(X) + 2]

  if (deg == 2) {
    b2 <- params[length(X) + 3]
  } else if (deg == 1) {
    b2 <- 0
  }

  if (!is.null(M)) {
    bm1 <- params[(2 + deg + length(X)):(length(M) + deg + length(X) + 1)]
    if (deg == 2) {
      bm2 <- params[
        (length(M) + 2 + deg + length(X)):(2 * length(M) + deg + length(X) + 1)
      ]
    } else {
      bm2 <- NULL
    }
  } else {
    bm1 <- bm2 <- NULL
  }

  if (deg == 2) {
    betavec <- c(b1, b2, bm1, bm2)
  } else {
    betavec <- c(b1, bm1)
  }

  #data$rcdose_ameras <- rowMeans(dosemat)

  mus <- pmax(
    exp(pmin(a0 + Xlinpred, 7e1)) *
      pmax(
        exposureRR(
          params = betavec,
          D = "rcdose_ameras",
          M = M,
          data = data,
          doseRRmod = doseRRmod,
          deg = deg
        ),
        loglim
      ) *
      offset,
    loglim
  )
  if (any(mus < 0)) {
    stop("RR <0, please check supplied transformation")
  }

  #ls <- sum(data[,Y]*log(mus)-mus-lfactorial(data[,Y]))
  ls <- sum(
    data[, Y] *
      log(mus) -
      mus -
      ifelse(data[, Y] > 0, data[, Y] * log(data[, Y]) - data[, Y], 0)
  ) # Stirling's approximation like Mark

  derivs <- dRRdD(
    params = betavec,
    D = "rcdose_ameras",
    M = M,
    data = data,
    doseRRmod = doseRRmod,
    deg = deg
  )
  dmdd <- exp(pmin(a0 + Xlinpred, 7e1)) * derivs$first * offset
  dmdd2 <- exp(pmin(a0 + Xlinpred, 7e1)) * derivs$second * offset

  v <- (data[, Y] / mus - 1) * dmdd

  # center dose for covariance
  Xc <- dosemat - data[, "rcdose_ameras"]
  Xt_v <- crossprod(Xc, v)

  term1 <- sum(Xt_v^2) / (ncol(dosemat) - 1)

  # diagonal correction
  corrterm <- (data[, Y] / mus - 1) * dmdd2 - data[, Y] / mus^2 * dmdd^2

  # diag(Kmat) = row variances of dosemat
  row_var <- rowSums(Xc^2) / (ncol(dosemat) - 1)

  term2 <- sum(row_var * corrterm)

  val <- term1 + term2

  return(-1 * (ls + log(max(1 + .5 * val, loglim))))
}


loglik.gaussian <- function(
  params,
  D,
  Y,
  X = NULL,
  M = NULL,
  data,
  ERC = FALSE,
  Kmat = NULL,
  deg = 1,
  loglim = 1e-30,
  transform = NULL,
  ...
) {
  # params = c(a0, a1, ..., ap, b1, b2, (bm1), (bm2), sigma)

  if (length(params) != deg * length(M) + deg + length(X) + 2) {
    stop("Parameter vector length mismatch")
  }
  if (ERC & is.null(Kmat)) {
    stop("Kmat necessary for ERC")
  }
  if (ERC & length(D) > 1) {
    stop(
      "ERC only works with one supplied dose (i.e., the mean across replicates)"
    )
  }

  if (!is.null(transform)) {
    if (is.function(transform)) {
      params <- transform(params = params, ...)
    } else {
      stop("transform should be a function")
    }
  }

  a0 <- params[1]

  if (!is.null(X)) {
    a <- params[2:(length(X) + 1)]
    Xlinpred <- c(as.matrix(data[, X]) %*% a)
  } else {
    Xlinpred <- 0
  }

  b1 <- params[length(X) + 2]

  if (deg == 2) {
    b2 <- params[length(X) + 3]
  } else if (deg == 1) {
    b2 <- 0
  }

  if (!is.null(M)) {
    bm1 <- params[(2 + deg + length(X)):(length(M) + deg + length(X) + 1)]
    if (deg == 2) {
      bm2 <- params[
        (length(M) + 2 + deg + length(X)):(2 * length(M) + deg + length(X) + 1)
      ]
    } else {
      bm2 <- NULL
    }
  } else {
    bm1 <- bm2 <- NULL
  }

  if (deg == 2) {
    betavec <- c(b1, b2, bm1, bm2)
  } else {
    betavec <- c(b1, bm1)
  }

  sigma <- params[deg * length(M) + deg + length(X) + 2]

  #L <- (1/(2*pi*sigma^2))^(length(Y)/2)*exp(-sum((Y-alpha-beta1*D-beta2*D^2)^2)/(2*sigma^2))

  # Here I use mu = a0 + X^Ta + RR - 1 with RR using the linear ERR
  mus <- a0 +
    Xlinpred +
    exposureRR(
      params = betavec,
      D = D,
      M = M,
      data = data,
      doseRRmod = "ERR",
      deg = deg
    ) -
    1

  ls <- -log(2 * pi * sigma^2) *
    nrow(data) /
    2 -
    colSums((data[, Y] - as.matrix(mus, ncol = length(D)))^2) / (2 * sigma^2)

  ls <- as.numeric(ls)

  if (ERC) {
    # See above, dmu/dD = dRR/dD

    derivs <- dRRdD(
      params = betavec,
      D = D,
      M = M,
      data = data,
      doseRRmod = "ERR",
      deg = deg
    )
    dmdd <- derivs$first
    dmdd2 <- derivs$second

    mymat <- (tcrossprod(dmdd) / sigma^2) *
      (tcrossprod(data[, Y] - mus) /
        sigma^2 -
        diag(1, nrow = nrow(data), ncol = nrow(data)))
    diag(mymat) <- diag(mymat) + dmdd2 / sigma^2 * (data[, Y] - mus)

    #res <- compute_weighted_sum_kahan(dmdd, dmdd2, data[,Y]-mus, Kmat, sigma)

    return(-1 * (ls + log(max(1 + .5 * sum(mymat * Kmat), loglim))))
    #return(-1*(l+log(max(1+.5*res,loglim))))
  } else {
    return(-1 * ls)
  }
}


loglik.clogit <- function(
  params,
  D,
  status,
  X = NULL,
  M = NULL,
  doseRRmod,
  designmat,
  data,
  deg = 1,
  ERC = FALSE,
  Kmat = NULL,
  loglim = 1e-30,
  transform = NULL,
  ...
) {
  # params = c(a1, ..., ap, b1, b2, (bm1), (bm2))
  #print(params)
  if (length(params) != deg * length(M) + deg + length(X)) {
    stop("Parameter vector length mismatch")
  }
  if (ERC & is.null(Kmat)) {
    stop("Kmat necessary for ERC")
  }
  if (ERC & length(D) > 1) {
    stop(
      "ERC only works with one supplied dose (i.e., the mean across replicates)"
    )
  }

  if (!is.null(transform)) {
    if (is.function(transform)) {
      params <- transform(params = params, ...)
    } else {
      stop("transform should be a function")
    }
  }

  if (!is.null(X)) {
    a <- params[1:length(X)]
    Xlinpred <- c(as.matrix(data[, X]) %*% a)
  } else {
    Xlinpred <- 0
  }

  b1 <- params[length(X) + 1]

  if (deg == 2) {
    b2 <- params[length(X) + 2]
  } else if (deg == 1) {
    b2 <- 0
  }

  if (!is.null(M)) {
    bm1 <- params[(1 + deg + length(X)):(length(M) + deg + length(X))]
    if (deg == 2) {
      bm2 <- params[
        (length(M) + 1 + deg + length(X)):(2 * length(M) + deg + length(X))
      ]
    } else {
      bm2 <- NULL
    }
  } else {
    bm1 <- bm2 <- NULL
  }

  if (deg == 2) {
    betavec <- c(b1, b2, bm1, bm2)
  } else {
    betavec <- c(b1, bm1)
  }

  RRs <- exp(pmin(Xlinpred, 7e1)) *
    exposureRR(
      params = betavec,
      D = D,
      M = M,
      data = data,
      doseRRmod = doseRRmod,
      deg = deg
    )

  if (any(RRs < 0)) {
    stop("RR <0, please check supplied transformation")
  }
  ls <- colSums(
    log(RRs[data[, status] == 1]) -
      log(designmat %*% as.matrix(RRs, ncol = length(D)))
  )
  ls <- as.numeric(ls)

  if (ERC) {
    derivs <- dRRdD(
      params = betavec,
      D = D,
      M = M,
      data = data,
      doseRRmod = doseRRmod,
      deg = deg
    )
    drdd <- exp(pmin(Xlinpred, 7e1)) * derivs$first
    drdd2 <- exp(pmin(Xlinpred, 7e1)) * derivs$second

    tmp <- c(dldd_clogit(designmat, RRs))
    dldd <- drdd / RRs * (data[, status] == 1) - tmp * drdd

    mymat <- compute_ERCmatrix_clogit(
      designmat,
      RRs,
      drdd,
      drdd2,
      as.integer(data[, status])
    )

    return(
      -1 *
        (ls + log(max(1 + .5 * sum((tcrossprod(dldd) + mymat) * Kmat), loglim)))
    )
  } else {
    return(-1 * ls)
  }
}

loglik.prophaz <- function(
  params,
  D,
  status,
  X = NULL,
  M = NULL,
  doseRRmod,
  data,
  deg = 1,
  entry = NULL,
  exit,
  ERC = FALSE,
  Kmat = NULL,
  loglim = 1e-30,
  transform = NULL,
  ...
) {
  # params = c(a1, ..., ap, b1, b2, (bm1), (bm2))
  #print(params)
  if (length(params) != deg * length(M) + deg + length(X)) {
    stop("Parameter vector length mismatch")
  }
  if (ERC & is.null(Kmat)) {
    stop("Kmat necessary for ERC")
  }
  if (ERC & length(D) > 1) {
    stop(
      "ERC only works with one supplied dose (i.e., the mean across replicates)"
    )
  }

  if (!is.null(transform)) {
    if (is.function(transform)) {
      params <- transform(params = params, ...)
    } else {
      stop("transform should be a function")
    }
  }

  if (!is.null(X)) {
    a <- params[1:length(X)]
    Xlinpred <- c(as.matrix(data[, X]) %*% a)
  } else {
    Xlinpred <- 0
  }

  b1 <- params[length(X) + 1]

  if (deg == 2) {
    b2 <- params[length(X) + 2]
  } else if (deg == 1) {
    b2 <- 0
  }

  if (!is.null(M)) {
    bm1 <- params[(1 + deg + length(X)):(length(M) + deg + length(X))]
    if (deg == 2) {
      bm2 <- params[
        (length(M) + 1 + deg + length(X)):(2 * length(M) + deg + length(X))
      ]
    } else {
      bm2 <- NULL
    }
  } else {
    bm1 <- bm2 <- NULL
  }

  if (deg == 2) {
    betavec <- c(b1, b2, bm1, bm2)
  } else {
    betavec <- c(b1, bm1)
  }

  #RRs <- exp(pmin(Xlinpred, 7e1))*exposureRR(params=betavec, D=D, M=M, data=data, doseRRmod=doseRRmod, deg=deg)
  e1 <- exposureRR(
    params = betavec,
    D = D,
    M = M,
    data = data,
    doseRRmod = doseRRmod,
    deg = deg
  )
  e1 <- as.matrix(e1, ncol = length(D))
  if (any(e1 < 0)) {
    stop("RR <0, please check supplied transformation")
  }

  logRRs <- Xlinpred + log(e1)
  #logRRs <- logRRs - max(logRRs) # Is not compatible with ERC calculation!!

  RRs <- exp(pmin(logRRs, 7e1))

  ord_exit <- order(data[[exit]])
  exit_t <- data[[exit]][ord_exit]
  status_ord <- data[[status]][ord_exit]
  RR_exit <- RRs[ord_exit, , drop = FALSE]

  if (!is.null(entry)) {
    ord_entry <- order(data[[entry]])
    entry_t <- data[[entry]][ord_entry]
    RR_entry <- RRs[ord_entry, , drop = FALSE]
  } else {
    entry_t <- rep(min(exit_t) - 1, nrow(data))
    RR_entry <- RRs
  }

  ls <- loglik_prophaz_rcpp(
    exit_t,
    entry_t,
    RR_entry,
    RR_exit,
    status_ord,
    loglim
  )

  if (ERC) {
    derivs <- dRRdD(
      params = betavec,
      D = D,
      M = M,
      data = data,
      doseRRmod = doseRRmod,
      deg = deg
    )
    drdd <- exp(pmin(Xlinpred, 7e1)) * derivs$first
    drdd2 <- exp(pmin(Xlinpred, 7e1)) * derivs$second

    drdd <- drdd[ord_exit]
    drdd2 <- drdd2[ord_exit]

    Kmat <- Kmat[ord_exit, ord_exit]

    if (!is.null(entry)) {
      entry_t2 <- data[[entry]][ord_exit]
    } else {
      entry_t2 <- entry_t
    }
    tmp <- c(dldd_prophaz(entry_t2, exit_t, status_ord, RR_exit))
    dldd <- drdd / RR_exit * (status_ord == 1) - tmp * drdd

    mymat <- compute_ERCmatrix_prophaz(
      entry_t2,
      exit_t,
      status_ord,
      RR_exit,
      drdd,
      drdd2
    )

    return(
      -1 *
        (ls + log(max(1 + .5 * sum((tcrossprod(dldd) + mymat) * Kmat), loglim)))
    )
  } else {
    return(-1 * ls)
  }
}


loglik.multinomial <- function(
  params,
  Y,
  D,
  M = NULL,
  X = NULL,
  data,
  doseRRmod,
  ERC = FALSE,
  Kmat = NULL,
  deg = 1,
  loglim = 1e-30,
  transform = NULL,
  ...
) {
  Z <- length(unique(data[, Y])) # number of outcome categories
  # params = c(a0^(1), (a)^(1), b1^(1), b2^(1), (bm1)^(1), (bm2)^(1), ..., a0^(Z-1), (a)^(Z-1), b1^(Z-1), b2^(Z-1), (bm1)^(Z-1), (bm2)^(Z-1) )
  if (length(params) != (Z - 1) * (deg * length(M) + deg + length(X) + 1)) {
    stop("Parameter vector length mismatch")
  }
  if (ERC & length(D) > 1) {
    stop(
      "ERC only works with one supplied dose (i.e., the mean across replicates)"
    )
  }

  if (!is.null(transform)) {
    if (is.function(transform)) {
      params <- transform(params = params, ...)
    } else {
      stop("transform should be a function")
    }
  }

  params <- matrix(params, ncol = Z - 1)
  params <- cbind(params, rep(0, nrow(params)))

  a0 <- params[1, ]

  if (!is.null(X)) {
    a <- params[2:(length(X) + 1), ]
    Xlinpred <- sweep(as.matrix(data[, X]) %*% a, 2, a0, "+")
  } else {
    Xlinpred <- matrix(a0, nrow = nrow(data), ncol = Z, byrow = TRUE)
  }

  betamat <- params[
    (length(X) + 2):(deg * length(M) + deg + length(X) + 1),
    ,
    drop = FALSE
  ]

  if (length(D) > 1) {
    myarray <- array(NA_real_, dim = c(nrow(data), length(D), Z))

    for (ii in 1:(Z - 1)) {
      myarray[,, ii] <- exposureRR(
        params = betamat[, ii],
        D = D,
        M = M,
        data = data,
        doseRRmod = doseRRmod,
        deg = deg
      )
    }
    myarray[,, Z] <- 1

    if (any(myarray < 0)) {
      stop("RR <0, please check supplied transformation")
    }

    Xlinpred_array <- array(exp(pmin(Xlinpred, 7e1)), dim = c(nrow(data), 1, Z))
    Xlinpred_array <- Xlinpred_array[, rep(1, length(D)), , drop = FALSE]

    myarray <- myarray * Xlinpred_array # RR(d) * exp(X^t a)

    Asums <- colSums(aperm(myarray, c(3, 1, 2)))
    Asums_array <- array(Asums, dim = c(nrow(data), length(D), 1))
    Asums_array <- Asums_array[,, rep(1, Z), drop = FALSE]

    prob_array <- myarray / Asums_array # array with probabilities, indexed by individual x dose replicate x outcome category

    Ymat <- as.matrix(model.matrix(~ data[, Y] - 1))
    Y_array <- array(Ymat, dim = c(nrow(data), 1, Z))
    Y_array <- Y_array[, rep(1, length(D)), , drop = FALSE]

    myarray2 <- log(prob_array) * Y_array
    ls <- colSums(aperm(myarray2, c(1, 3, 2)), dims = 2)
  } else {
    RRmat <- matrix(1, nrow = nrow(data), ncol = Z)

    for (ii in 1:(Z - 1)) {
      RRmat[, ii] <- exposureRR(
        params = betamat[, ii],
        D = D,
        M = M,
        data = data,
        doseRRmod = doseRRmod,
        deg = deg
      )
    }

    if (any(RRmat < 0)) {
      stop("RR <0, please check supplied transformation")
    }

    RRmat <- RRmat * exp(pmin(Xlinpred, 7e1))
    probmat <- RRmat / rowSums(RRmat)[, drop = FALSE]

    Ymat <- diag(Z)[as.integer(data[, Y]), ] #as.matrix(model.matrix(~data[,Y]-1))
    ls <- sum(log(probmat) * Ymat)
  }

  if (ERC) {
    # Only one dose supplied for ERC

    A <- RRmat

    drdd <- apply(betamat, 2, function(betavec) {
      dRRdD(
        params = betavec,
        D = D,
        M = M,
        data = data,
        doseRRmod = doseRRmod,
        deg = deg
      )$first
    })
    drdd2 <- apply(betamat, 2, function(betavec) {
      dRRdD(
        params = betavec,
        D = D,
        M = M,
        data = data,
        doseRRmod = doseRRmod,
        deg = deg
      )$second
    })

    exp_X <- exp(pmin(Xlinpred, 7e1))
    rowSums_A <- rowSums(A)

    probvec <- rowSums(Ymat * probmat) #diag(tcrossprod(Ymat, probmat))

    drdd_expX <- drdd * exp_X
    Ymat_drdd_expX <- rowSums(Ymat * drdd_expX) #diag(tcrossprod(Ymat, drdd_expX))
    cross_drdd_exp_X <- rowSums(drdd * exp_X) #diag(tcrossprod(drdd, exp_X))

    dpdd <- (Ymat_drdd_expX - probvec * cross_drdd_exp_X) / rowSums_A
    dpdd2 <- (rowSums(Ymat * exp_X * drdd2) -
      dpdd * cross_drdd_exp_X -
      probvec * rowSums(drdd2 * exp_X)) /
      rowSums_A +
      (cross_drdd_exp_X^2 * probvec - cross_drdd_exp_X * Ymat_drdd_expX) /
        rowSums_A^2

    mymat <- tcrossprod(dpdd / probvec)
    diag(mymat) <- dpdd2 / probvec

    return(-1 * (ls + log(max(1 + .5 * sum(mymat * Kmat), loglim))))
  } else {
    return(-1 * ls)
  }
}


exposureRR <- function(params, D, M, data, doseRRmod, deg) {
  # params = c(beta1, beta2, (beta_m1), (beta_m2))
  params[is.infinite(params) & params > 0] <- 7e1
  dosemat <- as.matrix(data[, D, drop = FALSE])
  b1 <- params[1]

  if (deg == 2) {
    b2 <- params[2]
  } else {
    b2 <- 0
  }

  if (!is.null(M)) {
    bm1 <- params[(1 + deg):(length(M) + deg)]
    Mlinpred1 <- c(as.matrix(data[, M]) %*% bm1)
    if (deg == 2) {
      bm2 <- params[(length(M) + 1 + deg):(2 * length(M) + deg)]
      Mlinpred2 <- c(as.matrix(data[, M]) %*% bm2)
    } else {
      Mlinpred2 <- 0
    }
  } else {
    Mlinpred1 <- Mlinpred2 <- 0
  }

  if (doseRRmod == "ERR") {
    return(
      1 +
        b1 * dosemat +
        b2 * dosemat^2 +
        Mlinpred1 * dosemat +
        Mlinpred2 * dosemat^2
    )
  } else if (doseRRmod == "EXP") {
    val <- b1 *
      dosemat +
      b2 * dosemat^2 +
      Mlinpred1 * dosemat +
      Mlinpred2 * dosemat^2
    val <- pmin(val, 7e1) # cap large finite values
    val[!is.finite(val)] <- 7e1 # replace NaN, Inf, -Inf
    return(exp(val))
  } else if (doseRRmod == "LINEXP") {
    val <- 1 +
      (b1 + Mlinpred1) * dosemat * exp(pmin((b2 + Mlinpred2) * dosemat, 7e1))
    val <- pmin(val, exp(7e1))
    val[!is.finite(val)] <- exp(7e1) # replace Inf/NaN with capped value
    return(val)
    #return(pmin(1+(b1+Mlinpred1)*dosemat*exp(pmin((b2+Mlinpred2)*dosemat, 8e1)), exp(8e1)))
  }
}


dRRdD <- function(params, D, M, data, doseRRmod, deg) {
  # params = c(beta1, beta2, (beta_m1), (beta_m2))
  params[is.infinite(params) & params > 0] <- 7e1
  b1 <- params[1]

  if (deg == 2) {
    b2 <- params[2]
  } else {
    b2 <- 0
  }

  if (!is.null(M)) {
    bm1 <- params[(1 + deg):(length(M) + deg)]
    Mlinpred1 <- c(as.matrix(data[, M]) %*% bm1)
    if (deg == 2) {
      bm2 <- params[(length(M) + 1 + deg):(2 * length(M) + deg)]
      Mlinpred2 <- c(as.matrix(data[, M]) %*% bm2)
    } else {
      Mlinpred2 <- 0
    }
  } else {
    Mlinpred1 <- Mlinpred2 <- 0
  }

  # First derivative
  if (doseRRmod == "ERR") {
    first <- b1 + 2 * b2 * data[, D] + Mlinpred1 + 2 * Mlinpred2 * data[, D]
  } else if (doseRRmod == "EXP") {
    first <- (b1 + 2 * b2 * data[, D] + Mlinpred1 + 2 * Mlinpred2 * data[, D]) *
      exp(pmin(
        b1 *
          data[, D] +
          b2 * data[, D]^2 +
          Mlinpred1 * data[, D] +
          Mlinpred2 * data[, D]^2,
        8e1
      ))
  } else if (doseRRmod == "LINEXP") {
    first <- (b1 + Mlinpred1) *
      exp((b2 + Mlinpred2) * data[, D]) +
      (b2 + Mlinpred2) *
        (b1 + Mlinpred1) *
        data[, D] *
        exp((b2 + Mlinpred2) * data[, D])
  }

  # Second derivative
  if (doseRRmod == "ERR") {
    second <- 2 * (b2 + Mlinpred2)

    if (length(second) == 1) second <- rep(second, length(first))
  } else if (doseRRmod == "EXP") {
    second <- 2 *
      (b2 + Mlinpred2) *
      exp(pmin(
        b1 *
          data[, D] +
          b2 * data[, D]^2 +
          Mlinpred1 * data[, D] +
          Mlinpred2 * data[, D]^2,
        8e1
      )) +
      (b1 + 2 * b2 * data[, D] + Mlinpred1 + 2 * Mlinpred2 * data[, D])^2 *
        exp(pmin(
          b1 *
            data[, D] +
            b2 * data[, D]^2 +
            Mlinpred1 * data[, D] +
            Mlinpred2 * data[, D]^2,
          8e1
        ))

    if (length(second) == 1) second <- rep(second, length(first))
  } else if (doseRRmod == "LINEXP") {
    second <- (b1 + Mlinpred1) *
      (b2 + Mlinpred2) *
      exp((b2 + Mlinpred2) * data[, D]) +
      (b2 + Mlinpred2) * first

    if (length(second) == 1) second <- rep(second, length(first))
  }

  return(list(first = first, second = second))
}

ecdfplot <- function(
  data,
  dosevars,
  xlab = "Dose",
  ylab = "Cumulative distribution",
  show.mean = TRUE,
  log.xaxis = TRUE
) {
  check_pkgs(c("dplyr", "ggplot2", "tidyr", "scales", "patchwork"))

  dose <- curve_id <- row_id <- row_mean <- NULL # To suppress R CMD CHECK notes about undefined global variables

  dosemat <- data[, dosevars, drop = FALSE]

  p1 <- dosemat |>
    dplyr::mutate(row_id = dplyr::row_number()) |>
    tidyr::pivot_longer(
      cols = -row_id,
      names_to = "curve_id",
      values_to = "dose"
    ) |>
    ggplot2::ggplot() +
    ggplot2::stat_ecdf(
      ggplot2::aes(x = dose, group = curve_id),
      color = "grey",
      alpha = 0.2
    ) +
    ggplot2::theme_minimal(base_size = 15) +
    ggplot2::xlab(xlab) +
    ggplot2::ylab(ylab) +
    ggplot2::ggtitle("Curves corresponding to realizations")
  if (show.mean) {
    means1 <- dosemat |>
      dplyr::mutate(
        row_mean = rowMeans(dplyr::across(dplyr::everything()), na.rm = TRUE)
      )
    p1 <- p1 +
      ggplot2::stat_ecdf(
        data = means1,
        ggplot2::aes(x = row_mean),
        color = "black",
        linewidth = 1.2
      )
  }
  if (log.xaxis) {
    p1 <- p1 + ggplot2::scale_x_log10(labels = scales::comma)
  }

  p2 <- as.data.frame(t(dosemat)) |>
    dplyr::mutate(row_id = dplyr::row_number()) |>
    tidyr::pivot_longer(
      cols = -row_id,
      names_to = "curve_id",
      values_to = "dose"
    ) |>
    ggplot2::ggplot() +
    ggplot2::stat_ecdf(
      ggplot2::aes(x = dose, group = curve_id),
      color = "grey",
      alpha = 0.2
    ) +
    ggplot2::theme_minimal(base_size = 15) +
    ggplot2::xlab(xlab) +
    ggplot2::ylab(ylab) +
    ggplot2::ggtitle("Curves corresponding to individuals")
  if (show.mean) {
    means2 <- as.data.frame(t(dosemat)) |>
      dplyr::mutate(
        row_mean = rowMeans(dplyr::across(dplyr::everything()), na.rm = TRUE)
      )
    p2 <- p2 +
      ggplot2::stat_ecdf(
        data = means2,
        ggplot2::aes(x = row_mean),
        color = "black",
        linewidth = 1.2
      )
  }

  if (log.xaxis) {
    p2 <- p2 + ggplot2::scale_x_log10(labels = scales::comma)
  }
  patchwork::wrap_plots(p1, p2)
}

parse_ameras_formula <- function(formula, data, family) {
  specials <- c("dose", "strata", "offset")
  X_formula <- collect_X(formula)

  if (family == "prophaz") {
    surv <- parse_surv_term(formula)
    formula[[2]] <- quote(.response)
    tt <- terms(formula, specials = specials)
    dose <- parse_dose_term(tt, data)

    return(list(
      Y = NULL,
      status = surv$status,
      entry = surv$entry,
      exit = surv$exit,
      dosevars = dose$dosevars,
      doseRRmod = dose$doseRRmod,
      deg = dose$deg,
      M = dose$M,
      X_formula = X_formula,
      offset = NULL,
      setnr = NULL
    ))
  } else if (family == "clogit") {
    tt <- terms(formula, specials = specials)
    dose <- parse_dose_term(tt, data)
    strata <- parse_strata_term(tt)

    if (is.null(strata$setnr)) {
      stop("Formula for family='clogit' must contain a strata() term")
    }

    return(list(
      Y = as.character(formula[[2]]),
      status = NULL,
      entry = NULL,
      exit = NULL,
      dosevars = dose$dosevars,
      doseRRmod = dose$doseRRmod,
      deg = dose$deg,
      M = dose$M,
      X_formula = X_formula,
      offset = NULL,
      setnr = strata$setnr
    ))
  } else if (family == "poisson") {
    tt <- terms(formula, specials = specials)
    dose <- parse_dose_term(tt, data)
    off <- parse_offset_term(tt)

    return(list(
      Y = as.character(formula[[2]]),
      status = NULL,
      entry = NULL,
      exit = NULL,
      dosevars = dose$dosevars,
      doseRRmod = dose$doseRRmod,
      deg = dose$deg,
      M = dose$M,
      X_formula = X_formula,
      offset = off$offset,
      setnr = NULL
    ))
  } else {
    tt <- terms(formula, specials = specials)
    dose <- parse_dose_term(tt, data)

    return(list(
      Y = as.character(formula[[2]]),
      status = NULL,
      entry = NULL,
      exit = NULL,
      dosevars = dose$dosevars,
      doseRRmod = dose$doseRRmod,
      deg = dose$deg,
      M = dose$M,
      X_formula = X_formula,
      offset = NULL,
      setnr = NULL
    ))
  }
}


parse_dose_term <- function(tt, data) {
  special_idx <- attr(tt, "specials")$dose
  if (is.null(special_idx)) {
    stop("Formula must contain a dose() term")
  }

  rhs <- attr(tt, "variables")
  dose_call <- rhs[[special_idx + 1]]
  dose_args <- as.list(dose_call[-1])

  nms <- names(dose_args)
  if (is.null(nms)) {
    nms <- rep("", length(dose_args))
  }
  named_idx <- !is.na(nms) & nzchar(nms)
  named_args <- dose_args[named_idx]
  sel_args <- dose_args[!named_idx]

  dosevars <- resolve_dose_selection(sel_args, data)
  doseRRmod <- if (!is.null(named_args$model)) {
    as.character(named_args$model)
  } else {
    "ERR"
  }
  deg <- if (!is.null(named_args$deg)) as.integer(named_args$deg) else 1
  M <- if (!is.null(named_args$modifier)) {
    parse_modifier(named_args$modifier)
  } else {
    NULL
  }

  list(
    dosevars = dosevars,
    doseRRmod = doseRRmod,
    deg = deg,
    M = M
  )
}


parse_modifier <- function(expr) {
  if (is.null(expr)) {
    return(NULL)
  }
  if (is.call(expr)) {
    fn <- as.character(expr[[1]])
    if (fn %in% c("*", ":", "^")) {
      stop(
        "Interactions in modifier are not currently supported. ",
        "Please create interaction terms manually as new columns in your ",
        "data frame before calling ameras(). ",
        "For example: data$M1M2 <- data$M1 * data$M2"
      )
    }
  }
  all.vars(expr)
}


resolve_dose_selection <- function(sel_args, data) {
  sel_expr <- if (length(sel_args) == 1) {
    sel_args[[1]]
  } else {
    as.call(c(list(quote(c)), sel_args))
  }

  idx <- tidyselect::eval_select(sel_expr, data)
  colnames(data)[idx]
}


collect_X <- function(formula) {
  specials <- c("dose", "strata", "offset", "Surv")
  rhs <- formula[[3]]

  if (has_no_intercept(rhs)) {
    stop(
      "Removing the intercept via -1 or +0 is not currently supported. "
    )
  }

  remove_specials <- function(expr) {
    if (is.symbol(expr)) {
      return(expr)
    }
    if (is.call(expr)) {
      fn <- as.character(expr[[1]])
      if (fn %in% specials) {
        return(NULL)
      }
      if (fn %in% c("+", "-", "*", ":", "^")) {
        args <- lapply(as.list(expr)[-1], remove_specials)
        args <- Filter(Negate(is.null), args)
        if (!length(args)) {
          return(NULL)
        }
        if (length(args) == 1) {
          return(args[[1]])
        }
        return(as.call(c(list(expr[[1]]), args)))
      }
    }
    expr
  }

  cleaned_rhs <- remove_specials(rhs)
  if (is.null(cleaned_rhs)) {
    return(NULL)
  }

  # Return the cleaned RHS as a formula for later use by model.matrix
  as.formula(paste("~", deparse(cleaned_rhs, width.cutoff = 500L)))
}


has_no_intercept <- function(expr) {
  if (is.call(expr)) {
    fn <- as.character(expr[[1]])
    if (fn == "-") {
      # Check for -1
      args <- as.list(expr)[-1]
      if (any(sapply(args, function(a) is.numeric(a) && a == 1))) {
        return(TRUE)
      }
    }
    if (fn == "+") {
      # Check for +0
      args <- as.list(expr)[-1]
      if (any(sapply(args, function(a) is.numeric(a) && a == 0))) {
        return(TRUE)
      }
    }
    # Recurse into sub-expressions
    return(any(sapply(as.list(expr)[-1], has_no_intercept)))
  }
  FALSE
}


parse_surv_term <- function(formula) {
  lhs <- formula[[2]]

  if (!is.call(lhs) || as.character(lhs[[1]]) != "Surv") {
    stop("Left hand side must be a Surv() term for family='prophaz'")
  }

  args <- as.list(lhs[-1])

  if (length(args) == 2) {
    # Surv(exit, status)
    list(
      entry = NULL,
      exit = as.character(args[[1]]),
      status = as.character(args[[2]])
    )
  } else if (length(args) == 3) {
    # Surv(entry, exit, status)
    list(
      entry = as.character(args[[1]]),
      exit = as.character(args[[2]]),
      status = as.character(args[[3]])
    )
  } else {
    stop(
      "Surv() must have either 2 arguments (exit, status) or ",
      "3 arguments (entry, exit, status)"
    )
  }
}

parse_strata_term <- function(tt) {
  special_idx <- attr(tt, "specials")$strata
  if (is.null(special_idx)) {
    return(list(setnr = NULL))
  }

  rhs <- attr(tt, "variables")
  strata_call <- rhs[[special_idx + 1]]
  strata_args <- as.list(strata_call[-1])

  if (length(strata_args) != 1) {
    stop("strata() must contain exactly one variable")
  }

  list(setnr = as.character(strata_args[[1]]))
}


parse_offset_term <- function(tt) {
  special_idx <- attr(tt, "specials")$offset
  if (is.null(special_idx)) {
    return(list(offset = NULL))
  }

  rhs <- attr(tt, "variables")
  offset_call <- rhs[[special_idx + 1]]
  offset_args <- as.list(offset_call[-1])

  if (length(offset_args) != 1) {
    stop("offset() must contain exactly one variable")
  }

  list(offset = as.character(offset_args[[1]]))
}

check_df <- function(x, nm = "data") {
  if (!is.data.frame(x)) {
    stop(paste0("ERROR: ", nm, " must be a data frame"))
  }
  if (!nrow(x)) {
    stop(paste0("ERROR: ", nm, " has no rows"))
  }
  if (!ncol(x)) {
    stop(paste0("ERROR: ", nm, " has no columns"))
  }
  NULL
}

check_family <- function(x, nm = "family") {
  valid <- c(
    "gaussian",
    "binomial",
    "poisson",
    "prophaz",
    "multinomial",
    "clogit"
  )
  check_char_vec(x, nm, valid = valid, def = NULL, len = 1)
}

check_doseRRmod <- function(x, nm = "doseRRmod") {
  valid <- c("ERR", "EXP", "LINEXP")
  check_char_vec(x, nm, valid = valid, def = NULL, len = 1)
}

check_Y <- function(v, data, family) {
  nm <- "Y"
  check_vars(data, v, nm, minlen = 1, maxlen = 0)
  vec <- data[, v, drop = TRUE]
  if (family != "multinomial") {
    binary <- nonneg <- integer <- 0
    if (family %in% c("binomial", "prophaz", "clogit")) {
      binary <- 1
    } else if (family == "poisson") {
      nonneg <- 1
      integer <- 1
    }
    check_num_vec(vec, nm, binary = binary, nonneg = nonneg, integer = integer)
    NULL
  } else {
    # multinomial
    check_factor_vec(vec, nm)
    NULL
  }
}

check_D <- function(vars, data, methods) {
  nm <- "dosevars"
  check_vars(data, vars, nm, minlen = 1, maxlen = 0)
  for (v in vars) {
    vec <- data[, v, drop = TRUE]
    nm2 <- paste0(nm, ":", v)
    check_num_vec(vec, nm2)
  }
  if (length(vars) == 1 & any(c("ERC", "MCML", "FMA", "BMA") %in% methods)) {
    stop(
      "Multiple exposure replicates required for ERC, MCML, FMA, and BMA. With a single exposure vector, use RC"
    )
  }
  NULL
}

check_M <- function(vars, data) {
  nm <- "M"
  check_vars(data, vars, nm, minlen = 0, maxlen = 0)
  for (v in vars) {
    vec <- data[, v, drop = TRUE]
    nm2 <- paste0(nm, ":", v)
    check_num_vec(vec, nm2, binary = 1)
  }
  NULL
}

check_X <- function(vars, data) {
  nm <- "X"
  if (is.null(vars)) {
    return(NULL)
  }
  for (v in vars) {
    vec <- data[, v, drop = TRUE]
    nm2 <- paste0(nm, ":", v)
    check_num_vec(vec, nm2)
  }
  NULL
}

check_offset <- function(v, data) {
  if (!length(v)) {
    return(NULL)
  }
  nm <- "offset"
  check_vars(data, v, nm, minlen = 0, maxlen = 0)
  check_num_vec(data[, v, drop = TRUE], nm, nonneg = 1)

  NULL
}

check_setnr <- function(v, data) {
  nm <- "setnr"
  check_vars(data, v, nm, minlen = 1, maxlen = 1)
  check_num_vec(data[, v, drop = TRUE], nm, nonneg = 1, integer = 1)

  nset_noncontributing <- sum(table(data[, v, drop = TRUE]) == 1)
  if (nset_noncontributing > 0) {
    warning(paste0(
      "Data contains ",
      nset_noncontributing,
      " matched sets of size 1 that do not contribute to model estimation"
    ))
  }

  NULL
}

check_entry_exit <- function(entry, exit, data) {
  nm1 <- "entry"
  nm2 <- "exit"
  check_vars(data, entry, nm1, minlen = 0, maxlen = 0)
  check_vars(data, exit, nm2, minlen = 1, maxlen = 1)
  vec2 <- data[, exit, drop = TRUE]
  check_num_vec(vec2, nm2)
  if (length(entry)) {
    vec1 <- data[, entry, drop = TRUE]
    check_num_vec(vec1, nm1)
    tmp <- entry > exit
    tmp[is.na(tmp)] <- FALSE
    if (any(tmp)) stop(paste0("ERROR: ", nm1, " > ", nm2, " for some values"))
  }

  NULL
}

check_methods <- function(x) {
  nm <- "methods"
  valid <- c("RC", "ERC", "MCML", "FMA", "BMA")
  ret <- check_char_vec(x, nm, valid = valid, def = "RC", len = 0)
  ret <- unique(ret)
  ret
}

check_deg <- function(x) {
  if (!length(x)) {
    return(2)
  }
  nm <- "deg"
  check_integer(x, nm, minlen = 1, maxlen = 1, min = 1, max = 2)
  x
}

check_inpar <- function(x, family, M, X, deg, multinom_levels = 0) {
  if (!length(x)) {
    return(NULL)
  }
  nm <- "inpar"
  if (family == "gaussian") {
    len <- 2 + length(X) + length(M) * deg + deg
  } else if (family %in% c("binomial", "poisson")) {
    len <- 1 + length(X) + length(M) * deg + deg
  } else if (family %in% c("prophaz", "clogit")) {
    len <- length(X) + length(M) * deg + deg
  } else if (family == "multinomial") {
    len <- (multinom_levels - 1) * (1 + length(X) + length(M) * deg + deg)
  } else {
    stop("ERROR")
  }
  check_num_vec(x, nm, binary = 0, nonneg = 0, integer = 0, len = len)
  x
}


check_factor_vec <- function(x, nm, len = 0) {
  if (!is.factor(x)) {
    stop(paste0("ERROR: ", nm, " must be numeric"))
  }
  if (len && (len != length(x))) {
    stop(paste0("ERROR: ", nm, " must be a numeric vector of length ", len))
  }

  if (length(levels(x)) < 3) {
    stop(paste0("ERROR: ", nm, " must have at least 3 levels"))
  }

  if (length(levels(x)) > length(unique(x))) {
    stop(paste0("ERROR: ", nm, " contains unused levels"))
  }

  NULL
}

check_num_vec <- function(x, nm, binary = 0, nonneg = 0, integer = 0, len = 0) {
  if (!is.numeric(x)) {
    stop(paste0("ERROR: ", nm, " must be numeric"))
  }
  if (len && (len != length(x))) {
    stop(paste0("ERROR: ", nm, " must be a numeric vector of length ", len))
  }
  tmp <- !is.finite(x)
  if (any(tmp)) {
    stop(paste0("ERROR: ", nm, " must contain finite values"))
  }
  if (binary) {
    tmp <- !(x %in% 0:1)
    if (any(tmp)) {
      stop(paste0("ERROR: ", nm, " must contain binary (0 - 1) values"))
    }
  }
  if (nonneg) {
    tmp <- x < 0
    if (any(tmp)) {
      stop(paste0("ERROR: ", nm, " must contain non-negative values"))
    }
  }
  if (integer) {
    check_integer(x, nm, minlen = 0, maxlen = 0, min = NULL, max = NULL)
  }

  NULL
}

check_vars <- function(data, vars, nm, minlen = 0, maxlen = 0) {
  nv <- length(vars)
  if (minlen && (minlen == maxlen) && (nv != minlen)) {
    stop(paste0("ERROR: ", nm, " must have length ", minlen))
  }
  if (nv < minlen) {
    stop(paste0("ERROR: ", nm, " must have length >= ", minlen))
  }
  if (!nv) {
    return(NULL)
  }
  if (!is.vector(vars)) {
    stop(paste0(
      "ERROR: ",
      nm,
      " must be a vector of indices or variable names"
    ))
  }

  nc <- ncol(data)
  cx <- colnames(data)
  if (is.numeric(vars)) {
    check_integer(vars, nm, minlen = minlen, maxlen = nc, min = 1, max = nc)
  } else if (is.character(vars)) {
    check_char_vec(vars, nm, valid = cx, def = NULL)
  } else {
    stop(paste0(
      "ERROR: ",
      nm,
      " must be a vector of indices or variable names"
    ))
  }
  if (any(duplicated(vars))) {
    stop(paste0("ERROR: ", nm, " contains duplicated values"))
  }

  unique(vars)
}


check_char_vec <- function(x, nm, valid = NULL, def = NULL, len = 0) {
  n <- length(x)
  if (len && (n != len)) {
    stop(paste0("ERROR: ", nm, " must have length ", len))
  }
  if (!n) {
    return(def)
  }
  if (!is.character(x)) {
    stop(paste0("ERROR: ", nm, " must be character"))
  }
  if (length(valid)) {
    tmp <- !(x %in% valid)
    if (any(tmp)) {
      err <- getCharVecStr(x[tmp])
      stop(paste0("ERROR ", nm, " contains invalid values: ", err))
    }
  }
  x
}


required_vars <- function(m) {
  vars <- c(m$dosevars, m$X, m$M)

  if (m$family %in% c("gaussian", "binomial", "poisson", "multinomial")) {
    vars <- c(vars, m$Y)
  }

  if (m$family == "poisson" && !is.null(m$offset)) {
    vars <- c(vars, m$offset)
  }

  if (m$family == "prophaz") {
    vars <- c(vars, m$status, m$exit)
    if (!is.null(m$entry)) vars <- c(vars, m$entry)
  }

  if (m$family == "clogit") {
    vars <- c(vars, m$status, m$setnr)
  }

  vars[!is.null(vars)]
}

check_integer <- function(
  x,
  nm,
  minlen = 1,
  maxlen = 0,
  min = NULL,
  max = NULL
) {
  n <- length(x)
  if (minlen && (minlen == maxlen) && (n != minlen)) {
    stop(paste0("ERROR: ", nm, " must have length ", minlen))
  }
  if (minlen && (n < minlen)) {
    stop(paste0("ERROR: ", nm, " must have length >= ", minlen))
  }
  if (!is.numeric(x)) {
    stop(paste0("ERROR: ", nm, " must be integer"))
  }
  if (any(!is.finite(x))) {
    stop(paste0("ERROR: ", nm, " must be integer"))
  }
  if (any(x != floor(x))) {
    stop(paste0("ERROR: ", nm, " must be integer"))
  }
  if (length(min) && any(x < min)) {
    stop(paste0("ERROR: ", nm, " must be >= ", min))
  }
  if (length(max) && any(x > max)) {
    stop(paste0("ERROR: ", nm, " must be <= ", max))
  }

  NULL
}


check_string <- function(obj, valid, parm) {
  # obj:   A character string (length 1)
  # valid: Character vector of valid values
  # parm:  The name of the argument being checked

  errFlag <- 0

  # Check for errors
  if (!isString(obj)) {
    errFlag <- 1
  }
  if (!errFlag) {
    obj <- trimws(obj)
    if (!(obj %in% valid)) errFlag <- 1
  }
  if (errFlag) {
    msg <- getCharVecStr(valid)
    msg <- paste0("ERROR: ", parm, " contains the invalid values ", msg)
    stop(msg)
  }

  obj
} # END: check.string

getVarNumbers <- function(vars, data) {
  if (!length(vars)) {
    return(NULL)
  }
  if (is.numeric(vars)) {
    return(vars)
  }
  cx <- colnames(data)
  ret <- match(vars, cx)
  ret
}


getCharVecStr <- function(x, sep = ",") {
  ret <- paste0("'", x, "'")
  ret <- paste0(ret, collapse = sep)
  ret
}


# Function to check that an object is a string
isString <- function(obj) {
  if ((length(obj) == 1) && is.character(obj)) {
    ret <- TRUE
  } else {
    ret <- FALSE
  }

  ret
} # END: isString

make_base_loglik_fn <- function(object, method_fit, data) {
  m <- object$model
  ERC <- method_fit$ERC
  if (is.null(ERC)) {
    ERC <- FALSE
  }
  Kmat <- if (ERC && m$family != "poisson") {
    cov(t(data[, m$dosevars, drop = FALSE]))
  } else {
    NULL
  }

  if (m$family == "gaussian") {
    function(params, D) {
      do.call(
        loglik.gaussian,
        c(
          list(
            params = params,
            D = D,
            X = m$X,
            Y = m$Y,
            M = m$M,
            data = data,
            deg = m$deg,
            ERC = ERC,
            Kmat = Kmat,
            loglim = m$loglim,
            transform = object$transform
          ),
          object$other.args
        )
      )
    }
  } else if (m$family == "poisson") {
    if (ERC) {
      function(params, D) {
        do.call(
          loglik.poisson.erc,
          c(
            list(
              params = params,
              D = m$dosevars,
              X = m$X,
              Y = m$Y,
              M = m$M,
              offset = m$offset,
              doseRRmod = m$doseRRmod,
              data = data,
              deg = m$deg,
              loglim = m$loglim,
              transform = object$transform
            ),
            object$other.args
          )
        )
      }
    } else {
      function(params, D) {
        do.call(
          loglik.poisson,
          c(
            list(
              params = params,
              D = D,
              X = m$X,
              Y = m$Y,
              M = m$M,
              offset = m$offset,
              doseRRmod = m$doseRRmod,
              data = data,
              deg = m$deg,
              loglim = m$loglim,
              transform = object$transform
            ),
            object$other.args
          )
        )
      }
    }
  } else if (m$family == "binomial") {
    function(params, D) {
      do.call(
        loglik.binomial,
        c(
          list(
            params = params,
            D = D,
            X = m$X,
            Y = m$Y,
            M = m$M,
            doseRRmod = m$doseRRmod,
            data = data,
            deg = m$deg,
            ERC = ERC,
            Kmat = Kmat,
            loglim = m$loglim,
            transform = object$transform
          ),
          object$other.args
        )
      )
    }
  } else if (m$family == "prophaz") {
    function(params, D) {
      do.call(
        loglik.prophaz,
        c(
          list(
            params = params,
            D = D,
            X = m$X,
            status = m$status,
            entry = m$entry,
            exit = m$exit,
            M = m$M,
            doseRRmod = m$doseRRmod,
            data = data,
            deg = m$deg,
            ERC = ERC,
            Kmat = Kmat,
            loglim = m$loglim,
            transform = object$transform
          ),
          object$other.args
        )
      )
    }
  } else if (m$family == "clogit") {
    designmat <- t(model.matrix(~ as.factor(m$data[, m$setnr]) - 1))
    function(params, D) {
      do.call(
        loglik.clogit,
        c(
          list(
            params = params,
            D = D,
            X = m$X,
            status = m$status,
            designmat = designmat,
            M = m$M,
            doseRRmod = m$doseRRmod,
            data = data,
            deg = m$deg,
            ERC = ERC,
            Kmat = Kmat,
            loglim = m$loglim,
            transform = object$transform
          ),
          object$other.args
        )
      )
    }
  } else if (m$family == "multinomial") {
    function(params, D) {
      do.call(
        loglik.multinomial,
        c(
          list(
            params = params,
            D = D,
            X = m$X,
            Y = m$Y,
            M = m$M,
            doseRRmod = m$doseRRmod,
            data = data,
            deg = m$deg,
            ERC = ERC,
            Kmat = Kmat,
            loglim = m$loglim,
            transform = object$transform
          ),
          object$other.args
        )
      )
    }
  } else {
    stop("Unknown family: ", m$family)
  }
}


make_loglik_fn <- function(object, method_name, method_fit, data) {
  m <- object$model
  base_fn <- make_base_loglik_fn(object, method_fit, data)

  if (method_name == "MCML") {
    function(params) {
      logliks <- base_fn(params, D = m$dosevars)
      shifted <- pmax(pmin(-logliks - max(-logliks), 7e1), -7e1)
      -log(mean(exp(shifted))) - max(-logliks)
    }
  } else if (method_name %in% c("RC", "ERC")) {
    # Poisson ERC uses dosevars directly inside loglik.poisson.erc
    # so D argument is ignored in that case, but we pass it for consistency
    function(params) {
      base_fn(params, D = "rcdose_ameras")
    }
  } else {
    stop("make_loglik_fn not applicable for method: ", method_name)
  }
}


resolve_data <- function(object, data = NULL) {
  if (!is.null(object$model$data)) {
    return(object$model$data)
  }

  if (is.null(data)) {
    stop(
      "Data not stored on object (keep.data=FALSE). ",
      "Please supply data argument to confint()."
    )
  }

  check_df(data)

  m <- object$model

  # Check dose columns separately for a clearer error message
  missing_dose <- setdiff(m$dosevars, colnames(data))
  if (length(missing_dose)) {
    stop(
      "Supplied data is missing dose columns present during fitting: ",
      paste(missing_dose, collapse = ", ")
    )
  }

  # Check remaining required variables excluding X and dose
  # X is checked separately via X_formula, dose already checked above
  needed <- setdiff(required_vars(m), c(m$dosevars, m$X))
  missing_vars <- setdiff(needed, colnames(data))
  if (length(missing_vars)) {
    stop(
      "Supplied data is missing columns present during fitting: ",
      paste(missing_vars, collapse = ", ")
    )
  }

  # Check raw X variables referenced in the X formula
  if (!is.null(m$X_formula)) {
    X_raw_vars <- all.vars(m$X_formula)
    missing_X <- setdiff(X_raw_vars, colnames(data))
    if (length(missing_X)) {
      stop(
        "Supplied data is missing X variables present during fitting: ",
        paste(missing_X, collapse = ", ")
      )
    }
  }

  # Check dose columns are numeric and finite
  for (v in m$dosevars) {
    check_num_vec(data[, v, drop = TRUE], nm = paste0("dosevars:", v))
  }

  if (nrow(data) != object$num.rows) {
    warning(
      "Supplied data has ",
      nrow(data),
      " rows but original data had ",
      object$num.rows,
      " rows."
    )
  }

  # Re-expand X formula on supplied data
  if (!is.null(m$X_formula)) {
    X_matrix <- model.matrix(m$X_formula, data = data)[, -1, drop = FALSE]
    X_colnames <- colnames(X_matrix)
    new_cols <- setdiff(X_colnames, colnames(data))
    if (length(new_cols)) {
      data[, new_cols] <- X_matrix[, new_cols, drop = FALSE]
    }
  }

  data$rcdose_ameras <- rowMeans(data[, m$dosevars, drop = FALSE])

  data
}


check_pkgs <- function(pkgs) {
  missing <- pkgs[!vapply(pkgs, requireNamespace, logical(1), quietly = TRUE)]
  if (length(missing)) {
    stop("Please install required packages: ", paste(missing, collapse = ", "))
  }
}


`%||%` <- function(x, y) if (is.null(x)) y else x

ameras.fma <- function(
  family,
  dosevars,
  data,
  deg,
  transform = NULL,
  transform.jacobian = NULL,
  Y = NULL,
  M = NULL,
  X = NULL,
  offset = NULL,
  inpar = NULL,
  entry = NULL,
  exit = NULL,
  status = NULL,
  setnr = setnr,
  unweighted = NULL,
  doseRRmod = NULL,
  MFMA = 100000,
  optim.method = "Nelder-Mead",
  control = list(reltol = 1e-10),
  ...
) {
  if (is.null(unweighted)) {
    unweighted <- FALSE
  } else if (!is.logical(unweighted)) {
    stop("unweighted should be TRUE or FALSE")
  }

  t0 <- proc.time()
  if (family == "gaussian") {
    if (is.null(Y)) {
      stop("Y is required for family=gaussian")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 2 + length(X) + length(M) * deg + deg)
    }

    FMAfits <- lapply(1:length(dosevars), function(Xi) {
      fit.FMAi <- optim(
        inpar,
        loglik.gaussian,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        data = data,
        deg = deg,
        ERC = FALSE,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )
      if (optim.method == "Nelder-Mead") {
        fit.FMAi <- optim(
          fit.FMAi$par,
          loglik.gaussian,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
      }

      fit.FMAi$hessian <- numDeriv::hessian(
        func = loglik.gaussian,
        x = fit.FMAi$par,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        data = data,
        deg = deg,
        ERC = FALSE,
        transform = transform,
        ...
      )

      if (
        det(fit.FMAi$hessian) != 0 &
          rcond(fit.FMAi$hessian) > .Machine$double.eps &
          fit.FMAi$convergence == 0 &
          all(eigen(fit.FMAi$hessian)$values > 0)
      ) {
        include <- TRUE
      } else {
        include <- FALSE
      }

      list(
        coef = fit.FMAi$par,
        hess = fit.FMAi$hessian,
        AIC = 2 * fit.FMAi$value + 2 * (2 + length(X) + length(M) * deg + deg),
        convergence = fit.FMAi$convergence,
        include = include
      )
    })

    parnames <- c(
      "(Intercept)",
      names(data[, X, drop = FALSE]),
      c("dose", "dose_squared")[1:deg]
    )
    if (!is.null(M)) {
      parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
      if (deg == 2) {
        parnames <- c(
          parnames,
          paste0("dose_squared:", names(data[, M, drop = FALSE]))
        )
      }
    }
    parnames <- c(parnames, "sigma")
  } else if (family == "binomial") {
    if (is.null(Y)) {
      stop("Y is required for family=binomial")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
    }

    FMAfits <- lapply(1:length(dosevars), function(Xi) {
      fit.FMAi <- optim(
        inpar,
        loglik.binomial,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = FALSE,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )
      if (optim.method == "Nelder-Mead") {
        fit.FMAi <- optim(
          fit.FMAi$par,
          loglik.binomial,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
      }
      fit.FMAi$hessian <- numDeriv::hessian(
        func = loglik.binomial,
        x = fit.FMAi$par,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = FALSE,
        transform = transform,
        ...
      )

      if (
        det(fit.FMAi$hessian) != 0 &
          rcond(fit.FMAi$hessian) > .Machine$double.eps &
          fit.FMAi$convergence == 0 &
          all(eigen(fit.FMAi$hessian)$values > 0)
      ) {
        include <- TRUE
      } else {
        include <- FALSE
      }

      list(
        coef = fit.FMAi$par,
        hess = fit.FMAi$hessian,
        AIC = 2 * fit.FMAi$value + 2 * (1 + length(X) + length(M) * deg + deg),
        convergence = fit.FMAi$convergence,
        include = include
      )
    })

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "poisson") {
    if (is.null(Y)) {
      stop("Y is required for family=poisson")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
    }

    FMAfits <- lapply(1:length(dosevars), function(Xi) {
      fit.FMAi <- optim(
        inpar,
        loglik.poisson,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        offset = offset,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )
      if (optim.method == "Nelder-Mead") {
        fit.FMAi <- optim(
          fit.FMAi$par,
          loglik.poisson,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          offset = offset,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
      }
      fit.FMAi$hessian <- numDeriv::hessian(
        func = loglik.poisson,
        x = fit.FMAi$par,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        offset = offset,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        transform = transform,
        ...
      )

      if (
        det(fit.FMAi$hessian) != 0 &
          rcond(fit.FMAi$hessian) > .Machine$double.eps &
          fit.FMAi$convergence == 0 &
          all(eigen(fit.FMAi$hessian)$values > 0)
      ) {
        include <- TRUE
      } else {
        include <- FALSE
      }

      list(
        coef = fit.FMAi$par,
        hess = fit.FMAi$hessian,
        AIC = 2 * fit.FMAi$value + 2 * (1 + length(X) + length(M) * deg + deg),
        convergence = fit.FMAi$convergence,
        include = include
      )
    })

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "clogit") {
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=clogit")
    }
    if (is.null(status)) {
      stop("status is required for family=clogit")
    }

    designmat <- t(model.matrix(~ as.factor(data[, setnr]) - 1))

    if (is.null(inpar)) {
      inpar <- rep(0, length(X) + length(M) * deg + deg)
    }

    FMAfits <- lapply(1:length(dosevars), function(Xi) {
      if (length(X) + length(M) * deg + deg == 1) {
        # Optimize 1-dimensional model: use optimize instead of optim
        fit0 <- optimize(
          f = loglik.clogit,
          lower = -20,
          upper = 5,
          D = dosevars[Xi],
          status = status,
          X = X,
          M = M,
          designmat = designmat,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          ...
        )
        fit.FMAi <- list(
          par = fit0$minimum,
          value = fit0$objective,
          convergence = 0,
          hessian = numDeriv::hessian(
            func = loglik.clogit,
            x = fit0$minimum,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            designmat = designmat,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            ...
          )
        )
      } else {
        fit.FMAi <- optim(
          inpar,
          loglik.clogit,
          D = dosevars[Xi],
          status = status,
          X = X,
          M = M,
          designmat = designmat,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = optim.method,
          control = control,
          ...
        )
        if (optim.method == "Nelder-Mead") {
          fit.FMAi <- optim(
            fit.FMAi$par,
            loglik.clogit,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            designmat = designmat,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            method = "BFGS",
            control = control,
            ...
          )
        }
        fit.FMAi$hessian <- numDeriv::hessian(
          func = loglik.clogit,
          x = fit.FMAi$par,
          D = dosevars[Xi],
          status = status,
          X = X,
          M = M,
          designmat = designmat,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          ...
        )
      }

      if (
        det(fit.FMAi$hessian) != 0 &
          rcond(fit.FMAi$hessian) > .Machine$double.eps &
          fit.FMAi$convergence == 0 &
          all(eigen(fit.FMAi$hessian)$values > 0)
      ) {
        include <- TRUE
      } else {
        include <- FALSE
      }

      list(
        coef = fit.FMAi$par,
        hess = fit.FMAi$hessian,
        AIC = 2 * fit.FMAi$value + 2 * (length(X) + length(M) * deg + deg),
        convergence = fit.FMAi$convergence,
        include = include
      )
    })

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "prophaz") {
    if (is.null(exit)) {
      stop("exit is required for family=prophaz")
    }
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=prophaz")
    }
    if (is.null(status)) {
      stop("status is required for family=prophaz")
    }

    if (is.null(inpar)) {
      inpar <- rep(0, length(X) + length(M) * deg + deg)
    }

    FMAfits <- lapply(1:length(dosevars), function(Xi) {
      if (length(X) + length(M) * deg + deg == 1) {
        # Optimize 1-dimensional model: use optimize instead of optim
        fit0 <- optimize(
          f = loglik.prophaz,
          lower = -20,
          upper = 5,
          D = dosevars[Xi],
          status = status,
          X = X,
          M = M,
          entry = entry,
          exit = exit,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          ...
        )
        fit.FMAi <- list(
          par = fit0$minimum,
          value = fit0$objective,
          convergence = 0,
          hessian = numDeriv::hessian(
            func = loglik.prophaz,
            x = fit0$minimum,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            entry = entry,
            exit = exit,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            ...
          )
        )
      } else {
        fit.FMAi <- optim(
          inpar,
          loglik.prophaz,
          D = dosevars[Xi],
          status = status,
          X = X,
          M = M,
          entry = entry,
          exit = exit,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = optim.method,
          control = control,
          ...
        )
        if (optim.method == "Nelder-Mead") {
          fit.FMAi <- optim(
            fit.FMAi$par,
            loglik.prophaz,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            entry = entry,
            exit = exit,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            method = "BFGS",
            control = control,
            ...
          )
        }
        fit.FMAi$hessian <- numDeriv::hessian(
          func = loglik.prophaz,
          x = fit.FMAi$par,
          D = dosevars[Xi],
          status = status,
          X = X,
          M = M,
          entry = entry,
          exit = exit,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          ...
        )
      }

      if (
        det(fit.FMAi$hessian) != 0 &
          rcond(fit.FMAi$hessian) > .Machine$double.eps &
          fit.FMAi$convergence == 0 &
          all(eigen(fit.FMAi$hessian)$values > 0)
      ) {
        include <- TRUE
      } else {
        include <- FALSE
      }

      list(
        coef = fit.FMAi$par,
        hess = fit.FMAi$hessian,
        AIC = 2 * fit.FMAi$value + 2 * (length(X) + length(M) * deg + deg),
        convergence = fit.FMAi$convergence,
        include = include
      )
    })

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }
  } else if (family == "multinomial") {
    if (is.null(Y)) {
      stop("Y is required for family=multinomial")
    }

    if (is.null(inpar)) {
      inpar <- rep(
        0,
        (length(unique(data[, Y])) - 1) *
          (1 + length(X) + length(M) * deg + deg)
      )
    }

    FMAfits <- lapply(1:length(dosevars), function(Xi) {
      fit.FMAi <- optim(
        inpar,
        loglik.multinomial,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = FALSE,
        transform = transform,
        method = optim.method,
        control = control,
        ...
      )
      if (optim.method == "Nelder-Mead") {
        fit.FMAi <- optim(
          fit.FMAi$par,
          loglik.multinomial,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = "BFGS",
          control = control,
          ...
        )
      }
      fit.FMAi$hessian <- numDeriv::hessian(
        func = loglik.multinomial,
        x = fit.FMAi$par,
        D = dosevars[Xi],
        X = X,
        Y = Y,
        M = M,
        doseRRmod = doseRRmod,
        data = data,
        deg = deg,
        ERC = FALSE,
        transform = transform,
        ...
      )

      if (
        det(fit.FMAi$hessian) != 0 &
          rcond(fit.FMAi$hessian) > .Machine$double.eps &
          fit.FMAi$convergence == 0 &
          all(eigen(fit.FMAi$hessian)$values > 0)
      ) {
        include <- TRUE
      } else {
        include <- FALSE
      }

      list(
        coef = fit.FMAi$par,
        hess = fit.FMAi$hessian,
        AIC = 2 * fit.FMAi$value + 2 * (1 + length(X) + length(M) * deg + deg),
        convergence = fit.FMAi$convergence,
        include = include
      )
    })

    if (doseRRmod != "LINEXP") {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    } else {
      parnames <- c(
        "(Intercept)",
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    }

    mylv <- levels(data[, Y])

    mylv <- mylv[-length(mylv)]

    parnames <- do.call(
      "c",
      lapply(mylv, function(lv) paste0("(", lv, ")_", parnames))
    )
  }

  #Extra exclusion check for asymmetric variance matrix
  for (iii in 1:length(FMAfits)) {
    fit.FMAi <- FMAfits[[iii]]
    if (fit.FMAi$include) {
      if (!is.null(transform) & !is.null(transform.jacobian)) {
        if (is.function(transform) & is.function(transform.jacobian)) {
          jac <- transform.jacobian(fit.FMAi$coef, ...)
          samplevar <- jac %*% solve(fit.FMAi$hess) %*% t(jac)
        } else {
          stop("transform and transform.jacobian should be functions")
        }
      } else {
        samplevar <- solve(fit.FMAi$hess)
      }
      if (!isSymmetric(samplevar)) {
        FMAfits[[iii]]$include <- FALSE
      }
    }
  }

  #allfits <- FMAfits
  included.replicates <- which(sapply(FMAfits, function(x) x$include))

  if (length(included.replicates) > 0) {
    FMAfits <- FMAfits[sapply(FMAfits, function(x) x$include)]

    meanAIC <- mean(sapply(FMAfits, function(x) x$AIC))

    FMAfits <- lapply(FMAfits, function(x) {
      x$AIC_cent <- x$AIC - meanAIC
      x
    })

    FMAfits <- lapply(FMAfits, function(x) {
      if (unweighted) {
        x$wgt <- 1 / length(FMAfits)
      } else {
        x$wgt <- exp(-.5 * x$AIC_cent) /
          sum(sapply(FMAfits, function(y) exp(-.5 * y$AIC_cent)))
      }
      x$M <- max(round(x$wgt * MFMA), 1)
      x
    })

    FMAsamples <- lapply(
      FMAfits,
      function(fit.FMAi, ...) {
        if (!is.null(transform) & !is.null(transform.jacobian)) {
          if (is.function(transform) & is.function(transform.jacobian)) {
            samplemeans <- transform(fit.FMAi$coef, ...)
            jac <- transform.jacobian(fit.FMAi$coef, ...)
            samplevar <- jac %*% solve(fit.FMAi$hess) %*% t(jac)
          } else {
            stop("transform and transform.jacobian should be functions")
          }
        } else {
          samplemeans <- fit.FMAi$coef
          samplevar <- solve(fit.FMAi$hess)
        }

        return(rmvnorm(n = fit.FMAi$M, mean = samplemeans, sigma = samplevar))

        # if(isSymmetric(samplevar)){
        #   return(rmvnorm(n=fit.FMAi$M, mean=samplemeans, sigma=samplevar))
        # } else{
        #   return(NULL)
        # }
      },
      ...
    )

    FMAsamples <- do.call("rbind", FMAsamples)

    coefs <- colMeans(FMAsamples)
    sd <- apply(FMAsamples, 2, sd)

    FMAsamples <- as.data.frame(FMAsamples)
    names(coefs) <- names(sd) <- names(FMAsamples) <- parnames

    included.samples <- nrow(FMAsamples)
  } else {
    coefs <- sd <- NA * FMAfits[[1]]$coef
    names(coefs) <- names(sd) <- parnames

    included.samples <- 0
  }

  t1 <- proc.time()
  timedif <- t1 - t0
  runtime <- paste(
    round(as.numeric(as.difftime(timedif["elapsed"], units = "secs")), 1),
    "seconds"
  )

  prc_excluded <- round(
    100 * (1 - length(included.replicates) / length(dosevars)),
    1
  )
  if (length(included.replicates) / length(dosevars) < .8) {
    warning(paste0(
      "WARNING: ",
      prc_excluded,
      "% of replicates excluded from model averaging. Try different bounds or starting values."
    ))
  }

  out <- list(
    coefficients = coefs,
    sd = sd,
    included.replicates = included.replicates,
    included.samples = included.samples,
    samples = FMAsamples,
    #includedfits=FMAfits,
    #allfits=allfits,
    runtime = runtime
  )

  return(out)
}


ameras.bma <- function(
  family,
  dosevars,
  data,
  deg,
  Y = NULL,
  M = NULL,
  X = NULL,
  offset = NULL,
  entry = NULL,
  exit = NULL,
  status = NULL,
  setnr = setnr,
  transform = NULL,
  inpar = NULL,
  doseRRmod = NULL,
  ERRprior = "doubleexponential",
  prophaz_numints = 10,
  nburnin = 1000,
  niter = 5000,
  included.replicates = 1:length(dosevars),
  nchains = 2,
  thin = 10,
  optim.method = "Nelder-Mead",
  control = list(reltol = 1e-10),
  ...
) {
  # Remove build warnings, local functions may need to be pulled out from this function
  # HPDinterval <- K <- Mlen <- Mmat <- N <- Xlen <- Xmat <- a <- as.mcmc <-
  #   b <- bm <- bm1 <- bm2 <- buildMCMC <- nsets <- setmat <-
  #   col.ind <- compileNimble <- configureMCMC <- delta <-
  #   dosemat <- equals <- h0 <- inprod <- nimbleModel <- runMCMC <- NULL

  ndoses <- length(dosevars)

  if (family != "gaussian") {
    if (doseRRmod %in% c("ERR", "LINEXP") & is.null(ERRprior)) {
      stop("Please specify prior for ERR parameters")
    } else if (doseRRmod %in% c("ERR", "LINEXP") & !is.null(ERRprior)) {
      if (
        !(ERRprior %in%
          c(
            "truncated_normal",
            "truncated_horseshoe",
            "truncated_doubleexponential",
            "normal",
            "horseshoe",
            "doubleexponential"
          ))
      ) {
        stop(
          "Incorrect prior for ERR parameters specified, should be truncated_normal, truncated_horseshoe, truncated_doubleexponential, normal, horseshoe, or doubleexponential"
        )
      }
    }
  }

  t0 <- proc.time()

  if (family == "gaussian") {
    if (is.null(Y)) {
      stop("Y is required for family=gaussian")
    }

    if (is.null(included.replicates)) {
      if (is.null(inpar)) {
        inpar <- rep(0, 2 + length(X) + length(M) * deg + deg)
      }

      toInclude <- sapply(1:length(dosevars), function(Xi) {
        fit.FMAi <- optim(
          inpar,
          loglik.gaussian,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = optim.method,
          control = control,
          ...
        )
        if (optim.method == "Nelder-Mead") {
          fit.FMAi <- optim(
            fit.FMAi$par,
            loglik.gaussian,
            D = dosevars[Xi],
            X = X,
            Y = Y,
            M = M,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            method = "BFGS",
            control = control,
            ...
          )
        }
        fit.FMAi$hessian <- numDeriv::hessian(
          func = loglik.gaussian,
          x = fit.FMAi$par,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          ...
        )

        if (
          det(fit.FMAi$hessian) != 0 &
            rcond(fit.FMAi$hessian) > .Machine$double.eps &
            fit.FMAi$convergence == 0 &
            all(eigen(fit.FMAi$hessian)$values > 0)
        ) {
          include <- TRUE
        } else {
          include <- FALSE
        }

        return(include)
      })
      included.replicates <- which(toInclude == TRUE)
    }
    dosevars <- dosevars[included.replicates]

    nimbledata <- list(Y = data[, Y], dosemat = data[, dosevars])

    nimbleinits <- function() {
      L <- list(
        a0 = rnorm(1),
        b = rexp(deg),
        sigma = rexp(1),
        col.ind = sample(1:length(dosevars), 1)
      )

      if (length(X) > 0) {
        L <- c(L, list(a = rnorm(length(X))))
      }

      if (length(M) > 0) {
        if (deg == 1) {
          L <- c(L, list(bm = rexp(length(M))))
        } else if (deg > 1) {
          L <- c(L, list(bm1 = rexp(length(M)), bm2 = rexp(length(M))))
        }
      }
      L
    }
  } else if (family == "binomial") {
    if (is.null(Y)) {
      stop("Y is required for family=binomial")
    }
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=binomial")
    }

    if (is.null(included.replicates)) {
      if (is.null(inpar)) {
        inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
      }

      toInclude <- sapply(1:length(dosevars), function(Xi) {
        fit.FMAi <- optim(
          inpar,
          loglik.binomial,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = optim.method,
          control = control,
          ...
        )
        if (optim.method == "Nelder-Mead") {
          fit.FMAi <- optim(
            fit.FMAi$par,
            loglik.binomial,
            D = dosevars[Xi],
            X = X,
            Y = Y,
            M = M,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            method = "BFGS",
            control = control,
            ...
          )
        }
        fit.FMAi$hessian <- numDeriv::hessian(
          func = loglik.binomial,
          x = fit.FMAi$par,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          ...
        )

        if (
          det(fit.FMAi$hessian) != 0 &
            rcond(fit.FMAi$hessian) > .Machine$double.eps &
            fit.FMAi$convergence == 0 &
            all(eigen(fit.FMAi$hessian)$values > 0)
        ) {
          include <- TRUE
        } else {
          include <- FALSE
        }

        return(include)
      })
      included.replicates <- which(toInclude == TRUE)
    }

    dosevars <- dosevars[included.replicates]

    nimbledata <- list(Y = data[, Y], dosemat = data[, dosevars])

    nimbleinits <- function() {
      L <- list(
        a0 = rnorm(1),
        b = rexp(deg),
        col.ind = sample(1:length(dosevars), 1)
      )

      if (length(X) > 0) {
        L <- c(L, list(a = rnorm(length(X))))
      }

      if (length(M) > 0) {
        if (deg == 1) {
          L <- c(L, list(bm = rexp(length(M))))
        } else if (deg > 1) {
          L <- c(L, list(bm1 = rexp(length(M)), bm2 = rexp(length(M))))
        }
      }
      L
    }
  } else if (family == "poisson") {
    if (is.null(Y)) {
      stop("Y is required for family=poisson")
    }
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=poisson")
    }

    if (is.null(offset)) {
      P <- rep(1, nrow(data))
    } else {
      P <- data[, offset]
    }

    if (is.null(included.replicates)) {
      if (is.null(inpar)) {
        inpar <- rep(0, 1 + length(X) + length(M) * deg + deg)
      }

      toInclude <- sapply(1:length(dosevars), function(Xi) {
        fit.FMAi <- optim(
          inpar,
          loglik.poisson,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          offset = offset,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          transform = transform,
          method = optim.method,
          control = control,
          ...
        )
        if (optim.method == "Nelder-Mead") {
          fit.FMAi <- optim(
            fit.FMAi$par,
            loglik.poisson,
            D = dosevars[Xi],
            X = X,
            Y = Y,
            M = M,
            offset = offset,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            transform = transform,
            method = "BFGS",
            control = control,
            ...
          )
        }
        fit.FMAi$hessian <- numDeriv::hessian(
          func = loglik.poisson,
          x = fit.FMAi$par,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          offset = offset,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          transform = transform,
          ...
        )

        if (
          det(fit.FMAi$hessian) != 0 &
            rcond(fit.FMAi$hessian) > .Machine$double.eps &
            fit.FMAi$convergence == 0 &
            all(eigen(fit.FMAi$hessian)$values > 0)
        ) {
          include <- TRUE
        } else {
          include <- FALSE
        }

        return(include)
      })
      included.replicates <- which(toInclude == TRUE)
    }
    dosevars <- dosevars[included.replicates]

    nimbledata <- list(Y = data[, Y], dosemat = data[, dosevars])

    nimbleinits <- function() {
      L <- list(
        a0 = rnorm(1),
        b = rexp(deg),
        col.ind = sample(1:length(dosevars), 1)
      )

      if (length(X) > 0) {
        L <- c(L, list(a = rnorm(length(X))))
      }

      if (length(M) > 0) {
        if (deg == 1) {
          L <- c(L, list(bm = rexp(length(M))))
        } else if (deg > 1) {
          L <- c(L, list(bm1 = rexp(length(M)), bm2 = rexp(length(M))))
        }
      }
      L
    }
  } else if (family == "prophaz") {
    if (is.null(exit)) {
      stop("exit is required for family=prophaz")
    }
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=prophaz")
    }
    if (is.null(status)) {
      stop("status is required for family=prophaz")
    }

    if (is.null(included.replicates)) {
      if (is.null(inpar)) {
        inpar <- rep(0, length(X) + length(M) * deg + deg)
      }

      toInclude <- sapply(1:length(dosevars), function(Xi) {
        if (length(X) + length(M) * deg + deg == 1) {
          # Optimize 1-dimensional model: use optimize instead of optim
          fit0 <- optimize(
            f = loglik.prophaz,
            lower = -20,
            upper = 5,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            entry = entry,
            exit = exit,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            ...
          )
          fit.FMAi <- list(
            par = fit0$minimum,
            value = fit0$objective,
            convergence = 0,
            hessian = numDeriv::hessian(
              func = loglik.prophaz,
              x = fit0$minimum,
              D = dosevars[Xi],
              status = status,
              X = X,
              M = M,
              entry = entry,
              exit = exit,
              doseRRmod = doseRRmod,
              data = data,
              deg = deg,
              ERC = FALSE,
              transform = transform,
              ...
            )
          )
        } else {
          fit.FMAi <- optim(
            inpar,
            loglik.prophaz,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            entry = entry,
            exit = exit,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            method = optim.method,
            control = control,
            ...
          )
          if (optim.method == "Nelder-Mead") {
            fit.FMAi <- optim(
              fit.FMAi$par,
              loglik.prophaz,
              D = dosevars[Xi],
              status = status,
              X = X,
              M = M,
              entry = entry,
              exit = exit,
              doseRRmod = doseRRmod,
              data = data,
              deg = deg,
              ERC = FALSE,
              transform = transform,
              method = "BFGS",
              control = control,
              ...
            )
          }
          fit.FMAi$hessian <- numDeriv::hessian(
            func = loglik.prophaz,
            x = fit.FMAi$par,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            entry = entry,
            exit = exit,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            ...
          )
        }

        if (
          det(fit.FMAi$hessian) != 0 &
            rcond(fit.FMAi$hessian) > .Machine$double.eps &
            fit.FMAi$convergence == 0 &
            all(eigen(fit.FMAi$hessian)$values > 0)
        ) {
          include <- TRUE
        } else {
          include <- FALSE
        }

        return(include)
      })
      included.replicates <- which(toInclude == TRUE)
    }

    dosevars <- dosevars[included.replicates]

    prophaz_timepoints <- as.numeric(quantile(
      data[data[, status] == 1, exit],
      probs = seq(0, 1, length.out = prophaz_numints + 1)
    )) # define cut points using quantiles of event time distribution
    if (is.null(entry)) {
      prophaz_timepoints[1] <- 0
    } else {
      prophaz_timepoints[1] <- min(data[, entry])
    }

    prophaz_timepoints[prophaz_numints + 1] <- max(data[, exit]) + .001

    if (!is.null(entry)) {
      int.entry <- as.numeric(cut(
        data[, entry],
        breaks = prophaz_timepoints,
        right = FALSE
      )) # indicator for which interval entry time belongs to
    } else {
      int.entry <- rep(1, nrow(data))
    }

    int.exit <- as.numeric(cut(
      data[, exit],
      breaks = prophaz_timepoints,
      right = FALSE
    )) # indicator for which interval exit time belongs to

    nimbledata <- list(
      delta = data[, status],
      exit = data[, exit],
      dosemat = data[, dosevars],
      zeros = rep(0, nrow(data))
    )
    if (!is.null(entry)) {
      nimbledata <- c(nimbledata, list(entry = data[, entry]))
    } else {
      nimbledata <- c(nimbledata, list(entry = rep(0, nrow(data))))
    }

    nimbleinits <- function() {
      L <- list(
        b = rexp(deg),
        col.ind = sample(1:length(dosevars), 1),
        h0 = runif(prophaz_numints, .1)
      )

      if (length(X) > 0) {
        L <- c(L, list(a = rnorm(length(X))))
      }

      if (length(M) > 0) {
        if (deg == 1) {
          L <- c(L, list(bm = rexp(length(M))))
        } else if (deg > 1) {
          L <- c(L, list(bm1 = rexp(length(M)), bm2 = rexp(length(M))))
        }
      }
      L
    }
  } else if (family == "clogit") {
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=clogit")
    }

    # Remove sets of size 1
    set_counts <- table(data[, setnr])
    valid_sets <- as.numeric(names(set_counts[set_counts > 1]))
    data <- data[data[, setnr] %in% valid_sets, ]

    # Reorder and determine indexing for nimble
    data <- data[order(data[, setnr]), ]
    nsets <- length(unique(data[, setnr]))
    set_sizes <- as.numeric(table(data[, setnr]))
    set_start <- cumsum(c(1, head(set_sizes, -1)))
    set_end <- cumsum(set_sizes)

    if (is.null(included.replicates)) {
      designmat <- t(model.matrix(~ as.factor(data[, setnr]) - 1))

      if (is.null(inpar)) {
        inpar <- rep(0, length(X) + length(M) * deg + deg)
      }

      toInclude <- sapply(1:length(dosevars), function(Xi) {
        if (length(X) + length(M) * deg + deg == 1) {
          # Optimize 1-dimensional model: use optimize instead of optim
          fit0 <- optimize(
            f = loglik.clogit,
            lower = -20,
            upper = 5,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            designmat = designmat,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            ...
          )
          fit.FMAi <- list(
            par = fit0$minimum,
            value = fit0$objective,
            convergence = 0,
            hessian = numDeriv::hessian(
              func = loglik.clogit,
              x = fit0$minimum,
              D = dosevars[Xi],
              status = status,
              X = X,
              M = M,
              designmat = designmat,
              doseRRmod = doseRRmod,
              data = data,
              deg = deg,
              ERC = FALSE,
              transform = transform,
              ...
            )
          )
        } else {
          fit.FMAi <- optim(
            inpar,
            loglik.clogit,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            designmat = designmat,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            method = optim.method,
            control = control,
            ...
          )
          if (optim.method == "Nelder-Mead") {
            fit.FMAi <- optim(
              fit.FMAi$par,
              loglik.clogit,
              D = dosevars[Xi],
              status = status,
              X = X,
              M = M,
              designmat = designmat,
              doseRRmod = doseRRmod,
              data = data,
              deg = deg,
              ERC = FALSE,
              transform = transform,
              method = "BFGS",
              control = control,
              ...
            )
          }
          fit.FMAi$hessian <- numDeriv::hessian(
            func = loglik.clogit,
            x = fit.FMAi$par,
            D = dosevars[Xi],
            status = status,
            X = X,
            M = M,
            designmat = designmat,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            ...
          )
        }

        if (
          det(fit.FMAi$hessian) != 0 &
            rcond(fit.FMAi$hessian) > .Machine$double.eps &
            fit.FMAi$convergence == 0 &
            all(eigen(fit.FMAi$hessian)$values > 0)
        ) {
          include <- TRUE
        } else {
          include <- FALSE
        }

        return(include)
      })
      included.replicates <- which(toInclude == TRUE)
    }

    dosevars <- dosevars[included.replicates]

    nimbledata <- list(
      Y = as.numeric(data[, status]),
      dosemat = data[, dosevars]
    )

    nimbleinits <- function() {
      L <- list(b = rexp(deg), col.ind = sample(1:length(dosevars), 1))

      if (length(X) > 0) {
        L <- c(L, list(a = rnorm(length(X))))
      }

      if (length(M) > 0) {
        if (deg == 1) {
          L <- c(L, list(bm = rexp(length(M))))
        } else if (deg > 1) {
          L <- c(L, list(bm1 = rexp(length(M)), bm2 = rexp(length(M))))
        }
      }
      L
    }
  } else if (family == "multinomial") {
    if (is.null(Y)) {
      stop("Y is required for family=multinomial")
    }
    if (is.null(doseRRmod)) {
      stop("doseRRmod is required for family=multinomial")
    }

    Z <- nlevels(data[, Y])
    if (is.null(included.replicates)) {
      if (is.null(inpar)) {
        inpar <- rep(0, (Z - 1) * (1 + length(X) + length(M) * deg + deg))
      }

      toInclude <- sapply(1:length(dosevars), function(Xi) {
        fit.FMAi <- optim(
          inpar,
          loglik.multinomial,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          method = optim.method,
          control = control,
          ...
        )
        if (optim.method == "Nelder-Mead") {
          fit.FMAi <- optim(
            fit.FMAi$par,
            loglik.multinomial,
            D = dosevars[Xi],
            X = X,
            Y = Y,
            M = M,
            doseRRmod = doseRRmod,
            data = data,
            deg = deg,
            ERC = FALSE,
            transform = transform,
            method = "BFGS",
            control = control,
            ...
          )
        }
        fit.FMAi$hessian <- numDeriv::hessian(
          func = loglik.multinomial,
          x = fit.FMAi$par,
          D = dosevars[Xi],
          X = X,
          Y = Y,
          M = M,
          doseRRmod = doseRRmod,
          data = data,
          deg = deg,
          ERC = FALSE,
          transform = transform,
          ...
        )

        if (
          det(fit.FMAi$hessian) != 0 &
            rcond(fit.FMAi$hessian) > .Machine$double.eps &
            fit.FMAi$convergence == 0 &
            all(eigen(fit.FMAi$hessian)$values > 0)
        ) {
          include <- TRUE
        } else {
          include <- FALSE
        }

        return(include)
      })
      included.replicates <- which(toInclude == TRUE)
    }

    dosevars <- dosevars[included.replicates]

    nimbledata <- list(
      Y = model.matrix(~ data[, Y] - 1),
      dosemat = data[, dosevars]
    )

    nimbleinits <- function() {
      L <- list(
        a0 = rnorm(Z - 1),
        b = matrix(rexp(deg * (Z - 1)), ncol = Z - 1),
        col.ind = sample(1:length(dosevars), 1)
      )
      if (deg == 1) {
        L$b <- as.vector(L$b)
      }

      if (length(X) > 0) {
        L <- c(L, list(a = matrix(rnorm((Z - 1) * length(X)), ncol = Z - 1)))
      }

      if (length(M) > 0) {
        if (deg == 1) {
          L <- c(L, list(bm = matrix(rexp((Z - 1) * length(M)), ncol = Z - 1)))
        } else if (deg > 1) {
          L <- c(
            L,
            list(
              bm1 = matrix(rexp((Z - 1) * length(M)), ncol = Z - 1),
              bm2 = matrix(rexp((Z - 1) * length(M)), ncol = Z - 1)
            )
          )
        }
      }
      L
    }
  }

  if (!(family %in% c("prophaz", "clogit"))) {
    mons <- c("a0", "b", "col.ind")
  } else {
    mons <- c("b", "col.ind")
  }

  if (family == "gaussian") {
    doseRRmod <- "EXP"
  }
  nimbleconst <- list(
    N = nrow(data),
    K = length(dosevars),
    deg = deg,
    Xlen = length(X),
    Mlen = length(M),
    w = rep(1 / length(dosevars), length(dosevars)),
    doseRRmod = doseRRmod,
    family = family
  )
  if (family == "poisson") {
    nimbleconst <- c(nimbleconst, list(P = P))
  }
  if (family == "multinomial") {
    nimbleconst <- c(nimbleconst, list(Z = Z))
  }
  if (length(X) > 0) {
    nimbleconst <- c(nimbleconst, list(Xmat = data[, X]))
    mons <- c(mons, "a")
  }
  if (length(M) > 0) {
    nimbleconst <- c(nimbleconst, list(Mmat = data[, M]))
    if (deg == 1) {
      mons <- c(mons, "bm")
    } else if (deg > 1) {
      mons <- c(mons, "bm1", "bm2")
    }
  }
  if (family == "gaussian") {
    mons <- c(mons, "sigma")
  }
  if (family == "prophaz") {
    nimbleconst <- c(
      nimbleconst,
      list(
        prophaz_timepoints = prophaz_timepoints,
        int.exit = int.exit,
        int.entry = int.entry,
        prophaz_numints = prophaz_numints
      )
    )
    mons <- c(mons, "h0")
  }
  if (family == "clogit") {
    nimbleconst <- c(
      nimbleconst,
      list(nsets = nsets, set_start = set_start, set_end = set_end)
    )
  }

  if (doseRRmod %in% c("ERR", "LINEXP")) {
    nimbleconst <- c(nimbleconst, list(ERRprior = ERRprior))
  }

  mymod <- nimbleModel(
    nimblemod,
    data = nimbledata,
    constants = nimbleconst,
    inits = list(col.ind = sample(1:length(dosevars), 1))
  )

  mymod_C <- compileNimble(mymod)

  mymod <- configureMCMC(mymod, monitors = mons, thin = thin, print = FALSE)

  mymod$removeSamplers('col.ind')
  mymod$addSampler(
    target = 'col.ind',
    type = boundedSliceSampler,
    control = list(
      K = length(dosevars),
      adaptive = FALSE,
      sliceWidth = round(length(dosevars) / 2)
    )
  )

  mymod_MCMC <- buildMCMC(mymod)
  mymod_compiled <- compileNimble(mymod_MCMC, project = mymod_C)

  nimblesamples <- runMCMC(
    mymod_compiled,
    nburnin = nburnin,
    niter = niter,
    nchains = nchains,
    inits = nimbleinits
  )

  if (family != "multinomial") {
    if (!(family %in% c("prophaz", "clogit"))) {
      pars <- "a0"
    } else {
      pars <- NULL
    }

    if (length(X) == 1) {
      pars <- c(pars, "a")
    } else if (length(X) > 1) {
      pars <- c(pars, paste0("a[", 1:length(X), "]"))
    }
    if (deg == 1) {
      pars <- c(pars, "b")
    } else if (deg > 1) {
      pars <- c(pars, "b[1]", "b[2]")
    }
    if (length(M) == 1) {
      if (deg == 1) {
        pars <- c(pars, "bm")
      } else if (deg > 1) {
        pars <- c(pars, "bm1", "bm2")
      }
    } else if (length(M) > 1) {
      if (deg == 1) {
        pars <- c(pars, paste0("bm[", 1:length(M), "]"))
      } else if (deg > 1) {
        pars <- c(
          pars,
          paste0("bm1[", 1:length(M), "]"),
          paste0("bm2[", 1:length(M), "]")
        )
      }
    }
    if (family == "gaussian") {
      pars <- c(pars, "sigma")
    }
    if (family == "prophaz") {
      pars <- c(pars, paste0("h0[", 1:prophaz_numints, "]"))
    }
  } else {
    # multinomial
    pars <- NULL
    for (z in 1:(Z - 1)) {
      pars <- c(pars, paste0("a0[", z, "]"))

      if (length(X) == 1) {
        pars <- c(pars, paste0("a[", z, "]"))
      } else if (length(X) > 1) {
        pars <- c(pars, paste0("a[", 1:length(X), ", ", z, "]"))
      }

      if (deg == 1) {
        pars <- c(pars, paste0("b[", z, "]"))
      } else {
        pars <- c(pars, paste0("b[1, ", z, "]"), paste0("b[2, ", z, "]"))
      }

      if (length(M) == 1) {
        if (deg == 1) {
          pars <- c(pars, paste0("bm[", z, "]"))
        } else {
          pars <- c(pars, paste0("bm1[", z, "]"), paste0("bm2[", z, "]"))
        }
      } else if (length(M) > 1) {
        if (deg == 1) {
          pars <- c(pars, paste0("bm[", 1:length(M), ", ", z, "]"))
        } else {
          pars <- c(
            pars,
            paste0("bm1[", 1:length(M), ", ", z, "]"),
            paste0("bm2[", 1:length(M), ", ", z, "]")
          )
        }
      }
    }
  }

  if (!(family %in% c("prophaz", "clogit"))) {
    parnames <- "(Intercept)"
  } else {
    parnames <- NULL
  }
  if (!is.null(doseRRmod)) {
    if (doseRRmod == "LINEXP") {
      parnames <- c(
        parnames,
        names(data[, X, drop = FALSE]),
        c("dose_linear", "dose_exponential")
      )
      if (!is.null(M)) {
        parnames <- c(
          parnames,
          paste0("dose_linear:", names(data[, M, drop = FALSE]))
        )
        parnames <- c(
          parnames,
          paste0("dose_exponential:", names(data[, M, drop = FALSE]))
        )
      }
    } else {
      parnames <- c(
        parnames,
        names(data[, X, drop = FALSE]),
        c("dose", "dose_squared")[1:deg]
      )
      if (!is.null(M)) {
        parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
        if (deg == 2) {
          parnames <- c(
            parnames,
            paste0("dose_squared:", names(data[, M, drop = FALSE]))
          )
        }
      }
    }
  } else {
    parnames <- c(
      parnames,
      names(data[, X, drop = FALSE]),
      c("dose", "dose_squared")[1:deg]
    )
    if (!is.null(M)) {
      parnames <- c(parnames, paste0("dose:", names(data[, M, drop = FALSE])))
      if (deg == 2) {
        parnames <- c(
          parnames,
          paste0("dose_squared:", names(data[, M, drop = FALSE]))
        )
      }
    }
  }
  if (family == "gaussian") {
    parnames <- c(parnames, "sigma")
  }
  if (family == "prophaz") {
    parnames <- c(parnames, paste0("h0[", 1:prophaz_numints, "]"))
  }

  if (doseRRmod == "LINEXP") {
    parnames <- sub("\\bdose\\b", "dose_linear", parnames)
    parnames <- sub("\\bdose_squared\\b", "dose_exponential", parnames)
  }

  if (family == "multinomial") {
    mylv <- levels(data[, Y])

    mylv <- mylv[-length(mylv)]

    parnames <- do.call(
      "c",
      lapply(mylv, function(lv) paste0("(", lv, ")_", parnames))
    )
  }

  if (nchains > 1) {
    nimblesamples.stacked <- do.call("rbind", nimblesamples)
    for (ichain in 1:length(nimblesamples)) {
      nimblesamples[[ichain]] <- nimblesamples[[ichain]][, c(pars, "col.ind")]
      colnames(nimblesamples[[ichain]]) <- c(parnames, "col.ind")
    }
  } else if (nchains == 1) {
    nimblesamples.stacked <- nimblesamples
    nimblesamples <- nimblesamples[, c(pars, "col.ind")]
    colnames(nimblesamples) <- c(parnames, "col.ind")
  }
  nimblesamples.stacked <- nimblesamples.stacked[, pars]

  coef <- colMeans(nimblesamples.stacked)
  names(coef) <- parnames

  sd <- apply(nimblesamples.stacked, 2, sd)
  names(sd) <- parnames

  mcmcsum <- MCMCsummary(nimblesamples)
  Rhat <- mcmcsum[-nrow(mcmcsum), c("Rhat", "n.eff")]
  if (nchains > 1) {
    if (any(Rhat$Rhat > 1.05)) {
      warning(
        "WARNING: Potential problems with MCMC convergence, consider using longer chains"
      )
    }
  } else {
    warning(
      "WARNING: MCMC convergence cannot be assessed using a single chains"
    )
  }

  # if(CI=="percentile"){
  #   CIlower=apply(nimblesamples.stacked, 2, function(x) quantile(x, .025))
  #   CIupper=apply(nimblesamples.stacked, 2, function(x) quantile(x, .975))
  #   CIresult <- data.frame(lower=CIlower, upper=CIupper)
  # } else if(CI=="hpd"){
  #   CIresult <- as.data.frame(HPDinterval(as.mcmc(nimblesamples.stacked)))
  # }

  # rownames(CIresult) <- parnames

  t1 <- proc.time()
  timedif <- t1 - t0
  runtime <- paste(
    round(as.numeric(as.difftime(timedif["elapsed"], units = "secs")), 1),
    "seconds"
  )

  prc_excluded <- round(100 * (1 - length(included.replicates) / ndoses), 1)
  if (length(included.replicates) / ndoses < .8) {
    warning(paste0(
      "WARNING: ",
      prc_excluded,
      "% of replicates excluded from model averaging. Try different bounds or starting values."
    ))
  }

  out <- list(
    coefficients = coef,
    sd = sd,
    #CI=CIresult,
    Rhat = Rhat,
    samples = nimblesamples,
    included.replicates = included.replicates,
    runtime = runtime
  )
  if (family == "prophaz") {
    out <- c(
      out,
      list(
        prophaz_timepoints = prophaz_timepoints
      )
    )
  }
  return(out)
}

snapshot_combinations <- list(
  list(family = "binomial", Y = "Y.binomial", doseRRmod = "ERR", deg = 1),
  list(family = "binomial", Y = "Y.binomial", doseRRmod = "ERR", deg = 2),
  list(family = "binomial", Y = "Y.binomial", doseRRmod = "EXP", deg = 1),
  list(family = "binomial", Y = "Y.binomial", doseRRmod = "EXP", deg = 2),
  list(family = "binomial", Y = "Y.binomial", doseRRmod = "LINEXP", deg = 2),
  list(family = "poisson", Y = "Y.poisson", doseRRmod = "ERR", deg = 1),
  list(family = "poisson", Y = "Y.poisson", doseRRmod = "EXP", deg = 2),
  list(family = "gaussian", Y = "Y.gaussian", doseRRmod = NULL, deg = 1),
  list(family = "gaussian", Y = "Y.gaussian", doseRRmod = NULL, deg = 2),
  list(family = "prophaz", Y = "status", doseRRmod = "EXP", deg = 1),
  list(family = "multinomial", Y = "Y.multinomial", doseRRmod = "EXP", deg = 1),
  list(family = "clogit", Y = "Y.clogit", doseRRmod = "EXP", deg = 1)
)

covariate_combinations <- list(
  list(X = NULL, M = NULL),
  list(X = "X1", M = NULL),
  list(X = c("X1", "X2"), M = NULL),
  list(X = NULL, M = "M1"),
  list(X = "X1", M = "M1"),
  list(X = c("X1", "X2"), M = "M1"),
  list(X = NULL, M = c("M1", "M2")),
  list(X = "X1", M = c("M1", "M2")),
  list(X = c("X1", "X2"), M = c("M1", "M2"))
)

all_methods <- c("RC", "ERC", "MCML", "FMA", "BMA")

fit_combination <- function(
  family,
  Y,
  doseRRmod = NULL,
  deg,
  X,
  M,
  methods = "RC",
  data,
  niter.BMA = 1000,
  nburnin.BMA = 200,
  nthin.BMA = 1,
  transform = NULL,
  transform.jacobian = NULL,
  ...
) {
  if (length(M) > 0) {
    Mstring <- paste(M, collapse = "+")
  } else {
    Mstring <- "NULL"
  }
  if (length(X) > 0) {
    Xstring <- paste("+", X, collapse = "+")
  } else {
    Xstring <- ""
  }

  if (family == "gaussian") {
    myform <- as.formula(paste0(
      Y,
      "~dose(V1:V10, modifier=",
      Mstring,
      ", deg=",
      deg,
      ")",
      Xstring
    ))
  } else if (family == "prophaz") {
    myform <- as.formula(paste0(
      "Surv(time,",
      Y,
      ")",
      "~dose(V1:V10, modifier=",
      Mstring,
      ", deg=",
      deg,
      ",model=",
      doseRRmod,
      ")",
      Xstring
    ))
  } else if (family == "clogit") {
    myform <- as.formula(paste0(
      Y,
      "~dose(V1:V10, modifier=",
      Mstring,
      ", deg=",
      deg,
      ",model=",
      doseRRmod,
      ")",
      Xstring,
      "+ strata(setnr)"
    ))
  } else {
    myform <- as.formula(paste0(
      Y,
      "~dose(V1:V10, modifier=",
      Mstring,
      ", deg=",
      deg,
      ",model=",
      doseRRmod,
      ")",
      Xstring
    ))
  }

  args <- c(
    list(
      formula = myform,
      data = data,
      family = family,
      methods = methods,
      nburnin.BMA = nburnin.BMA,
      niter.BMA = niter.BMA,
      nthin.BMA = nthin.BMA,
      transform = transform,
      transform.jacobian = transform.jacobian
    ),
    list(...)
  )

  suppressWarnings(do.call(ameras, args))
}

set.seed(123)
data("data", package = "ameras")


for (method in c("RC", "ERC", "MCML")) {
  test_that(paste("proflik/wald.transformed snapshot:", method), {
    if (method %in% c("ERC", "MCML")) {
      skip_on_cran()
    }

    fit <- fit_combination(
      family = "binomial",
      Y = "Y.binomial",
      doseRRmod = "ERR",
      deg = 2,
      X = NULL,
      M = NULL,
      data = data,
      methods = method
    )
    fit1 <- confint(fit, type = c("proflik"))
    expect_snapshot(fit1[[method]]$coefficients)
    expect_snapshot(fit1[[method]]$sd)
    expect_snapshot(fit1[[method]]$CI)

    fit2 <- confint(fit, type = c("wald.transformed"))
    expect_snapshot(fit2[[method]]$coefficients)
    expect_snapshot(fit2[[method]]$sd)
    expect_snapshot(fit2[[method]]$CI)
  })
}


for (method in c("FMA", "BMA")) {
  test_that(paste("percentile/hpd snapshot:", method), {
    skip_on_cran()

    fit <- fit_combination(
      family = "binomial",
      Y = "Y.binomial",
      doseRRmod = "ERR",
      deg = 2,
      X = NULL,
      M = NULL,
      data = data,
      methods = method
    )
    fit1 <- confint(fit, type = c("percentile"))
    expect_snapshot(fit1[[method]]$coefficients)
    expect_snapshot(fit1[[method]]$sd)
    expect_snapshot(fit1[[method]]$CI)

    fit2 <- confint(fit, type = c("hpd"))
    expect_snapshot(fit2[[method]]$coefficients)
    expect_snapshot(fit2[[method]]$sd)
    expect_snapshot(fit2[[method]]$CI)
  })
}

set.seed(123)
data("data", package = "ameras")


# Gaussian, basic model, all methods

for (method in all_methods) {
  test_that(paste("gaussian snapshot:", method), {
    if (method == "BMA") {
      skip_on_cran()
    }

    fit <- fit_combination(
      family = "gaussian",
      Y = "Y.gaussian",
      deg = 2,
      X = NULL,
      M = NULL,
      data = data,
      methods = method,
      niter.BMA = 1000,
      nburnin.BMA = 500
    )
    fit <- confint(fit, type = c("wald.orig", "percentile"))
    expect_snapshot_value(
      coef(fit),
      tolerance = 1e-4,
      style = "deparse"
    )
    expect_snapshot_value(fit[[method]]$sd, tolerance = 1e-4, style = "deparse")
    expect_snapshot_value(fit[[method]]$CI, tolerance = 1e-4, style = "deparse")
    expect_snapshot(print(summary(fit)$summary_table, digits = 2))
  })
}

set.seed(123)
data("data", package = "ameras")


# Multinomial, basic model, all methods

for (method in all_methods) {
  test_that(paste("multinomial snapshot:", method), {
    if (method %in% c("ERC", "MCML", "BMA")) {
      skip_on_cran()
    }

    fit <- fit_combination(
      family = "multinomial",
      Y = "Y.multinomial",
      doseRRmod = "EXP",
      deg = 2,
      X = NULL,
      M = NULL,
      data = data,
      methods = method,
      niter.BMA = 1000,
      nburnin.BMA = 500
    )
    fit <- confint(fit, type = c("wald.orig", "percentile"))
    expect_snapshot(fit[[method]]$coefficients)
    expect_snapshot(fit[[method]]$sd)
    expect_snapshot(fit[[method]]$CI)
  })
}

transform1 <- function(
  params,
  index.t = 1:length(params),
  lowlimit = rep(0, length(index.t)),
  boundcheck = FALSE,
  boundtol = 1e-3,
  ...
) {
  # Transforms from reparametrized to original scale, i.e., xi to beta

  if (length(index.t) != length(lowlimit)) {
    stop("Length mismatch between index.t and lowlimit")
  }
  if (any(!(index.t %in% 1:length(params)))) {
    stop("Incorrect indices for transformation specified")
  }
  params[index.t] <- exp(params[index.t]) + lowlimit
  if (boundcheck) {
    if (any(params[index.t] - lowlimit < boundtol)) {
      warning(paste0(
        "WARNING: one or multiple parameter estimates within ",
        boundtol,
        " of lower bounds. Try different bounds or starting values."
      ))
    }
  }
  return(params)
}
transform1.inv <- function(
  params,
  index.t = 1:length(params),
  lowlimit = rep(0, length(index.t)),
  ...
) {
  # Transforms from original scale to reparametrized, i.e., beta to xi
  if (length(index.t) != length(lowlimit)) {
    stop("Length mismatch between index.t and lowlimit")
  }
  if (any(!(index.t %in% 1:length(params)))) {
    stop("Incorrect indices for transformation specified")
  }
  params[index.t] <- log(params[index.t] - lowlimit)
  return(params)
}
transform1.jacobian <- function(params, index.t = 1:length(params), ...) {
  # Transforms from reparametrized to original scale, i.e., xi to beta
  if (any(!(index.t %in% 1:length(params)))) {
    stop("Incorrect indices for transformation specified")
  }
  grad <- rep(1, length(params))
  grad[index.t] <- exp(params[index.t])
  if (length(params) > 1) {
    return(diag(grad))
  } else {
    return(matrix(grad))
  }
}

new_amerasfit <- function(x = list()) {
  stopifnot(is.list(x))
  structure(
    x,
    class = "amerasfit"
  )
}


print.amerasfit <- function(x, digits = max(3, getOption("digits") - 3), ...) {
  object0 <- x[intersect(names(x), c("RC", "ERC", "MCML", "FMA", "BMA"))]

  coefs <- coef.amerasfit(x, ...)

  runtime_table <- do.call(
    "rbind",
    lapply(1:length(object0), function(i) {
      y <- object0[[i]]
      method <- names(object0)[i]

      runtime <- as.numeric(strsplit(y$runtime, " seconds")[[1]])

      res <- data.frame(Method = method, Runtime = runtime)
      rownames(res) <- NULL
      res
    })
  )

  total_runtime_seconds <- sum(sapply(object0, function(x) {
    as.numeric(strsplit(x$runtime, " seconds")[[1]])
  }))

  cat("Call:\n")
  print(x$call)

  cat(paste0("\nNumber of rows: ", x$num.rows, "\n"))
  cat(paste0("Number of dose replicates: ", x$num.replicates, "\n"))

  cat(paste0("\nTotal run time: ", total_runtime_seconds, " seconds\n\n"))

  cat("Runtime in seconds by method:\n\n")
  print(format(runtime_table, digits = digits, nsmall = 1), row.names = FALSE)

  cat("\nEstimated model parameters:\n\n")
  print(format(coefs, digits = digits, nsmall = 1), row.names = TRUE)

  invisible(x)
}


coef.amerasfit <- function(object, ...) {
  object <- object[intersect(
    names(object),
    c("RC", "ERC", "MCML", "FMA", "BMA")
  )]

  bma <- ("BMA" %in% names(object))

  res <- as.data.frame(do.call(
    "cbind",
    lapply(1:length(object), function(i) {
      y <- object[[i]]

      if (bma) {
        tmp <- c(
          y$coefficients,
          object$BMA$coefficients[-(1:length(y$coefficients))] * NA
        )
      } else {
        tmp <- y$coefficients
      }

      tmp
    })
  ))
  names(res) <- names(object)
  res
}


summary.amerasfit <- function(object, ...) {
  object0 <- object[intersect(
    names(object),
    c("RC", "ERC", "MCML", "FMA", "BMA")
  )]

  bma <- ("BMA" %in% names(object0))

  summary_table <- do.call(
    "rbind",
    lapply(1:length(object0), function(i) {
      y <- object0[[i]]
      method <- names(object0)[i]

      coef <- y$coefficients

      se <- y$sd

      res <- data.frame(
        Method = method,
        Term = names(coef),
        Estimate = coef,
        SE = se
      )

      if (object$CI.computed) {
        CI <- y$CI
        CI.lowerbound <- CI.upperbound <- coef * NA
        CI.lowerbound[match(rownames(CI), names(coef))] <- CI$lower
        CI.upperbound[match(rownames(CI), names(coef))] <- CI$upper

        res <- cbind(
          res,
          data.frame(
            CI.lowerbound = CI.lowerbound,
            CI.upperbound = CI.upperbound
          )
        )
        # Add p-value columns for profile likelihood intervals. If not used, these will be removed after
        res <- cbind(res, data.frame(pval.lower = NA, pval.upper = NA))

        if ("pval.lower" %in% names(CI) && "pval.upper" %in% names(CI)) {
          pval.lower <- pval.upper <- coef * NA
          pval.lower[match(rownames(CI), names(coef))] <- CI$pval.lower
          pval.upper[match(rownames(CI), names(coef))] <- CI$pval.upper
          res$pval.lower <- pval.lower
          res$pval.upper <- pval.upper
          #res <- cbind(res, data.frame(pval.lower = pval.lower, pval.upper = pval.upper))
        }
      }

      if (bma) {
        if (method == "BMA") {
          res <- cbind(res, y$Rhat)
        } else {
          res <- cbind(res, data.frame(Rhat = NA, n.eff = NA))
        }
      }
      rownames(res) <- NULL
      res
    })
  )

  # Remove p-value columns if not used
  if (object$CI.computed) {
    if (
      all(is.na(summary_table$pval.lower)) &
        all(is.na(summary_table$pval.upper))
    ) {
      summary_table <- summary_table[,
        -which(names(summary_table) %in% c("pval.lower", "pval.upper"))
      ]
    }
  }
  runtime_table <- do.call(
    "rbind",
    lapply(1:length(object0), function(i) {
      y <- object0[[i]]
      method <- names(object0)[i]

      runtime <- as.numeric(strsplit(y$runtime, " seconds")[[1]])

      res <- data.frame(Method = method, Runtime = runtime)
      rownames(res) <- NULL
      res
    })
  )

  total_runtime_seconds <- sum(sapply(object0, function(x) {
    as.numeric(strsplit(x$runtime, " seconds")[[1]])
  }))

  ans <- list(
    call = object$call,
    summary_table = summary_table,
    runtime_table = runtime_table,
    total_runtime_seconds = total_runtime_seconds,
    CI.computed = object$CI.computed
  )

  class(ans) <- "summary.amerasfit"
  return(ans)
}


print.summary.amerasfit <- function(
  x,
  digits = max(3, getOption("digits") - 3),
  ...
) {
  cat("Call:\n")
  print(x$call)
  cat(paste0("\nTotal run time: ", x$total_runtime_seconds, " seconds\n\n"))

  cat("Runtime in seconds by method:\n\n")
  print(format(x$runtime_table, digits = digits, nsmall = 1), row.names = FALSE)

  cat("\nSummary of coefficients by method:\n\n")
  print(format(x$summary_table, digits = digits, nsmall = 2), row.names = FALSE)

  if (!x$CI.computed) {
    cat(
      "\nNote: confidence intervals not yet computed.",
      "Use confint() to add them."
    )
  }

  invisible(x)
}

traceplot <- function(object, ...) {
  UseMethod("traceplot")
}

traceplot.amerasfit <- function(
  object,
  iter = 5000,
  Rhat = TRUE,
  n.eff = TRUE,
  pdf = FALSE,
  ...
) {
  if ("BMA" %in% names(object)) {
    MCMCtrace(
      object$BMA$samples,
      iter = iter,
      Rhat = Rhat,
      n.eff = n.eff,
      pdf = pdf,
      ...
    )
  } else {
    stop(
      "ERROR: traceplot() requires BMA output in the provided 'amerasfit' object."
    )
  }
}


confint.amerasfit <- function(
  object,
  parm = "dose",
  level = 0.95,
  type = c("proflik", "percentile"),
  maxit.profCI = 20,
  tol.profCI = 1e-2,
  data = NULL,
  ...
) {
  # Validate level
  if (!is.numeric(level) || length(level) != 1 || level <= 0 || level >= 1) {
    stop("level must be a single numeric value between 0 and 1")
  }

  fitobj <- object[intersect(
    names(object),
    c("RC", "ERC", "MCML", "FMA", "BMA")
  )]

  res <- NULL
  for (it in 1:length(fitobj)) {
    method <- names(fitobj)[it]

    if (method %in% c("FMA", "BMA")) {
      type.i <- match.arg(
        type,
        choices = c("percentile", "hpd"),
        several.ok = TRUE
      )
      if (length(type.i) != 1) {
        stop(
          "ERROR: Exactly one CI type (either 'percentile' or 'hpd') for FMA and/or BMA should be specified"
        )
      }

      if (method == "BMA") {
        pars <- names(fitobj[[it]]$coefficients)
        samples <- fitobj[[it]]$samples
        if (is.list(samples)) {
          samples <- do.call("rbind", samples)
        }
        samples <- samples[, pars]
      } else {
        samples <- fitobj[[it]]$samples
      }

      object[[method]]$CI <- compute_sample_CI(
        samples = samples,
        level = level,
        type = type.i
      )
    } else {
      # RC, ERC, MCML
      type.i <- match.arg(
        type,
        choices = c("wald.orig", "wald.transformed", "proflik"),
        several.ok = TRUE
      )
      if (length(type.i) != 1) {
        stop(
          "ERROR: Exactly one CI type (one of 'proflik', 'wald.orig', or 'wald.transformed') for RC, ERC, and/or MCML should be specified"
        )
      }
      if (!is.null(parm) && type.i != "proflik") {
        #warning("parm is only used when method='proflik' and will be ignored")
      }

      if (type.i == "wald.orig") {
        object[[method]]$CI <- compute_wald_CI(
          method_fit = fitobj[[it]],
          transform = object$transform,
          level = level,
          type = "orig",
          other.args = object$other.args
        )
      } else if (type.i == "wald.transformed") {
        object[[method]]$CI <- compute_wald_CI(
          method_fit = fitobj[[it]],
          transform = object$transform,
          level = level,
          type = "transformed",
          other.args = object$other.args
        )
      } else if (type.i == "proflik") {
        resolved_data <- resolve_data(object, data)
        object[[method]]$CI <- compute_proflik_CI(
          method_fit = fitobj[[it]],
          object = object,
          method_name = method,
          data = resolved_data,
          parm = parm,
          level = level,
          maxit.profCI = maxit.profCI,
          tol.profCI = tol.profCI,
          optim.method = object$model$optim.method,
          control = object$model$control
        )
        rm(resolved_data)
      }
    }
  }

  object$CI.computed <- TRUE

  return(object)
}

boundedSliceSampler <- nimbleFunction(
  contains = nimble::sampler_BASE,
  setup = function(model, mvSaved, target, control) {
    # Create internal slice sampler safely
    sliceSampler <- nimble::sampler_slice(
      model = model,
      mvSaved = mvSaved,
      target = target,
      control = control
    )

    # Optionally accept user-supplied K from control
    if (!is.null(control$K)) {
      K <- control$K
    } else {
      # Default: try to detect length of p if it exists
      if ("p" %in% model$getNodeNames()) {
        K <- length(model[['p']])
      } else {
        stop(
          "Please provide K (number of categories) in 'control' for boundedSliceSampler."
        )
      }
    }
  },
  run = function() {
    sliceSampler$run()

    value <- model[[target]]

    # Apply integer bounds
    if (value < 1) {
      value <- 1
    }
    if (value > K) {
      value <- K
    }

    model[[target]] <<- value
  },
  methods = list(
    reset = function() {
      sliceSampler$reset()
    }
  )
)


nimblemod <- nimbleCode({
  col.ind ~ dcat(w[1:K])
  for (k in 1:K) {
    vec[k] <- equals(col.ind, k)
  }

  if (family != "multinomial") {
    if (!(family %in% c("prophaz", "clogit"))) {
      a0 ~ dnorm(0, 0.001)
    } else {
      a0 <- 0
    }

    if (Xlen > 1) {
      for (k in 1:Xlen) {
        a[k] ~ dnorm(0, 0.001)
      }
    } else if (Xlen == 1) {
      a ~ dnorm(0, .001)
    }

    if (doseRRmod == "EXP" | family == "gaussian") {
      # Normal priors with large variance
      if (deg > 1) {
        for (k in 1:deg) {
          b[k] ~ dnorm(0, 0.001)
        }
      } else {
        b ~ dnorm(0, .001)
      }

      if (Mlen > 1) {
        if (deg == 1) {
          for (k in 1:Mlen) {
            bm[k] ~ dnorm(0, 0.001)
          }
        } else if (deg > 1) {
          for (k in 1:Mlen) {
            bm1[k] ~ dnorm(0, 0.001)
            bm2[k] ~ dnorm(0, 0.001)
          }
        }
      } else if (Mlen == 1) {
        if (deg == 1) {
          bm ~ dnorm(0, 0.001)
        } else if (deg > 1) {
          bm1 ~ dnorm(0, 0.001)
          bm2 ~ dnorm(0, 0.001)
        }
      }
    } else if (doseRRmod == "ERR") {
      if (ERRprior == "truncated_horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )
        if (deg > 1) {
          for (kk in 1:2) {
            # horseshoe prior
            lambda[kk] ~ T(dt(0, 1, df = 1), 0, )
            b[kk] ~ T(dnorm(0, sd = lambda[kk] * tau), 0, )
          }
        } else {
          lambda ~ T(dt(0, 1, df = 1), 0, )
          b ~ T(dnorm(0, sd = lambda * tau), 0, )
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (zz in 1:Mlen) {
              lambda_m[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm[zz] ~ T(dnorm(0, sd = lambda_m[zz] * tau), 0, )
            }
          } else if (deg > 1) {
            for (zz in 1:Mlen) {
              # horseshoe prior
              lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz] ~ T(dnorm(0, sd = lambda_m1[zz] * tau), 0, )
              lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz] ~ T(dnorm(0, sd = lambda_m2[zz] * tau), 0, )
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            lambda_m ~ T(dt(0, 1, df = 1), 0, )
            bm ~ T(dnorm(0, sd = lambda_m * tau), 0, )
          } else if (deg > 1) {
            lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
            bm1 ~ T(dnorm(0, sd = lambda_m1 * tau), 0, )
            lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
            bm2 ~ T(dnorm(0, sd = lambda_m2 * tau), 0, )
          }
        }
      } else if (ERRprior == "truncated_normal") {
        if (deg > 1) {
          for (k in 1:deg) {
            b[k] ~ T(dnorm(0, 0.001), 0, )
          }
        } else {
          b ~ T(dnorm(0, .001), 0, )
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (k in 1:Mlen) {
              bm[k] ~ T(dnorm(0, 0.001), 0, )
            }
          } else if (deg > 1) {
            for (k in 1:Mlen) {
              bm1[k] ~ T(dnorm(0, 0.001), 0, )
              bm2[k] ~ T(dnorm(0, 0.001), 0, )
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            bm ~ T(dnorm(0, 0.001), 0, )
          } else if (deg > 1) {
            bm1 ~ T(dnorm(0, 0.001), 0, )
            bm2 ~ T(dnorm(0, 0.001), 0, )
          }
        }
      } else if (ERRprior == "truncated_doubleexponential") {
        if (deg > 1) {
          for (kk in 1:2) {
            lambda[kk] ~ T(dt(0, 1, df = 1), 0, )
            b[kk] ~ T(ddexp(0.0, lambda[kk]), 0, )
          }
        } else {
          lambda ~ T(dt(0, 1, df = 1), 0, )
          b ~ T(ddexp(0.0, lambda), 0, )
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (zz in 1:Mlen) {
              lambda_m[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm[zz] ~ T(ddexp(0.0, lambda_m[zz]), 0, )
            }
          } else if (deg > 1) {
            for (zz in 1:Mlen) {
              lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz] ~ T(ddexp(0.0, lambda_m1[zz]), 0, )
              lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz] ~ T(ddexp(0.0, lambda_m2[zz]), 0, )
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            lambda_m ~ T(dt(0, 1, df = 1), 0, )
            bm ~ T(ddexp(0.0, lambda_m), 0, )
          } else if (deg > 1) {
            lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
            bm1 ~ T(ddexp(0.0, lambda_m1), 0, )
            lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
            bm2 ~ T(ddexp(0.0, lambda_m2), 0, )
          }
        }
      } else if (ERRprior == "horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )
        if (deg > 1) {
          for (kk in 1:2) {
            # horseshoe prior
            lambda[kk] ~ T(dt(0, 1, df = 1), 0, )
            b[kk] ~ dnorm(0, sd = lambda[kk] * tau)
          }
        } else {
          lambda ~ T(dt(0, 1, df = 1), 0, )
          b ~ dnorm(0, sd = lambda * tau)
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (zz in 1:Mlen) {
              lambda_m[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm[zz] ~ dnorm(0, sd = lambda_m[zz] * tau)
            }
          } else if (deg > 1) {
            for (zz in 1:Mlen) {
              # horseshoe prior
              lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz] ~ dnorm(0, sd = lambda_m1[zz] * tau)
              lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz] ~ dnorm(0, sd = lambda_m2[zz] * tau)
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            lambda_m ~ T(dt(0, 1, df = 1), 0, )
            bm ~ dnorm(0, sd = lambda_m * tau)
          } else if (deg > 1) {
            lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
            bm1 ~ dnorm(0, sd = lambda_m1 * tau)
            lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
            bm2 ~ dnorm(0, sd = lambda_m2 * tau)
          }
        }
      } else if (ERRprior == "normal") {
        if (deg > 1) {
          for (k in 1:deg) {
            b[k] ~ dnorm(0, 0.001)
          }
        } else {
          b ~ dnorm(0, .001)
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (k in 1:Mlen) {
              bm[k] ~ dnorm(0, 0.001)
            }
          } else if (deg > 1) {
            for (k in 1:Mlen) {
              bm1[k] ~ dnorm(0, 0.001)
              bm2[k] ~ dnorm(0, 0.001)
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            bm ~ dnorm(0, 0.001)
          } else if (deg > 1) {
            bm1 ~ dnorm(0, 0.001)
            bm2 ~ dnorm(0, 0.001)
          }
        }
      } else if (ERRprior == "doubleexponential") {
        if (deg > 1) {
          for (kk in 1:2) {
            lambda[kk] ~ T(dt(0, 1, df = 1), 0, )
            b[kk] ~ ddexp(0.0, lambda[kk])
          }
        } else {
          lambda ~ T(dt(0, 1, df = 1), 0, )
          b ~ ddexp(0.0, lambda)
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (zz in 1:Mlen) {
              lambda_m[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm[zz] ~ ddexp(0.0, lambda_m[zz])
            }
          } else if (deg > 1) {
            for (zz in 1:Mlen) {
              lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz] ~ ddexp(0.0, lambda_m1[zz])
              lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz] ~ ddexp(0.0, lambda_m2[zz])
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            lambda_m ~ T(dt(0, 1, df = 1), 0, )
            bm ~ ddexp(0.0, lambda_m)
          } else if (deg > 1) {
            lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
            bm1 ~ ddexp(0.0, lambda_m1)
            lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
            bm2 ~ ddexp(0.0, lambda_m2)
          }
        }
      }
    } else if (doseRRmod == "LINEXP") {
      if (ERRprior == "truncated_horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )

        # horseshoe prior
        lambda[1] ~ T(dt(0, 1, df = 1), 0, )
        b[1] ~ T(dnorm(0, sd = lambda[1] * tau), 0, )
        lambda[2] ~ T(dt(0, 1, df = 1), 0, )
        b[2] ~ dnorm(0, sd = lambda[2] * tau)

        if (Mlen > 1) {
          for (zz in 1:Mlen) {
            # horseshoe prior
            lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm1[zz] ~ T(dnorm(0, sd = lambda_m1[zz] * tau), 0, )
            lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm2[zz] ~ dnorm(0, sd = lambda_m2[zz] * tau)
          }
        } else if (Mlen == 1) {
          lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
          bm1 ~ T(dnorm(0, sd = lambda_m1 * tau), 0, )
          lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
          bm2 ~ dnorm(0, sd = lambda_m2 * tau)
        }
      } else if (ERRprior == "truncated_normal") {
        b[1] ~ T(dnorm(0, 0.001), 0, )
        b[2] ~ dnorm(0, 0.001)

        if (Mlen > 1) {
          for (k in 1:Mlen) {
            bm1[k] ~ T(dnorm(0, 0.001), 0, )
            bm2[k] ~ dnorm(0, 0.001)
          }
        } else if (Mlen == 1) {
          bm1 ~ T(dnorm(0, 0.001), 0, )
          bm2 ~ dnorm(0, 0.001)
        }
      } else if (ERRprior == "truncated_doubleexponential") {
        lambda[1] ~ T(dt(0, 1, df = 1), 0, )
        b[1] ~ T(ddexp(0.0, lambda[1]), 0, )
        lambda[2] ~ T(dt(0, 1, df = 1), 0, )
        b[2] ~ ddexp(0.0, lambda[2])

        if (Mlen > 1) {
          for (zz in 1:Mlen) {
            lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm1[zz] ~ T(ddexp(0.0, lambda_m1[zz]), 0, )
            lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm2[zz] ~ ddexp(0.0, lambda_m2[zz])
          }
        } else if (Mlen == 1) {
          lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
          bm1 ~ T(ddexp(0.0, lambda_m1), 0, )
          lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
          bm2 ~ ddexp(0.0, lambda_m2)
        }
      } else if (ERRprior == "horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )

        for (kk in 1:2) {
          # horseshoe prior
          lambda[kk] ~ T(dt(0, 1, df = 1), 0, )
          b[kk] ~ dnorm(0, sd = lambda[kk] * tau)
        }

        if (Mlen > 1) {
          for (zz in 1:Mlen) {
            # horseshoe prior
            lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm1[zz] ~ dnorm(0, sd = lambda_m1[zz] * tau)
            lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm2[zz] ~ dnorm(0, sd = lambda_m2[zz] * tau)
          }
        } else if (Mlen == 1) {
          lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
          bm1 ~ dnorm(0, sd = lambda_m1 * tau)
          lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
          bm2 ~ dnorm(0, sd = lambda_m2 * tau)
        }
      } else if (ERRprior == "normal") {
        for (k in 1:2) {
          b[k] ~ dnorm(0, 0.001)
        }

        if (Mlen > 1) {
          for (k in 1:Mlen) {
            bm1[k] ~ dnorm(0, 0.001)
            bm2[k] ~ dnorm(0, 0.001)
          }
        } else if (Mlen == 1) {
          bm1 ~ dnorm(0, 0.001)
          bm2 ~ dnorm(0, 0.001)
        }
      } else if (ERRprior == "doubleexponential") {
        for (kk in 1:2) {
          lambda[kk] ~ T(dt(0, 1, df = 1), 0, )
          b[kk] ~ ddexp(0.0, lambda[kk])
        }

        if (Mlen > 1) {
          for (zz in 1:Mlen) {
            lambda_m1[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm1[zz] ~ ddexp(0.0, lambda_m1[zz])
            lambda_m2[zz] ~ T(dt(0, 1, df = 1), 0, )
            bm2[zz] ~ ddexp(0.0, lambda_m2[zz])
          }
        } else if (Mlen == 1) {
          lambda_m1 ~ T(dt(0, 1, df = 1), 0, )
          bm1 ~ ddexp(0.0, lambda_m1)
          lambda_m2 ~ T(dt(0, 1, df = 1), 0, )
          bm2 ~ ddexp(0.0, lambda_m2)
        }
      }
    }
  } else {
    # Multinomial

    for (z in 1:(Z - 1)) {
      a0[z] ~ dnorm(0, 0.001)

      if (Xlen > 1) {
        for (k in 1:Xlen) {
          a[k, z] ~ dnorm(0, 0.001)
        }
      } else if (Xlen == 1) {
        a[z] ~ dnorm(0, .001)
      }
    }

    if (doseRRmod == "EXP") {
      # Normal priors with large variance
      if (deg > 1) {
        for (z in 1:(Z - 1)) {
          for (k in 1:deg) {
            b[k, z] ~ dnorm(0, 0.001)
          }
        }
      } else {
        for (z in 1:(Z - 1)) {
          b[z] ~ dnorm(0, .001)
        }
      }

      if (Mlen > 1) {
        if (deg == 1) {
          for (z in 1:(Z - 1)) {
            for (k in 1:Mlen) {
              bm[k, z] ~ dnorm(0, 0.001)
            }
          }
        } else if (deg > 1) {
          for (z in 1:(Z - 1)) {
            for (k in 1:Mlen) {
              bm1[k, z] ~ dnorm(0, 0.001)
              bm2[k, z] ~ dnorm(0, 0.001)
            }
          }
        }
      } else if (Mlen == 1) {
        if (deg == 1) {
          for (z in 1:(Z - 1)) {
            bm[z] ~ dnorm(0, 0.001)
          }
        } else if (deg > 1) {
          for (z in 1:(Z - 1)) {
            bm1[z] ~ dnorm(0, 0.001)
            bm2[z] ~ dnorm(0, 0.001)
          }
        }
      }
    } else if (doseRRmod == "ERR") {
      if (ERRprior == "truncated_horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )
        if (deg > 1) {
          for (z in 1:(Z - 1)) {
            for (kk in 1:2) {
              # horseshoe prior
              lambda[kk, z] ~ T(dt(0, 1, df = 1), 0, )
              b[kk, z] ~ T(dnorm(0, sd = lambda[kk, z] * tau), 0, )
            }
          }
        } else {
          for (z in 1:(Z - 1)) {
            lambda[z] ~ T(dt(0, 1, df = 1), 0, )
            b[z] ~ T(dnorm(0, sd = lambda[z] * tau), 0, )
          }
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                lambda_m[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm[zz, z] ~ T(dnorm(0, sd = lambda_m[zz, z] * tau), 0, )
              }
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                # horseshoe prior
                lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm1[zz, z] ~ T(dnorm(0, sd = lambda_m1[zz, z] * tau), 0, )
                lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm2[zz, z] ~ T(dnorm(0, sd = lambda_m2[zz, z] * tau), 0, )
              }
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              lambda_m[z] ~ T(dt(0, 1, df = 1), 0, )
              bm[z] ~ T(dnorm(0, sd = lambda_m[z] * tau), 0, )
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[z] ~ T(dnorm(0, sd = lambda_m1[z] * tau), 0, )
              lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[z] ~ T(dnorm(0, sd = lambda_m2[z] * tau), 0, )
            }
          }
        }
      } else if (ERRprior == "truncated_normal") {
        if (deg > 1) {
          for (z in 1:(Z - 1)) {
            for (k in 1:deg) {
              b[k, z] ~ T(dnorm(0, 0.001), 0, )
            }
          }
        } else {
          for (z in 1:(Z - 1)) {
            b[z] ~ T(dnorm(0, .001), 0, )
          }
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              for (k in 1:Mlen) {
                bm[k, z] ~ T(dnorm(0, 0.001), 0, )
              }
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              for (k in 1:Mlen) {
                bm1[k, z] ~ T(dnorm(0, 0.001), 0, )
                bm2[k, z] ~ T(dnorm(0, 0.001), 0, )
              }
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              bm[z] ~ T(dnorm(0, 0.001), 0, )
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              bm1[z] ~ T(dnorm(0, 0.001), 0, )
              bm2[z] ~ T(dnorm(0, 0.001), 0, )
            }
          }
        }
      } else if (ERRprior == "truncated_doubleexponential") {
        if (deg > 1) {
          for (z in 1:(Z - 1)) {
            for (kk in 1:2) {
              lambda[kk, z] ~ T(dt(0, 1, df = 1), 0, )
              b[kk, z] ~ T(ddexp(0.0, lambda[kk, z]), 0, )
            }
          }
        } else {
          for (z in 1:(Z - 1)) {
            lambda[z] ~ T(dt(0, 1, df = 1), 0, )
            b[z] ~ T(ddexp(0.0, lambda), 0, )
          }
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                lambda_m[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm[zz, z] ~ T(ddexp(0.0, lambda_m[zz, z]), 0, )
              }
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm1[zz, z] ~ T(ddexp(0.0, lambda_m1[zz, z]), 0, )
                lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm2[zz, z] ~ T(ddexp(0.0, lambda_m2[zz, z]), 0, )
              }
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              lambda_m[z] ~ T(dt(0, 1, df = 1), 0, )
              bm[z] ~ T(ddexp(0.0, lambda_m[z]), 0, )
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[z] ~ T(ddexp(0.0, lambda_m1[z]), 0, )
              lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[z] ~ T(ddexp(0.0, lambda_m2[z]), 0, )
            }
          }
        }
      } else if (ERRprior == "horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )
        if (deg > 1) {
          for (z in 1:(Z - 1)) {
            for (kk in 1:2) {
              # horseshoe prior
              lambda[kk, z] ~ T(dt(0, 1, df = 1), 0, )
              b[kk, z] ~ dnorm(0, sd = lambda[kk, z] * tau)
            }
          }
        } else {
          for (z in 1:(Z - 1)) {
            lambda[z] ~ T(dt(0, 1, df = 1), 0, )
            b[z] ~ dnorm(0, sd = lambda[z] * tau)
          }
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                lambda_m[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm[zz, z] ~ dnorm(0, sd = lambda_m[zz, z] * tau)
              }
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                # horseshoe prior
                lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm1[zz, z] ~ dnorm(0, sd = lambda_m1[zz, z] * tau)
                lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm2[zz, z] ~ dnorm(0, sd = lambda_m2[zz, z] * tau)
              }
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              lambda_m[z] ~ T(dt(0, 1, df = 1), 0, )
              bm[z] ~ dnorm(0, sd = lambda_m[z] * tau)
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[z] ~ dnorm(0, sd = lambda_m1[z] * tau)
              lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[z] ~ dnorm(0, sd = lambda_m2[z] * tau)
            }
          }
        }
      } else if (ERRprior == "normal") {
        if (deg > 1) {
          for (z in 1:(Z - 1)) {
            for (k in 1:deg) {
              b[k, z] ~ dnorm(0, 0.001)
            }
          }
        } else {
          for (z in 1:(Z - 1)) {
            b[z] ~ dnorm(0, .001)
          }
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              for (k in 1:Mlen) {
                bm[k, z] ~ dnorm(0, 0.001)
              }
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              for (k in 1:Mlen) {
                bm1[k, z] ~ dnorm(0, 0.001)
                bm2[k, z] ~ dnorm(0, 0.001)
              }
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              bm[z] ~ dnorm(0, 0.001)
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              bm1[z] ~ dnorm(0, 0.001)
              bm2[z] ~ dnorm(0, 0.001)
            }
          }
        }
      } else if (ERRprior == "doubleexponential") {
        if (deg > 1) {
          for (z in 1:(Z - 1)) {
            for (kk in 1:2) {
              lambda[kk, z] ~ T(dt(0, 1, df = 1), 0, )
              b[kk, z] ~ ddexp(0.0, lambda[kk, z])
            }
          }
        } else {
          for (z in 1:(Z - 1)) {
            lambda[z] ~ T(dt(0, 1, df = 1), 0, )
            b[z] ~ ddexp(0.0, lambda[z])
          }
        }

        if (Mlen > 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                lambda_m[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm[zz, z] ~ ddexp(0.0, lambda_m[zz, z])
              }
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              for (zz in 1:Mlen) {
                lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm1[zz, z] ~ ddexp(0.0, lambda_m1[zz, z])
                lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
                bm2[zz, z] ~ ddexp(0.0, lambda_m2[zz, z])
              }
            }
          }
        } else if (Mlen == 1) {
          if (deg == 1) {
            for (z in 1:(Z - 1)) {
              lambda_m[z] ~ T(dt(0, 1, df = 1), 0, )
              bm[z] ~ ddexp(0.0, lambda_m[z])
            }
          } else if (deg > 1) {
            for (z in 1:(Z - 1)) {
              lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[z] ~ ddexp(0.0, lambda_m1[z])
              lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[z] ~ ddexp(0.0, lambda_m2[z])
            }
          }
        }
      }
    } else if (doseRRmod == "LINEXP") {
      if (ERRprior == "truncated_horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )
        for (z in 1:(Z - 1)) {
          # horseshoe prior
          lambda[1, z] ~ T(dt(0, 1, df = 1), 0, )
          b[1, z] ~ T(dnorm(0, sd = lambda[1, z] * tau), 0, )
          lambda[2, z] ~ T(dt(0, 1, df = 1), 0, )
          b[2, z] ~ dnorm(0, sd = lambda[2, z] * tau)
        }

        if (Mlen > 1) {
          for (z in 1:(Z - 1)) {
            for (zz in 1:Mlen) {
              # horseshoe prior
              lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz, z] ~ T(dnorm(0, sd = lambda_m1[zz, z] * tau), 0, )
              lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz, z] ~ dnorm(0, sd = lambda_m2[zz, z] * tau)
            }
          }
        } else if (Mlen == 1) {
          for (z in 1:(Z - 1)) {
            lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
            bm1[z] ~ T(dnorm(0, sd = lambda_m1[z] * tau), 0, )
            lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
            bm2[z] ~ dnorm(0, sd = lambda_m2[z] * tau)
          }
        }
      } else if (ERRprior == "truncated_normal") {
        for (z in 1:(Z - 1)) {
          b[1, z] ~ T(dnorm(0, 0.001), 0, )
          b[2, z] ~ dnorm(0, 0.001)
        }
        if (Mlen > 1) {
          for (z in 1:(Z - 1)) {
            for (k in 1:Mlen) {
              bm1[k, z] ~ T(dnorm(0, 0.001), 0, )
              bm2[k, z] ~ dnorm(0, 0.001)
            }
          }
        } else if (Mlen == 1) {
          for (z in 1:(Z - 1)) {
            bm1[z] ~ T(dnorm(0, 0.001), 0, )
            bm2[z] ~ dnorm(0, 0.001)
          }
        }
      } else if (ERRprior == "truncated_doubleexponential") {
        for (z in 1:(Z - 1)) {
          lambda[1, z] ~ T(dt(0, 1, df = 1), 0, )
          b[1, z] ~ T(ddexp(0.0, lambda[1, z]), 0, )
          lambda[2, z] ~ T(dt(0, 1, df = 1), 0, )
          b[2, z] ~ ddexp(0.0, lambda[2, z])
        }
        if (Mlen > 1) {
          for (z in 1:(Z - 1)) {
            for (zz in 1:Mlen) {
              lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz, z] ~ T(ddexp(0.0, lambda_m1[zz, z]), 0, )
              lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz, z] ~ ddexp(0.0, lambda_m2[zz, z])
            }
          }
        } else if (Mlen == 1) {
          for (z in 1:(Z - 1)) {
            lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
            bm1[z] ~ T(ddexp(0.0, lambda_m1[z]), 0, )
            lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
            bm2[z] ~ ddexp(0.0, lambda_m2[z])
          }
        }
      } else if (ERRprior == "horseshoe") {
        tau ~ T(dt(0, 1, df = 1), 0, )
        for (z in 1:(Z - 1)) {
          for (kk in 1:2) {
            # horseshoe prior
            lambda[kk, z] ~ T(dt(0, 1, df = 1), 0, )
            b[kk, z] ~ dnorm(0, sd = lambda[kk, z] * tau)
          }
        }

        if (Mlen > 1) {
          for (z in 1:(Z - 1)) {
            for (zz in 1:Mlen) {
              # horseshoe prior
              lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz, z] ~ dnorm(0, sd = lambda_m1[zz, z] * tau)
              lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz, z] ~ dnorm(0, sd = lambda_m2[zz, z] * tau)
            }
          }
        } else if (Mlen == 1) {
          for (z in 1:(Z - 1)) {
            lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
            bm1[z] ~ dnorm(0, sd = lambda_m1[z] * tau)
            lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
            bm2[z] ~ dnorm(0, sd = lambda_m2[z] * tau)
          }
        }
      } else if (ERRprior == "normal") {
        for (z in 1:(Z - 1)) {
          for (k in 1:2) {
            b[k, z] ~ dnorm(0, 0.001)
          }
        }

        if (Mlen > 1) {
          for (z in 1:(Z - 1)) {
            for (k in 1:Mlen) {
              bm1[k, z] ~ dnorm(0, 0.001)
              bm2[k, z] ~ dnorm(0, 0.001)
            }
          }
        } else if (Mlen == 1) {
          for (z in 1:(Z - 1)) {
            bm1[z] ~ dnorm(0, 0.001)
            bm2[z] ~ dnorm(0, 0.001)
          }
        }
      } else if (ERRprior == "doubleexponential") {
        for (z in 1:(Z - 1)) {
          for (kk in 1:2) {
            lambda[kk, z] ~ T(dt(0, 1, df = 1), 0, )
            b[kk, z] ~ ddexp(0.0, lambda[kk, z])
          }
        }

        if (Mlen > 1) {
          for (z in 1:(Z - 1)) {
            for (zz in 1:Mlen) {
              lambda_m1[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm1[zz, z] ~ ddexp(0.0, lambda_m1[zz, z])
              lambda_m2[zz, z] ~ T(dt(0, 1, df = 1), 0, )
              bm2[zz, z] ~ ddexp(0.0, lambda_m2[zz, z])
            }
          }
        } else if (Mlen == 1) {
          for (z in 1:(Z - 1)) {
            lambda_m1[z] ~ T(dt(0, 1, df = 1), 0, )
            bm1[z] ~ ddexp(0.0, lambda_m1[z])
            lambda_m2[z] ~ T(dt(0, 1, df = 1), 0, )
            bm2[z] ~ ddexp(0.0, lambda_m2[z])
          }
        }
      }
    }
  }

  if (family == "gaussian") {
    sigma ~ T(dnorm(0, 0.001), 0, )
  }
  if (family == "prophaz") {
    for (j in 1:prophaz_numints) {
      h0[j] ~ dgamma(0.01, 0.01)
    }
  }

  # Likelihood

  if (family == "prophaz") {
    for (i in 1:N) {
      # Pieces of the cumulative baseline hazard function
      for (k in int.entry[i]:int.exit[i]) {
        start_t[i, k] <- max(prophaz_timepoints[k], entry[i])
        end_t[i, k] <- min(prophaz_timepoints[k + 1], exit[i])
        HH[i, k] <- max(end_t[i, k] - start_t[i, k], 0) * h0[k]
      }
      # Cumulative baseline hazard
      H[i] <- sum(HH[i, int.entry[i]:int.exit[i]])
    }
  }

  for (i in 1:N) {
    dose[i] <- inprod(dosemat[i, 1:K], vec[1:K])
    if (family != "multinomial") {
      if (doseRRmod != "LINEXP") {
        if (Mlen == 0) {
          if (deg == 2) {
            dosepart[i] <- b[1] * dose[i] + b[2] * dose[i] * dose[i]
          } else if (deg == 1) {
            dosepart[i] <- b * dose[i]
          }
        } else if (Mlen == 1) {
          if (deg == 2) {
            dosepart[i] <- (b[1] + bm1 * Mmat[i]) *
              dose[i] +
              (b[2] + bm2 * Mmat[i]) * dose[i] * dose[i]
          } else if (deg == 1) {
            dosepart[i] <- (b + bm * Mmat[i]) * dose[i]
          }
        } else {
          if (deg == 2) {
            dosepart[i] <- (b[1] + inprod(Mmat[i, ], bm1[1:Mlen])) *
              dose[i] +
              (b[2] + inprod(Mmat[i, ], bm2[1:Mlen])) * dose[i] * dose[i]
          } else if (deg == 1) {
            dosepart[i] <- (b + inprod(Mmat[i, ], bm[1:Mlen])) * dose[i]
          }
        }
      } else {
        # LINEXP
        if (Mlen == 0) {
          dosepart[i] <- b[1] * dose[i] * exp(b[2] * dose[i])
        } else if (Mlen == 1) {
          dosepart[i] <- (b[1] + bm1 * Mmat[i]) *
            dose[i] *
            exp((b[2] + bm2 * Mmat[i]) * dose[i])
        } else {
          dosepart[i] <- (b[1] + inprod(Mmat[i, ], bm1[1:Mlen])) *
            dose[i] *
            exp((b[2] + inprod(Mmat[i, ], bm2[1:Mlen])) * dose[i])
        }
      }

      if (Xlen > 1) {
        Xlinpred[i] <- a0 + inprod(Xmat[i, ], a[1:Xlen])
      } else if (Xlen == 1) {
        Xlinpred[i] <- a0 + Xmat[i] * a
      } else if (Xlen == 0) {
        Xlinpred[i] <- a0
      }
      if (family == "gaussian") {
        mu[i] <- dosepart[i] + Xlinpred[i]
        Y[i] ~ dnorm(mu[i], sd = sigma)
      } else if (family == "binomial") {
        if (doseRRmod %in% c("ERR", "LINEXP")) {
          mu[i] <- (1 + dosepart[i]) * exp(Xlinpred[i])
        } else if (doseRRmod == "EXP") {
          mu[i] <- exp(dosepart[i] + Xlinpred[i])
        }
        prob[i] <- mu[i] / (1 + mu[i])
        Y[i] ~ dbinom(prob[i], size = 1)
      } else if (family == "poisson") {
        if (doseRRmod %in% c("ERR", "LINEXP")) {
          mu[i] <- (1 + dosepart[i]) * exp(Xlinpred[i]) * P[i]
        } else if (doseRRmod == "EXP") {
          mu[i] <- exp(dosepart[i] + Xlinpred[i]) * P[i]
        }
        Y[i] ~ dpois(mu[i])
      } else if (family == "prophaz") {
        # Relative risk
        if (doseRRmod %in% c("ERR", "LINEXP")) {
          R[i] <- (1 + dosepart[i]) * exp(Xlinpred[i])
        } else if (doseRRmod == "EXP") {
          R[i] <- exp(dosepart[i] + Xlinpred[i])
        }

        # Log-hazard
        logHaz[i] <- log(h0[int.exit[i]] * R[i])
        # Log-survival
        logSurv[i] <- -H[i] * R[i]

        # Using the zeros trick
        phi[i] <- 100000 - delta[i] * logHaz[i] - logSurv[i]
        zeros[i] ~ dpois(phi[i])
      } else if (family == "clogit") {
        # Relative risk
        if (doseRRmod %in% c("ERR", "LINEXP")) {
          R[i] <- (1 + dosepart[i]) * exp(Xlinpred[i])
        } else if (doseRRmod == "EXP") {
          R[i] <- exp(dosepart[i] + Xlinpred[i])
        }
      }
    } else {
      # multinomial
      for (z in 1:(Z - 1)) {
        if (doseRRmod != "LINEXP") {
          if (Mlen == 0) {
            if (deg == 2) {
              dosepart[i, z] <- b[1, z] * dose[i] + b[2, z] * dose[i] * dose[i]
            } else if (deg == 1) {
              dosepart[i, z] <- b[z] * dose[i]
            }
          } else if (Mlen == 1) {
            if (deg == 2) {
              dosepart[i, z] <- (b[1, z] + bm1[z] * Mmat[i]) *
                dose[i] +
                (b[2, z] + bm2[z] * Mmat[i]) * dose[i] * dose[i]
            } else if (deg == 1) {
              dosepart[i, z] <- (b[z] + bm[z] * Mmat[i]) * dose[i]
            }
          } else {
            if (deg == 2) {
              dosepart[i, z] <- (b[1, z] + inprod(Mmat[i, ], bm1[1:Mlen, z])) *
                dose[i] +
                (b[2, z] + inprod(Mmat[i, ], bm2[1:Mlen, z])) *
                  dose[i] *
                  dose[i]
            } else if (deg == 1) {
              dosepart[i, z] <- (b[z] + inprod(Mmat[i, ], bm[1:Mlen, z])) *
                dose[i]
            }
          }
        } else {
          # LINEXP
          if (Mlen == 0) {
            dosepart[i, z] <- b[1, z] * dose[i] * exp(b[2, z] * dose[i])
          } else if (Mlen == 1) {
            dosepart[i, z] <- (b[1, z] + bm1[z] * Mmat[i]) *
              dose[i] *
              exp((b[2, z] + bm2[z] * Mmat[i]) * dose[i])
          } else {
            dosepart[i, z] <- (b[1, z] + inprod(Mmat[i, ], bm1[1:Mlen, z])) *
              dose[i] *
              exp((b[2, z] + inprod(Mmat[i, ], bm2[1:Mlen, z])) * dose[i])
          }
        }

        if (Xlen > 1) {
          Xlinpred[i, z] <- a0[z] + inprod(Xmat[i, ], a[1:Xlen, z])
        } else if (Xlen == 1) {
          Xlinpred[i, z] <- a0[z] + Xmat[i] * a[z]
        } else if (Xlen == 0) {
          Xlinpred[i, z] <- a0[z]
        }

        if (doseRRmod %in% c("ERR", "LINEXP")) {
          mu[i, z] <- (1 + dosepart[i, z]) * exp(Xlinpred[i, z])
        } else if (doseRRmod == "EXP") {
          mu[i, z] <- exp(dosepart[i, z] + Xlinpred[i, z])
        }
      }

      mu[i, Z] <- 1
      probs[i, 1:Z] <- mu[i, 1:Z] / sum(mu[i, 1:Z])
      Y[i, 1:Z] ~ dmulti(prob = probs[i, 1:Z], size = 1)
    }
  }

  if (family == "clogit") {
    for (setnr in 1:nsets) {
      #probs[setnr, 1:N] <- R[1:N]*setmat[setnr,1:N]/inprod(setmat[setnr, 1:N], R[1:N])
      #Ymat[setnr, 1:N] ~ dmulti(prob=probs[setnr, 1:N], size=1)
      probs[set_start[setnr]:set_end[setnr]] <- R[
        set_start[setnr]:set_end[setnr]
      ] /
        sum(R[set_start[setnr]:set_end[setnr]])
      Y[set_start[setnr]:set_end[setnr]] ~ dmulti(
        probs[set_start[setnr]:set_end[setnr]],
        size = 1
      )
    }
  }
})

dldd_clogit <- function(designmat, RRs) {
  nr <- nrow(designmat)
  nc <- ncol(designmat)
  ret <- as.numeric(rep(-9999, nc))

  # designmat must be passed in as a vector with columns stacked
  tmp <- .C(
    "C_dldd_clogit",
    as.numeric(designmat),
    as.integer(nr),
    as.integer(nc),
    as.numeric(RRs),
    ret = ret,
    PACKAGE = "ameras"
  )
  ret <- tmp$ret
  ret
}

dldd_prophaz <- function(entry, exit, status, RRs) {
  n <- length(entry) # Number of individuals
  ret <- as.numeric(rep(-9999, n)) # Initialize return vector with placeholders

  # Ensure all inputs are numeric vectors (entry, exit, status, RRs)
  entry <- as.numeric(entry)
  exit <- as.numeric(exit)
  status <- as.integer(status)
  RRs <- as.numeric(RRs)

  # Call the C function using .C
  tmp <- .C(
    "C_dldd_prophaz",
    as.numeric(entry),
    as.numeric(exit),
    as.integer(status),
    as.numeric(RRs),
    as.integer(n),
    ret = ret,
    PACKAGE = "ameras"
  )

  # Return the result as a numeric vector
  ret <- tmp$ret
  return(ret)
}


compute_ERCmatrix_clogit <- function(designmat, RRs, drdd, drdd2, status) {
  nr <- nrow(designmat)
  nc <- ncol(designmat)
  ret <- as.numeric(rep(-9999, nc * nc))

  tmp <- .C(
    "C_compute_ERCmatrix_clogit",
    as.numeric(designmat),
    as.integer(nr),
    as.integer(nc),
    as.numeric(RRs),
    as.numeric(drdd),
    as.numeric(drdd2),
    as.integer(status),
    ret = ret,
    PACKAGE = "ameras"
  )
  ret <- matrix(tmp$ret, nrow = nc, ncol = nc, byrow = FALSE)
  ret
}


compute_ERCmatrix_prophaz <- function(entry, exit, status, RRs, drdd, drdd2) {
  n <- length(entry) # Number of individuals
  ret <- as.numeric(rep(-9999, n * n)) # Initialize return vector with placeholders

  # Ensure all inputs are numeric vectors (entry, exit, status, RRs)
  entry <- as.numeric(entry)
  exit <- as.numeric(exit)
  status <- as.integer(status)
  RRs <- as.numeric(RRs)
  drdd <- as.numeric(drdd)
  drdd2 <- as.numeric(drdd2)

  # Call the C function using .C
  tmp <- .C(
    "C_compute_ERCmatrix_prophaz",
    as.numeric(entry),
    as.numeric(exit),
    as.integer(status),
    as.numeric(RRs),
    as.numeric(drdd),
    as.numeric(drdd2),
    as.integer(n),
    ret = ret,
    PACKAGE = "ameras"
  )

  # Return the result as a numeric vector
  ret <- matrix(tmp$ret, nrow = n, ncol = n, byrow = FALSE)
  ret
  return(ret)
}

loglik_prophaz_rcpp <- function(
  exit_t,
  entry_t,
  RR_entry,
  RR_exit,
  status_ord,
  loglim = 1e-30
) {
  n <- length(exit_t)
  K <- ncol(RR_exit)

  # sanity checks (minimal)
  stopifnot(
    length(entry_t) == n,
    length(status_ord) == n,
    nrow(RR_entry) == n,
    nrow(RR_exit) == n,
    ncol(RR_entry) == K
  )

  ret <- double(K)

  tmp <- .C(
    "C_loglik_prophaz_rcpp",
    exit_t = as.double(exit_t),
    entry_t = as.double(entry_t),
    RR_entry = as.double(RR_entry),
    RR_exit = as.double(RR_exit),
    status_ord = as.integer(status_ord),
    n = as.integer(n),
    K = as.integer(K),
    loglim = as.double(loglim),
    ret = ret,
    PACKAGE = "ameras"
  )

  tmp$ret
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

dldd_clogit <- function(designmat, RRs) {
    .Call(`_ameras_dldd_clogit`, designmat, RRs)
}

dldd_prophaz <- function(entry_t, exit_t, status_ord, RRs) {
    .Call(`_ameras_dldd_prophaz`, entry_t, exit_t, status_ord, RRs)
}

compute_ERCmatrix_prophaz <- function(entry_t, exit_t, status_ord, RRs, drdd, drdd2) {
    .Call(`_ameras_compute_ERCmatrix_prophaz`, entry_t, exit_t, status_ord, RRs, drdd, drdd2)
}

compute_ERCmatrix_clogit <- function(designmat, RRs, drdd, drdd2, status) {
    .Call(`_ameras_compute_ERCmatrix_clogit`, designmat, RRs, drdd, drdd2, status)
}

loglik_prophaz_rcpp <- function(exit_t, entry_t, RR_entry, RR_exit, status_ord, loglim) {
    .Call(`_ameras_loglik_prophaz_rcpp`, exit_t, entry_t, RR_entry, RR_exit, status_ord, loglim)
}

compute_sample_CI <- function(samples, type = "percentile", level = .95) {
  if (type == "percentile") {
    CIlower = apply(samples, 2, function(x) quantile(x, (1 - level) / 2))
    CIupper = apply(samples, 2, function(x) quantile(x, (1 + level) / 2))
    CIresult <- data.frame(lower = CIlower, upper = CIupper)
  } else if (type == "hpd") {
    CIresult <- as.data.frame(
      coda::HPDinterval(coda::as.mcmc(samples)),
      prob = level
    )
  }
  return(CIresult)
}

compute_wald_CI <- function(
  method_fit,
  level = .95,
  transform = NULL,
  other.args = NULL,
  type = "orig"
) {
  if (type == "transformed" & is.null(transform)) {
    stop(
      "No transformation specified, specify transformation or choose a different CI type"
    )
  }

  z <- qnorm(1 - (1 - level) / 2)
  hessian <- method_fit$optim$hessian

  if (type == "transformed") {
    # Check invertibility
    invertible <- det(hessian) != 0 &&
      rcond(hessian) > .Machine$double.eps &&
      all(eigen(hessian)$values > 0)

    if (!invertible) {
      warning(
        "Hessian not invertible, confidence intervals could not be obtained"
      )
      na_vec <- NA * method_fit$coefficients
      return(data.frame(lower = na_vec, upper = na_vec))
    }

    hess_inv <- solve(hessian)

    CIlower <- do.call(
      transform,
      c(
        list(params = method_fit$optim$par - z * sqrt(diag(hess_inv))),
        other.args
      )
    )
    CIupper <- do.call(
      transform,
      c(
        list(params = method_fit$optim$par + z * sqrt(diag(hess_inv))),
        other.args
      )
    )
  } else if (type == "orig") {
    CIlower <- method_fit$coefficients - z * sqrt(diag(method_fit$vcov))
    CIupper <- method_fit$coefficients + z * sqrt(diag(method_fit$vcov))
  }

  CIresult <- data.frame(
    lower = CIlower,
    upper = CIupper,
    row.names = names(method_fit$coefficients)
  )
}

compute_proflik_CI <- function(
  method_fit,
  object,
  method_name,
  data,
  parm = "dose",
  level = 0.95,
  maxit.profCI = 20,
  tol.profCI = 1e-2,
  optim.method = "Nelder-Mead",
  control = list(reltol = 1e-10)
) {
  alpha <- 1 - level
  optval <- -method_fit$loglik # stored as positive log-likelihood
  inpar <- method_fit$optim$par
  parnames <- names(method_fit$coefficients)

  # Determine which parameters to compute CIs for
  if (identical(parm, "dose")) {
    CIindices <- which(
      startsWith(parnames, "dose") |
        grepl(")_dose", parnames)
    )
  } else if (identical(parm, "all")) {
    CIindices <- seq_along(parnames)
  } else {
    CIindices <- which(parnames %in% parm)
    if (!length(CIindices)) {
      stop(
        "No parameters matched parm. Available parameters are: ",
        paste(parnames, collapse = ", ")
      )
    }
  }

  # Reconstruct the likelihood function from stored model
  loglik_fn <- make_loglik_fn(object, method_name, method_fit, data)

  # Use stored hessian to get sensible search bounds
  hessian <- method_fit$optim$hessian
  invertible <- det(hessian) != 0 &&
    rcond(hessian) > .Machine$double.eps &&
    all(eigen(hessian)$values > 0)

  if (invertible) {
    se <- sqrt(diag(solve(hessian)))
    lowlims <- inpar - 4 * se
    uplims <- inpar + 4 * se
  } else {
    warning(
      "Hessian not invertible, using default search bounds for profile ",
      "likelihood CI. Results may be unreliable."
    )
    lowlims <- rep(-20, length(inpar))
    uplims <- rep(10, length(inpar))
  }

  # Initialise output vectors
  CIlower <- rep(NA_real_, length(parnames))
  CIupper <- rep(NA_real_, length(parnames))
  pval_lower <- rep(NA_real_, length(parnames))
  pval_upper <- rep(NA_real_, length(parnames))
  iter_lower <- rep(NA_integer_, length(parnames))
  iter_upper <- rep(NA_integer_, length(parnames))

  for (myindex in seq_along(parnames)) {
    if (!(myindex %in% CIindices)) {
      next
    }

    message("Obtaining profile likelihood CI for ", parnames[myindex])

    profCI_one <- compute_proflik_ci_one(
      index = myindex,
      inpar = inpar,
      optval = optval,
      loglik_fn = loglik_fn,
      lowlim = lowlims[myindex],
      uplim = uplims[myindex],
      alpha = alpha,
      maxit.profCI = maxit.profCI,
      tol.profCI = tol.profCI,
      optim.method = optim.method,
      control = control,
      parname = parnames[myindex],
      transform = object$transform,
      other.args = object$other.args
    )

    CIlower[myindex] <- profCI_one$lower
    CIupper[myindex] <- profCI_one$upper
    pval_lower[myindex] <- profCI_one$pval.lower
    pval_upper[myindex] <- profCI_one$pval.upper
    iter_lower[myindex] <- profCI_one$iter.lower
    iter_upper[myindex] <- profCI_one$iter.upper
  }

  # Back-transform if needed
  if (!is.null(object$transform)) {
    CIlower <- do.call(
      object$transform,
      c(list(params = CIlower), object$other.args)
    )
    CIupper <- do.call(
      object$transform,
      c(list(params = CIupper), object$other.args)
    )
  }

  result <- data.frame(
    lower = CIlower,
    upper = CIupper,
    pval.lower = pval_lower,
    pval.upper = pval_upper,
    iter.lower = iter_lower,
    iter.upper = iter_upper,
    row.names = parnames
  )

  result[CIindices, , drop = FALSE]
}


compute_proflik_ci_one <- function(
  index,
  inpar,
  optval,
  loglik_fn,
  lowlim,
  uplim,
  alpha = 0.05,
  maxit.profCI = 20,
  tol.profCI = 1e-2,
  optim.method = "Nelder-Mead",
  control = list(reltol = 1e-10),
  parname = "parameter",
  transform = NULL,
  other.args = NULL
) {
  profCI_fn <- function(par) {
    sapply(par, function(mypar) {
      1 -
        pchisq(
          2 *
            (proflik(
              parvalue = mypar,
              index = index,
              fun = loglik_fn,
              inpar = inpar,
              optim.method = optim.method,
              control = control
            ) -
              optval),
          df = 1
        ) -
        alpha
    })
  }

  # Upper bound
  val_at_15 <- profCI_fn(15)
  if (val_at_15 > 0) {
    if (is.null(transform)) {
      warning(
        "Upper bound for ",
        parname,
        " is > 15 and may not exist. ",
        "Consider rescaling the variable."
      )
    } else {
      warning(
        "Upper bound for ",
        parname,
        " is > ",
        round(
          do.call(
            transform,
            c(list(params = rep(15, length(inpar))), other.args)
          )[index],
          1
        ),
        " and may not exist. ",
        "Consider rescaling the variable."
      )
    }

    uproot <- list(root = Inf, f.root = val_at_15, iter = NA_integer_)
  } else {
    uproot <- tryCatch(
      uniroot(
        profCI_fn,
        lower = inpar[index],
        upper = uplim,
        extendInt = "downX",
        maxiter = maxit.profCI,
        tol = tol.profCI
      ),
      error = function(e) {
        list(root = NA_real_, f.root = NA_real_, iter = NA_integer_)
      }
    )
    if (!is.na(uproot$f.root) && abs(uproot$f.root) > 0.005) {
      warning(
        "P-value for ",
        parname,
        " upper bound is more than 0.005 from ",
        alpha,
        ". Consider reducing tol.profCI or increasing maxit.profCI."
      )
    }
  }

  # Lower bound
  val_at_minus10 <- profCI_fn(-10)
  if (val_at_minus10 > 0) {
    if (is.null(transform)) {
      warning(
        "Lower bound for ",
        parname,
        " is < -10 and may not exist. ",
        "Consider rescaling the variable."
      )
    } else {
      warning(
        "Lower bound for ",
        parname,
        " is < ",
        round(
          do.call(
            transform,
            c(list(params = rep(-10, length(inpar))), other.args)
          )[index],
          1
        ),
        " and may not exist. ",
        "Consider rescaling the variable or using a different transformation."
      )
    }

    lowroot <- list(root = -Inf, f.root = val_at_minus10, iter = NA_integer_)
  } else {
    lowroot <- tryCatch(
      uniroot(
        profCI_fn,
        lower = lowlim,
        upper = inpar[index],
        extendInt = "upX",
        maxiter = maxit.profCI,
        tol = tol.profCI
      ),
      error = function(e) {
        list(root = NA_real_, f.root = NA_real_, iter = NA_integer_)
      }
    )
    if (!is.na(lowroot$f.root) && abs(lowroot$f.root) > 0.005) {
      warning(
        "P-value for ",
        parname,
        " lower bound is more than 0.005 from ",
        alpha,
        ". Consider reducing tol.profCI or increasing maxit.profCI."
      )
    }
  }

  list(
    lower = lowroot$root,
    upper = uproot$root,
    pval.lower = lowroot$f.root + alpha,
    pval.upper = uproot$f.root + alpha,
    iter.lower = lowroot$iter,
    iter.upper = uproot$iter
  )
}

proflik <- function(
  parvalue,
  index,
  fun,
  inpar,
  optim.method = optim.method,
  control = list(reltol = 1e-10),
  ...
) {
  if (length(inpar) == 1) {
    # 1-parameter model -> just return likelihood itself

    return(fun(params = parvalue, ...))
  } else if (length(inpar) == 2) {
    # 2-parameter model -> 1 parameter optimization for profile likelihood -> use optimize

    innerfun <- function(params, ...) {
      if (index > 1) {
        params <- c(params[1:(index - 1)], parvalue, params[-(1:(index - 1))])
      } else {
        params <- c(parvalue, params)
      }

      fun(params = params, ...)
    }

    innerfit <- optimize(innerfun, lower = -10, upper = 10, ...)
    return(innerfit$objective)
  } else {
    # 3+ parameter model -> 2+ parameter optimization for profile likelihood -> use optim

    innerfun <- function(params, ...) {
      if (index > 1) {
        params <- c(params[1:(index - 1)], parvalue, params[-(1:(index - 1))])
      } else {
        params <- c(parvalue, params)
      }

      fun(params = params, ...)
    }
    innerfit <- optim(
      par = inpar[-index],
      fn = innerfun,
      method = optim.method,
      control = control,
      ...
    )
    if (optim.method == "Nelder-Mead") {
      count0 <- innerfit$counts
      innerfit <- optim(
        par = innerfit$par,
        fn = innerfun,
        method = "BFGS",
        control = control,
        ...
      )
      innerfit$counts <- replace(count0, is.na(count0), 0) +
        replace(innerfit$counts, is.na(innerfit$counts), 0)
    }
    return(innerfit$value)
  }
}

---
title: "Relative risk models"
output: 
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Relative risk models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(ameras)
library(ggplot2)
data(data, package="ameras")
```

# Introduction

For non-Gaussian families, three relative risk models for the main exposure are supported, the usual exponential model $$RR_i=\exp(\beta_1 D_i+\beta_2 D_i^2+ \mathbf{M}_i^T \mathbf{\beta}_{m1}D_i + \mathbf{M}_i^T \mathbf{\beta}_{m2} D_i^2),$$ the linear(-quadratic) excess relative risk (ERR) model $$RR_i= 1+\beta_1 D_i+\beta_2 D_i^2 + \mathbf{M}_i^T \mathbf{\beta_{m1}}D_i + \mathbf{M}_i^T \mathbf{\beta}_{m2}D_i^2,$$ and the linear-exponential model $$ RR_i= 1+(\beta_1 + \mathbf{M}_i^T \mathbf{\beta}_{m1}) D_i \exp\{(\beta_2+  \mathbf{M}_i^T \mathbf{\beta}_{m2})D_i\}. $$ This vignette illustrates fitting the three models using regression calibration for logistic regression, but the same syntax applies to all other settings.

# Exponential relative risk

The usual exponential relative risk model is given by $RR_i=\exp(\beta_1 D_i+\beta_2 D_i^2+ \mathbf{M}_i^T \mathbf{\beta}_{m1}D_i + \mathbf{M}_i^T \mathbf{\beta}_{m2} D_i^2)$, where the quadratic and effect modification terms are optional (not fit by setting `deg=1` and not passing anything to `M`, respectively). This model is fit by setting `model="EXP"` as follows:

```{r modelfit.exp}
fit.ameras.exp <- ameras(Y.binomial~dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                         data=data, family="binomial", methods="RC")
summary(fit.ameras.exp)
```




# Linear excess relative risk

The linear excess relative risk model is given by $RR_i=1+\beta_1 D_i+\beta_2 D_i^2+ \mathbf{M}_i^T \mathbf{\beta}_{m1}D_i + \mathbf{M}_i^T \mathbf{\beta}_{m2} D_i^2$, where again the quadratic and effect modification terms are optional. In this case, no degree needs to be specified. This model is fit by setting `model="ERR"` as follows:

```{r modelfit.err}
fit.ameras.err <- ameras(Y.binomial~dose(V1:V10, deg=2, model="ERR")+X1+X2, 
                         data=data, family="binomial", methods="RC")
summary(fit.ameras.err)
```


# Linear-exponential relative risk

The linear-exponential relative risk model is given by $RR_i= 1+(\beta_1 + \mathbf{M}_i^T \mathbf{\beta}_{m1}) D_i \exp\{(\beta_2+  \mathbf{M}_i^T \mathbf{\beta}_{m2})D_i\}$, where the effect modification terms are optional. This model is fit by setting `model="LINEXP"` as follows:

```{r modelfit.linexp}
fit.ameras.linexp <- ameras(Y.binomial~dose(V1:V10, model="LINEXP")+X1+X2, 
                         data=data, family="binomial", methods="RC")
summary(fit.ameras.linexp)
```


# Comparison between models

To compare between models, it is easiest to do so visually:

```{r comparison, fig.width=7, fig.height=6}
ggplot(data.frame(x=c(0, 5)), aes(x))+
  theme_light(base_size=15)+
  xlab("Exposure")+
  ylab("Relative risk")+
  labs(col="Model", lty="Model") +
  theme(legend.position = "inside", 
        legend.position.inside = c(.2,.85),
        legend.box.background = element_rect(color = "black", fill = "white", linewidth = 1))+
  stat_function(aes(col="Linear-quadratic ERR", lty="Linear-quadratic ERR" ),fun=function(x){
    1+fit.ameras.err$RC$coefficients["dose"]*x + fit.ameras.err$RC$coefficients["dose_squared"]*x^2
  }, linewidth=1.2) + 
  stat_function(aes(col="Exponential", lty="Exponential"),fun=function(x){
    exp(fit.ameras.exp$RC$coefficients["dose"]*x + fit.ameras.exp$RC$coefficients["dose_squared"]*x^2)
  }, linewidth=1.2) +
  stat_function(aes(col="Linear-exponential", lty="Linear-exponential"),fun=function(x){
    1+fit.ameras.linexp$RC$coefficients["dose_linear"]*x * exp(fit.ameras.linexp$RC$coefficients["dose_exponential"]*x)
  }, linewidth=1.2)


```

---
title: "Confidence intervals"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Confidence intervals}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(ameras)

```

# Introduction
To compute confidence intervals, first fit the model using `ameras`, then use the `confint` method to attach confidence intervals.
Several types of confidence intervals are supported, which should be supplied to the `type` argument of `confint`, see below. When `confint` is called with `methods` containing at least one of `RC`, `ERC`, and `MCML` and at least one of `FMA` and `BMA`, `type` should be a vector of length 2 with one method for `RC`, `ERC` and `MCML` and one for `FMA` and `BMA`.

# Regression calibration, extended regression calibration, and Monte Carlo maximum likelihood

For (extended) regression calibration and Monte Carlo maximum likelihood, 
Wald and profile likelihood intervals can be obtained. When a parameter 
transformation $\mathbf\theta = h(\mathbf\eta)$ is used, `type="wald.transformed"`
yields the CI at significance level $\alpha$ of $h(\mathbf\eta \pm z_{1-\alpha/2} \mathbf V)$
where $z_{1-\alpha/2}$ is the $1-\alpha/2$-quantile of the standard normal distribution and $\mathbf V$ is the vector 
of standard deviations estimated using the inverse Hessian matrix (returned by `optim`), 
and `type="wald.orig"` uses the delta method to obtain the CI 
$h(\mathbf\eta)\pm z_{1-\alpha/2} \mathbf V_*$ where $\mathbf V_*$ is the vector of 
standard deviations estimated using $J H^{-1} J^T$ with $J$ the 
Jacobian of the transformation (obtained with `transform.jacobian`) and $H$ is the Hessian. When no 
transformation is used, `type="wald.orig"` should be used. 
The third option is `type="proflik"`, which uses the profile likelihood to 
compute confidence bounds. For this, the profile log (partial) likelihood for parameter $\theta_p$ is defined as
$$
    PL_p (\theta_p^*) = \max_{\mathbf\theta: \theta_p = \theta_p^*} \ell (\mathbf \theta),
$$
where $\ell$ is the log (partial) likelihood. Next, profile confidence intervals $(\theta_p^l, \theta_p^h)$ are obtained for parameter $\theta_p$ at significance level $\alpha$ by solving 
$-2 \{PL_p(\theta_p^*) - \ell(\hat{\mathbf{\theta}})\}=\chi^2_{1,1-\alpha}$ using the bisection method, with $\hat{\mathbf{\theta}}$ the maximum likelihood estimate. Note that profile likelihoods are more computationally intensive to obtain. For this reason, `confint` offers the option to only determine them for the exposure-related parameters, which is the default setting. To obtain profile likelihood intervals for all parameters, use `parm = "all"`.

To illustrate, we determine the three types of confidence intervals for a regression calibration analysis using the example data, using a linear excess relative risk model with the default exponential transformation (see [Parameter transformations](transformations.html)).

```{r fits1, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
data(data, package="ameras")
dosevars <- paste0("V", 1:10)
fit.ameras <- ameras(Y.binomial~dose(V1:V10, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods=c("RC"))

fit.ameras.waldorig <- confint(fit.ameras, type="wald.orig")
fit.ameras.waldtransformed <- confint(fit.ameras, type="wald.transformed")
fit.ameras.proflik <- confint(fit.ameras, type="proflik", parm="all")

summary(fit.ameras.waldorig)
summary(fit.ameras.waldtransformed)
summary(fit.ameras.proflik)
```

# Frequentist and Bayesian model averaging

For frequentist and Bayesian model averaging methods, the options are `percentile` which uses equal-tailed quantiles, and `hpd` which computes highest posterior density intervals using `HPDinterval` from the `coda` package, using either the FMA samples or Bayesian posterior samples.

Again, we use the example data to illustrate.

```{r fits2, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(123)
fit.ameras2 <- ameras(Y.binomial~dose(V1:V10, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods=c("FMA"))

fit.ameras.hpd <- confint(fit.ameras2, type="hpd")
fit.ameras.percentile <- confint(fit.ameras2, type="percentile")

summary(fit.ameras.hpd)
summary(fit.ameras.percentile)
```

---
title: "Parallel FMA with future"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Parallel FMA with future}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
)

future_available <- requireNamespace("future", quietly = TRUE) &&
  requireNamespace("future.apply", quietly = TRUE)
run_parallel_examples <- future_available
```

```{r setup}
library(ameras)
```

```{r data}
data(data, package = "ameras")
```

# Introduction

Frequentist model averaging (`FMA`) fits the specified model once for each dose
realization, computes AIC weights across those realization-specific fits, and
then samples from the fitted normal approximations. The realization-specific
fits are independent of one another, so they are a natural place to use
parallel computation.

`ameras()` supports optional parallel FMA through the
[`future`](https://future.futureverse.org/reference/plan.html) framework and
[`future.apply`](https://future.apply.futureverse.org/reference/future_lapply.html).
`ameras()` does not set a future plan. Instead, the user decides whether FMA
should run sequentially or in parallel by setting `future::plan()` before
calling `ameras()`.

If `future.apply` is not installed, `ameras()` falls back to the ordinary
sequential `lapply()` path.

For more detail on future plans and future-backed apply loops, see
`?future::plan` and `?future.apply::future_lapply`, or the linked Futureverse
documentation.

# Choosing a future plan

The default future plan is sequential, so this call uses the ordinary
single-process FMA path.

```{r sequential-plan, eval = FALSE}
future::plan(future::sequential)

fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
              data = data,
              family = "gaussian",
              methods = "FMA")
```

For parallel FMA on a local machine, `multisession` is usually the most portable
choice. It starts separate background R sessions and works across the main
platforms supported by R, including Windows, macOS, and Linux.

```{r multisession-plan, eval = FALSE}
future::plan(future::multisession, workers = 2)

fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
              data = data,
              family = "gaussian",
              methods = "FMA")

future::plan(future::sequential)
```

The `multicore` backend can be efficient on systems that support forking, but it
is not available on Windows. For code intended to run on multiple platforms,
prefer `multisession`.

For more specialized settings, the future ecosystem also supports cluster and
HPC-oriented backends. See `?future::plan` or the future package documentation
for details.

# Controlling chunk size

By default, `future.apply` chooses how to split dose realizations into futures.
For FMA, this can be adjusted with `future.chunk.size.FMA`, the average number
of dose realizations handled by each future.

Let `K` be the number of dose realizations, `W` the number of workers, and `C`
the value of `future.chunk.size.FMA`. `K` and `C` determine the approximate
number of future tasks, `ceiling(K / C)`, and at most `W` of those tasks can run
at the same time.

For example, with `K = 100` dose realizations and `W = 4` workers:

| `future.chunk.size.FMA` | Approximate chunks | Interpretation |
|---:|---:|---|
| 1 | 100 | many small tasks; more load balancing, more overhead |
| 5 | 20 | moderate load balancing and overhead |
| 25 | 4 | about one task per worker |
| 100 | 1 | one task; only one worker is used |

```{r chunk-size, eval = FALSE}
future::plan(future::multisession, workers = 2)

fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
              data = data,
              family = "gaussian",
              methods = "FMA",
              future.chunk.size.FMA = 2)

future::plan(future::sequential)
```

Smaller chunks can improve load balancing when some realizations take longer to
fit than others. Larger chunks can reduce overhead and may be preferable when
each realization-specific fit is fast. If the realization-specific fits have
similar runtimes, a reasonable starting point is a chunk size near `ceiling(K /
W)` or somewhat smaller. If runtimes vary substantially across realizations,
smaller chunks may keep workers busier.

# Example: sequential and parallel FMA

The example below compares the same Gaussian FMA fit with a sequential plan and
a two-worker multisession plan. The seed is reset before each call so that the
final FMA sampling step is comparable.

```{r compare-plans, eval = run_parallel_examples}
old_plan <- future::plan()
on.exit(future::plan(old_plan), add = TRUE)

future::plan(future::sequential)
set.seed(2024)
sequential_elapsed <- system.time(
  sequential_fit <- suppressWarnings(
    ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
           data = data,
           family = "gaussian",
           methods = "FMA",
           MFMA = 10000)
  )
)[["elapsed"]]

future::plan(future::multisession, workers = 2)
set.seed(2024)
parallel_elapsed <- system.time(
  parallel_fit <- suppressWarnings(
    ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
           data = data,
           family = "gaussian",
           methods = "FMA",
           MFMA = 10000,
           future.chunk.size.FMA = 2)
  )
)[["elapsed"]]

future::plan(old_plan)
```

Compare the coefficient estimates and elapsed runtimes.

```{r compare-results, eval = run_parallel_examples}
coefficient_comparison <- data.frame(
  term = names(sequential_fit$FMA$coefficients),
  sequential = unname(sequential_fit$FMA$coefficients),
  parallel = unname(parallel_fit$FMA$coefficients),
  row.names = NULL
)

coefficient_comparison[-1] <- round(coefficient_comparison[-1], 4)
coefficient_comparison

data.frame(
  plan = c("sequential", "multisession"),
  elapsed_seconds = round(c(sequential_elapsed, parallel_elapsed), 2)
)
```

The estimates should agree up to ordinary numerical precision. Parallel
execution is not guaranteed to be faster for small examples because starting
workers and transferring data have overhead. It is most useful when the
realization-specific fits are slow enough for that overhead to matter less.

# Reproducibility

The parallelized part of FMA is the realization-specific optimization step,
which is deterministic for ordinary model fits. The random FMA sampling step is
performed afterward in the main R process. For reproducible FMA samples, set the
seed before calling `ameras()`, just as in the sequential case.

```{r reproducibility, eval = FALSE}
future::plan(future::multisession, workers = 2)

set.seed(2024)
fit <- ameras(Y.gaussian ~ dose(V1:V10) + X1 + X2,
              data = data,
              family = "gaussian",
              methods = "FMA")

future::plan(future::sequential)
```

User-supplied functions passed through `transform`, `transform.jacobian`, or
other runtime arguments should also be deterministic. Random number generation
inside those functions is not recommended.

# Practical recommendations

Use parallel FMA when a direct `ameras(..., methods = "FMA")` call is feasible
in one R session, but the realization-specific fits are slow enough that using
multiple workers is worthwhile.

Start conservatively:

- use `future::plan(future::multisession, workers = 2)` for a portable local
  parallel plan;
- increase `workers` only after checking memory use;
- try `future.chunk.size.FMA = 2` or larger when each realization-specific fit
  is fast;
- reset the plan to `future::sequential` after the fit if the rest of the
  session should not use parallel futures.

With `multisession`, each worker is a separate R process. Large datasets may
therefore be copied to multiple workers. If memory is tight, use fewer workers,
increase the chunk size, or consider the manual FMA workflow.

# Parallel FMA or manual FMA?

The parallel option shown here is best when the built-in FMA workflow fits in
memory and you want a regular `amerasfit` object returned by `ameras()`.

The [manual FMA vignette](manualfma.html) is intended for larger or more
operationally complicated analyses where it is useful to split the
realization-specific fits across separate jobs, save intermediate fit
summaries, or resume work after partial completion. Manual FMA is more flexible,
but it requires more bookkeeping and does not return a regular `amerasfit`
object.

In short:

- use built-in parallel FMA for ordinary analyses that benefit from local
  parallel workers;
- use manual FMA when the analysis must be broken into independent chunks or
  run outside a single R session.

---
title: "Fitting models and displaying output"
output: 
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Fitting models and displaying output}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r, include = FALSE}
  knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
  )
```
  
```{r setup}
  library(ameras)
```

# Introduction

All functionality of the package is included in the function `ameras`. This vignette demonstrates the basic functionality and the generated output. For more information on obtaining confidence/credible intervals, see [Confidence intervals](confidenceintervals.html).

## Model specification

Models are specified through formulas of the form `Y~dose(dose_expression, model="ERR", deg=1, modifier=M1+M2)+X1+X2`. Here `dose_expression` specifies the dose realization columns and is parsed by `eval_select` from the `tidyselect` package. Useful examples are `D1:D1000` if the doses are in a sequence of columns with sequential names such as `D1`-`D1000`, and `all_of(dosevars)` where `dosevars` is a vector with the names of all dose columns. Further, `model` specifies, for non-Gaussian families, whether to use the exponential dose-response model (`model="EXP"`), the linear-exponential model (`model="LINEXP"`) or the linear ERR model (`model="ERR"`). Next, `deg` is used to specify whether a quadratic dose term should (`deg=2`) or should not (`deg=1`) be estimated for the exponential or linear ERR dose-response model. The `modifier` term is optional and used to specify binary effect modification variables. Finally, `X1` and `X2` above represent optional additional covariates, which can include factor variables and interactions such as `X1*X2`. Note that interactions in the modifier term are not allowed, e.g. `M1*M2`. When `deg`, `modifier`, and `model` are not supplied, the defaults are `deg=1`, no effect modifiers, and `model="ERR"`.


## Example data

The included example dataset contains 3,000 individuals with 10 exposure realizations `V1`-`V10`, binary covariates `X1` and `X2` and effect modifiers `M1` and `M2`, and outcomes of all types (`Y.gaussian`, `Y.binomial`, `Y.poisson`, `status`, `time`, `Y.multinomial`, `Y.clogit`, and `setnr`).

```{r}
data(data, package="ameras")
head(data)
```



# Linear regression & displaying output

Now we fit all methods to the data through one function call:
```{r modelfit.linreg, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(12345)
fit.ameras.linreg <- ameras(Y.gaussian~dose(V1:V10)+X1+X2, data=data, 
                            family="gaussian", niter.BMA=5000, nburnin.BMA=1000,
                            methods=c("RC", "ERC", "MCML", "FMA", "BMA"))
```


The output is a list object with a `call` component, and one component for each method, each being a list:
```{r, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
str(fit.ameras.linreg)
```

Access the results for e.g. regression calibration as follows:
```{r, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
fit.ameras.linreg$RC
```

For a summary of results, use `summary` (note that `Rhat` and `n.eff` only apply to BMA results):

```{r, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
summary(fit.ameras.linreg)
```


To extract model coefficients and compare between methods, use `coef`:
```{r, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
coef(fit.ameras.linreg)
```

To produce traceplots and density plots for BMA results, use `traceplot`:
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
traceplot(fit.ameras.linreg)
```


# Logistic regression

To fit a logistic regression model, the syntax is very similar. However, `ameras` offers three functional forms for modeling the exposure-outcome relationship. Here, we will illustrate the standard exponential relationship `model="EXP"`. For more information, see the vignette [Relative risk models](relativeriskmodels.html).

First, we fit models including a quadratic exposure term by setting `deg=2`.
```{r modelfit.logreg, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(33521)
fit.ameras.logreg <- ameras(Y.binomial~dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                            data=data, family="binomial", 
                            methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                            niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.logreg)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.logreg)
```


Without the quadratic term (`deg=1`):
```{r modelfit.logreg.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(3521216)
fit.ameras.logreg.lin <- ameras(Y.binomial~dose(V1:V10, deg=1, model="EXP")+X1+X2,  
                                data=data, family="binomial", 
                                methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.logreg.lin)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.logreg.lin)
```



# Poisson regression

For Poisson regression, an offset can be optionally used by including an `offset` term in the formula, e.g., `Y~dose(V1:V10)~X1+offset(PYR)`.
Here, we show models without using an offset, first including a quadratic exposure term:

```{r modelfit.poisson, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(332101)
fit.ameras.poisson <- ameras(Y.poisson~dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                             data=data, family="poisson", 
                             methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                             niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.poisson)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.poisson)
```



Without the quadratic term (`deg=1`):
```{r modelfit.poisson.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(24252)
fit.ameras.poisson.lin <- ameras(Y.poisson~dose(V1:V10, deg=1, model="EXP")+X1+X2, 
                                 data=data, family="poisson", 
                                 methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                 niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.poisson.lin)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.poisson.lin)
```


# Proportional hazards regression
Proportional hazards regression uses syntax similar to the `survival` package, with models specified using formulas with `Surv(exit, status)` or `Surv(entry, exit, status)` on the left hand side. Note that BMA fits a piecewise constant baseline hazard `h0` as the proportional hazards model is not directly supported. By default, the observed time interval is divided into 10 intervals using quantiles of the observed event times among cases. This number of such intervals can be specified through the `prophaz.numints.BMA` argument. The BMA output contains the `prophaz.numints.BMA+1` cutpoints defining the intervals in addition to `h0`.


Again, we first fit models including a quadratic exposure term by setting `deg=2`.

```{r modelfit.prophaz, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(332120000)
fit.ameras.prophaz <- ameras(Surv(time, status)~
                               dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                             data=data, family="prophaz", 
                             methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                             niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.prophaz)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.prophaz)
```

The BMA output now contains the intervals with piecewise constant baseline hazards, corresponding to the estimates `h0`:
```{r, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
fit.ameras.prophaz$BMA$prophaz.timepoints
```



Without the quadratic term (`deg=1`):
```{r modelfit.prophaz.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(24978252)
fit.ameras.prophaz.lin <- ameras(Surv(time, status)~
                                   dose(V1:V10, deg=1, model="EXP")+X1+X2, 
                                 data=data, family="prophaz", 
                                 methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                 niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.prophaz.lin)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.prophaz.lin)
```



# Multinomial logistic regression

For multinomial logistic regression, the last category (in the case of the example data, `Y.multinomial='3'`) is used as the referent category.

Again, we first fit models including a quadratic exposure term by setting `deg=2`.
```{r modelfit.multinomial, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(33)
fit.ameras.multinomial <- ameras(Y.multinomial~
                                   dose(V1:V10, deg=2, model="EXP")+X1+X2, 
                                 data=data, family="multinomial",
                                 methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                 niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.multinomial)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.multinomial)
```



Without the quadratic term (`deg=1`):

```{r modelfit.multinomial.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(44)
fit.ameras.multinomial.lin <- ameras(Y.multinomial~
                                       dose(V1:V10, deg=1, model="EXP")+X1+X2, 
                                     data=data, family="multinomial",
                                     methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                     niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.multinomial.lin)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.multinomial.lin)
```


# Conditional logistic regression

For conditional logistic regression, the set number variable is specified through a `strata` term in the formula. Again, we first fit models including a quadratic exposure term by setting `deg=2`.
```{r modelfit.clogit, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(3301)
fit.ameras.clogit <- ameras(Y.clogit~dose(V1:V10, deg=2, model="EXP")+X1+X2+
                              strata(setnr), data=data, family="clogit", 
                            methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                            niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.clogit)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.clogit)
```



Without the quadratic term (`deg=1`):

```{r modelfit.clogit.lin, cache=TRUE, eval = identical(Sys.getenv("NOT_CRAN"), "true")}
set.seed(4401)
fit.ameras.clogit.lin <- ameras(Y.clogit~dose(V1:V10, deg=2, model="EXP")+X1+X2+
                                  strata(setnr), data=data, family="clogit", 
                                methods=c("RC", "ERC", "MCML", "FMA", "BMA"), 
                                niter.BMA = 5000, nburnin.BMA = 1000)
```
```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
summary(fit.ameras.clogit.lin)
```

```{r, fig.fullwidth=TRUE, fig.show="hold", out.width='100%', fig.width=6, fig.height=8, eval = identical(Sys.getenv("NOT_CRAN"), "true")} 
coef(fit.ameras.clogit.lin)
```

---
title: "Parameter transformations"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Parameter transformations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r, include = FALSE}
  knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
  )
```
  
```{r setup}
  library(ameras)
```

# Introduction

A transformation can be used to reparametrize parameters internally (i.e., such that the likelihoods are evaluated at `transform(parameters)`, where `parameters` are unconstrained), and should be specified when fitting linear excess relative risk and linear-exponential models to ensure nonnegative odds/risk/hazard. The included function `transform1` applies an exponential transformation to the desired parameters, see below. When supplying a function to `transform`, this should be a function of the full parameter vector, returning a full (transformed) parameter vector. In particular, the full parameter vector contains parameters in the following order (see `?ameras` for the model definitions): $\alpha_0, \mathbf \alpha, \beta_1, \beta_2, \mathbf \beta_{m1}, \mathbf \beta_{m2}, \sigma$, where $\mathbf \alpha$, $\mathbf\beta_{m1}$ and $\mathbf \beta_{m2}$ can be vectors, with lengths matching `X` and `M`, respectively. $\sigma$ is only included for the linear model (Gaussian family), and no intercept is included for the Cox and conditional logistic models. For the multinomial model, the full parameter vector is the concatenation of $Z-1$ parameter vectors in the order as given above, where $Z$ is the number of outcome categories. When no transformation is specified and the linear ERR model is used, `transform1` is used for ERR parameters $\beta_1$ and $\beta_2$ by default, with lower limits $-1/max(D)$ for linear dose-response and $(0,-1/max(D^2))$ for linear-quadratic dose-response, respectively (see below). For the linear-exponential model, a lower limit of 0 is used for $\beta_1$, and no transformation is used for $\beta_2$. If effect modifiers `M` are specified, no transformation is used for those parameters. When negative RRs are obtained during optimization, an error will be generated and a different transformation or bounds should be used. All output is returned in the original parametrization given in `?ameras`. The Jacobian of the transformation (`transform.jacobian`) is required when using a transformation with methods other than BMA. For `transform1`, the Jacobian is given by `transform1.jacobian`. 



# Exponential transformation using transform1

The included function `transform1` applies the exponential transformation $f(\theta_i)=\exp(\theta_i)+LB_i$ to one or multiple components of parameter vector $\mathbf \theta$, where $LB_i$ are lower limits that can be different for each component. In particular, a vector of indices of parameters to be transformed and a vector of corresponding lower bounds LB can be supplied to arguments `index.t` and `lowlimit`, respectively, resulting in transformed parameters $f(\theta_i)=\exp(\theta_i)+\text{LB}_i$.


In particular, `transform1` and `transform1.jacobian` are defined as follows:
```{r}
transform1 <- function(params, index.t=1:length(params), lowlimit=rep(0,length(index.t)), 
                       boundcheck=FALSE, boundtol=1e-3, ...){
  
  if(length(index.t)!=length(lowlimit)) 
    stop("Length mismatch between index.t and lowlimit")
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  params[index.t] <- exp(params[index.t]) + lowlimit
  if(boundcheck){
    if(any(params[index.t]-lowlimit < boundtol)) 
      warning(paste0("WARNING: one or multiple parameter estimates within ", boundtol, " of 
        lower bounds. Try different bounds or starting values."))
  }
  return(params)
}

transform1.jacobian <- function(params, index.t=1:length(params), ...){ 
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  grad <- rep(1, length(params))
  grad[index.t] <- exp(params[index.t])
  if(length(params)>1){
    return(diag(grad))
  } else{
    return(matrix(grad))
  }
}
```

# Defining a custom transformation

If you wish to supply your own transformation, it is helpful to start from the definition of `transform1`. It is also important to keep the following in mind:

1. At a minimum, the function should take arguments `params` and `...`, where `params` is the full parameter vector. Parameters are in a specific order (see above), in case of doubt it is always possible to run `ameras` with the default settings to verify correct the order from the result.
2. Extra arguments can be used and should be supplied when calling `ameras`
3. Optional: if `boundcheck` is an argument to the function, it should be a logical. When transforming parameters after the optimization, the transformation is called with `boundcheck=TRUE`. If `boundcheck` is not an argument of the function, this is ignored.
4. Also define the Jacobian of the transformation, keeping in mind point 1 above.

See the definition of `transform1` above for an example of how to apply the transformation only to specific parameters, and how to use extra arguments.

As an example, suppose instead of the exponential transformation from `transform1`, for the parameters $\beta_1$ and $\beta_2$ we wish to use the sigmoid transformation $f: \mathbb{R} \rightarrow (a,b)$ given by $$f_{a_i,b_i}(\theta_i)= a_i + (b_i-a_i) \frac{1}{1+\exp(-\theta_i)}.$$ Then, using `transform1` as a starting point, we can define the transformation as follows (note that since $\mathbf{a}$ and $\mathbf{b}$ act as bounds, we use an updated bound check):

```{r}
transform.sigmoid <- function(params, index.t=1:length(params), a=rep(0,length(index.t)), 
                              b=rep(1,length(index.t)), boundcheck=FALSE, boundtol=1e-3, ...){
  
  if(length(index.t)!=length(a) | length(index.t) != length(b)) 
    stop("Length mismatch between index.t, a, and b")
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  
  params[index.t] <- a + (b-a) * 1/(1+exp(-1*params[index.t]))
  if(boundcheck){
    if(any( (params[index.t]-a < boundtol) | (b-params[index.t] < boundtol))) 
      warning(paste0("WARNING: one or multiple parameter estimates within ", boundtol, 
      " of bounds. Try different bounds or starting values."))
  }
  return(params)
}
```

Next, noting that $\mathrm{d}f_{a_i,b_i}/d\theta_i = (b_i-a_i)\exp(-\theta_i)/\{1+\exp(-\theta_i) \}^2$, we can define the Jacobian as follows, using `transform1.jacobian` as a starting point:

```{r}
transform.sigmoid.jacobian <- function(params, index.t=1:length(params), 
                                       a=rep(0,length(index.t)), b=rep(1,length(index.t)), ...){ 
  if(length(index.t)!=length(a) | length(index.t) != length(b)) 
    stop("Length mismatch between index.t, a, and b")
  if(any(!(index.t %in% 1:length(params)))) 
    stop("Incorrect indices for transformation specified")
  grad <- rep(1, length(params))
  grad[index.t] <- (b-a)*exp(-1*params[index.t])/(1+exp(-1*params[index.t]))^2
  if(length(params)>1){
    return(diag(grad))
  } else{
    return(matrix(grad))
  }
}
```

Now let us try this transformation on the example data using regression calibration for a linear-quadratic ERR model. Note that all parameters are returned after transformation, and so there should be no difference between using `transform1` and `transform.sigmoid`. First, we fit the model using the sigmoid transformation:

```{r modelfit.sigmoid}
data(data, package="ameras")
fit.ameras.sigmoid <- ameras(Y.binomial~dose(V1:V10, deg=2, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods="RC", transform=transform.sigmoid, 
                            transform.jacobian=transform.sigmoid.jacobian, index.t=4:5)
summary(fit.ameras.sigmoid)
```

Next with default settings, using `transform1`:
```{r modelfit.transform1}
fit.ameras.transform1 <- ameras(Y.binomial~dose(V1:V10, deg=2, model="ERR")+X1+X2, data=data, 
                            family="binomial", methods="RC")
summary(fit.ameras.transform1)
```

As expected, there is no difference between the results.

---
title: "Effect modification"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Effect modification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
)
```

```{r setup}
library(ameras)
data(data, package = "ameras")
```

# Introduction

Effect modification is specified inside `dose()` with the `modifier` argument.
Formula modifiers can be binary numeric/logical variables or factors. Character
variables are not accepted as modifiers; convert them to factors first so the
level order is explicit.

There are two useful ways to code effect modification:

- `modifier = ~ M` uses reference-plus-contrast coding.
- `modifier = ~ 0 + M` or `modifier = ~ M - 1` uses subgroup-specific coding.

These are reparameterizations of the same model. They give the same fitted
likelihood when the same model and data are used, but the coefficients answer
different questions directly.

# Reference-plus-contrast coding

Reference-plus-contrast coding is the default formula-modifier coding. For a
binary modifier `M1`, the coefficient named `dose` is the dose effect in the
reference group, and `dose:M1` is the difference between the non-reference group
and the reference group.

```{r contrast-fit}
fit_contrast <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ M1) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_contrast)
```

This coding is convenient when the effect-modification contrast is the main
quantity of interest.

# Subgroup-specific coding

Subgroup-specific coding reports one dose-effect parameter per subgroup. For
the same binary modifier `M1`, the coefficient names are `dose[M1=0]` and
`dose[M1=1]`.

```{r subgroup-fit}
fit_subgroup <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ 0 + M1) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_subgroup)
```

The two codings are equivalent. The non-reference subgroup effect is the
reference effect plus the contrast:

```{r coding-equivalence}
contrast_coef <- fit_contrast$RC$coefficients
subgroup_coef <- fit_subgroup$RC$coefficients

data.frame(
  term = c("reference group", "non-reference group"),
  from_contrast = c(
    contrast_coef["dose"],
    contrast_coef["dose"] + contrast_coef["dose:M1"]
  ),
  from_subgroup = c(
    subgroup_coef["dose[M1=0]"],
    subgroup_coef["dose[M1=1]"]
  ),
  row.names = NULL
)
```

Subgroup-specific coding is usually the cleaner choice when subgroup-specific
effect estimates and confidence intervals are the main goal. For linear ERR and
linear-exponential models, it can also be more numerically robust because
default bounds are applied directly to each subgroup dose-effect parameter.
With reference-plus-contrast coding, contrast parameters are not bounded by the
default transformation.

# Factor modifiers

Factor modifiers use the existing level order of the factor. The first level is
the reference level.

```{r factor-setup}
data$M_factor <- factor(
  ifelse(data$M1 == 0, "unexposed modifier group", "modifier group"),
  levels = c("unexposed modifier group", "modifier group")
)
```

With `modifier = ~ M_factor`, ameras reports contrast terms for the
non-reference levels.

```{r factor-contrast}
fit_factor_contrast <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ M_factor) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_factor_contrast)
```

With `modifier = ~ 0 + M_factor`, ameras reports subgroup-specific effects for
all levels.

```{r factor-subgroup}
fit_factor_subgroup <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ 0 + M_factor) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_factor_subgroup)
```

Multi-level factors are supported. With reference-plus-contrast coding, ameras
reports one contrast for each non-reference level. With subgroup-specific
coding, ameras reports one dose-effect parameter for each factor level.

```{r three-level-factor}
data$M3 <- factor(
  rep(c("low", "middle", "high"), length.out = nrow(data)),
  levels = c("low", "middle", "high")
)

fit_three_level <- ameras(
  Y.gaussian ~ dose(V1:V10, modifier = ~ 0 + M3) + X1 + X2,
  data = data,
  family = "gaussian",
  methods = "RC"
)

coef(fit_three_level)
```

# Confidence intervals

For subgroup-specific confidence intervals, fit the model using
subgroup-specific coding and then call `confint()` as usual. For `RC`, `ERC`,
and `MCML`, profile likelihood intervals treat subgroup effects as ordinary
coefficient rows.

```{r subgroup-ci, eval = FALSE}
fit_subgroup <- confint(
  fit_subgroup,
  type = "proflik",
  parm = "dose"
)

summary(fit_subgroup)
```

For `FMA` and `BMA`, sample-based confidence or credible intervals are computed
from the sampled coefficients. When subgroup-specific coding is used, BMA
priors and MCMC sampling remain on the internal reference-plus-contrast scale,
but stored samples, summaries, diagnostics, trace plots, and sample-based
intervals are reported on the subgroup-specific scale.

# Choosing a coding

Use reference-plus-contrast coding when the contrast itself is the main
estimand. Use subgroup-specific coding when subgroup effects are the main
estimands, when subgroup-specific profile likelihood intervals are desired, or
when ERR/LINEXP optimization is sensitive to the unbounded contrast
parameterization.

Testing whether effect modification improves the model is a model-comparison
question. Fit a model without the modifier and a model with the modifier, then
compare the fitted likelihoods using the appropriate degrees of freedom.
`dose_lrt()` tests dose-related parameters, but its global dose test is not the
same as a test for effect modification.

---
title: "Standard analyses with one dose realization"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Standard analyses with one dose realization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE
)
```

```{r setup}
library(ameras)
```

# Introduction

The main purpose of `ameras` is to fit association models when multiple realizations of an exposure are available. The same interface can also be used for a standard analysis with a single observed dose variable.

By default, `ameras()` fits regression calibration (`RC`). With a single dose realization, the regression-calibrated dose is just the supplied dose column, so the examples below omit the `methods` argument and use the default `RC` fit.

For non-Gaussian families, `ameras()` defaults to a linear excess relative risk model (`model = "ERR"`). To compare with ordinary `glm()` fits, use the exponential relative risk model (`model = "EXP"`), which puts the dose coefficient on the usual logit or log-link scale.

```{r data}
data(data, package = "ameras")

## Use one exposure realization as the single observed dose.
data$D <- data$V1
head(data[c("D", "Y.gaussian", "Y.binomial", "Y.poisson", "X1", "X2")])
```

This small helper prints matching coefficients from an `ameras` fit and a standard R model. The dose coefficient is named `dose` by `ameras` and `D` by the standard model.

```{r helper}
compare_coefficients <- function(ameras_fit, standard_fit) {
  ameras_terms <- c("(Intercept)", "X1", "X2", "dose")
  standard_terms <- c("(Intercept)", "X1", "X2", "D")

  out <- data.frame(
    term = ameras_terms,
    ameras = unname(ameras_fit$RC$coefficients[ameras_terms]),
    standard = unname(stats::coef(standard_fit)[standard_terms])
  )

  out$ameras <- round(out$ameras, 4)
  out$standard <- round(out$standard, 4)
  out
}
```

# Gaussian outcome

For a Gaussian outcome, a single-dose `ameras()` fit gives the same regression coefficients as `lm()`, with only a small numerical difference. The `ameras` fit also estimates `sigma`, so the comparison below only shows the regression coefficients.

```{r gaussian}
fit_ameras_gaussian <- ameras(Y.gaussian ~ dose(D) + X1 + X2,
                              data = data,
                              family = "gaussian")

fit_lm <- lm(Y.gaussian ~ D + X1 + X2, data = data)

compare_coefficients(fit_ameras_gaussian, fit_lm)
```

The usual `amerasfit` methods are available, including `summary()`, `coef()`, `vcov()`, and `confint()`.

```{r gaussian-summary}
fit_ameras_gaussian <- confint(fit_ameras_gaussian,
                               type = "wald.orig",
                               print = FALSE)
summary(fit_ameras_gaussian)
```

# Binary outcome

For a binary outcome, specify `model = "EXP"` to match the usual logistic regression parameterization used by `glm(..., family = binomial())`.

```{r binomial}
fit_ameras_binomial <- ameras(Y.binomial ~ dose(D, model = "EXP") + X1 + X2,
                              data = data,
                              family = "binomial")

fit_glm_binomial <- glm(Y.binomial ~ D + X1 + X2,
                        data = data,
                        family = binomial())

compare_coefficients(fit_ameras_binomial, fit_glm_binomial)
```

# Count outcome

The same idea applies to Poisson regression. Again, `model = "EXP"` makes the `ameras` dose term directly comparable to the coefficient for `D` in `glm(..., family = poisson())`.

```{r poisson}
fit_ameras_poisson <- ameras(Y.poisson ~ dose(D, model = "EXP") + X1 + X2,
                             data = data,
                             family = "poisson")

fit_glm_poisson <- glm(Y.poisson ~ D + X1 + X2,
                       data = data,
                       family = poisson())

compare_coefficients(fit_ameras_poisson, fit_glm_poisson)
```

# A note on comparison with other software

For a single dose realization, or for `RC` with multiple realizations after replacing the dose by its realization-specific mean, point estimates should generally agree with other software when the same likelihood, dose-response model, covariates, and parameterization are used.

Standard errors and Wald confidence intervals may nevertheless differ. For `RC`, `ERC`, and `MCML`, `ameras` computes variance-covariance matrices from the observed information, using a numerical Hessian of the negative log-likelihood at the optimum. Other software may use expected Fisher information, a scoring-based approximation, a different treatment of parameter constraints, or a different internal parameter scale. These choices can lead to different standard errors even when coefficient estimates agree closely.

# Moving to multiple realizations

Once multiple exposure realizations are available, the formula changes only in the `dose()` term. For example, replacing `D` by `V1:V10` fits the default `RC` analysis using all 10 realizations:

```{r multiple-realizations}
fit_ameras_rc <- ameras(Y.binomial ~ dose(V1:V10, model = "EXP") + X1 + X2,
                        data = data,
                        family = "binomial")
summary(fit_ameras_rc)
```

Other methods can then be added through the `methods` argument:

```{r multiple-methods, eval = FALSE}
fit_ameras_all <- ameras(Y.binomial ~ dose(V1:V10, model = "EXP") + X1 + X2,
                         data = data,
                         family = "binomial",
                         methods = c("RC", "ERC", "MCML", "FMA", "BMA"))
```

# proflik/wald.transformed snapshot: RC

    Code
      fit1[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.69566358   0.03583481   0.27352443 

---

    Code
      fit1[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.08531710   0.20479501   0.07751758 

---

    Code
      fit1[[method]]$CI
    Output
                       lower     upper pval.lower pval.upper iter.lower iter.upper
      dose         0.0000000 0.4819361 0.86103023 0.05005070         NA         11
      dose_squared 0.1248245 0.3885281 0.04978527 0.04932973          7          7

---

    Code
      fit2[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.69566358   0.03583481   0.27352443 

---

    Code
      fit2[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.08531710   0.20479501   0.07751758 

---

    Code
      fit2[[method]]$CI
    Output
                           lower        upper
      (Intercept)  -8.628820e-01   -0.5284451
      dose          4.894522e-07 2623.6136590
      dose_squared  1.559819e-01    0.4746758

# proflik/wald.transformed snapshot: ERC

    Code
      fit1[[method]]$coefficients
    Output
        (Intercept)          dose  dose_squared 
      -6.934349e-01  5.339510e-08  3.150004e-01 

---

    Code
      fit1[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
      0.0525498471 0.0002282538 0.0469447230 

---

    Code
      fit1[[method]]$CI
    Output
                       lower     upper pval.lower pval.upper iter.lower iter.upper
      dose         0.0000000        NA 0.99231556         NA         NA         NA
      dose_squared 0.2316689 0.4201524 0.04930966 0.04946522          6          4

---

    Code
      fit2[[method]]$coefficients
    Output
        (Intercept)          dose  dose_squared 
      -6.934349e-01  5.339510e-08  3.150004e-01 

---

    Code
      fit2[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
      0.0525498471 0.0002282538 0.0469447230 

---

    Code
      fit2[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.7964307 -0.5904391
      dose          0.0000000        Inf
      dose_squared  0.2348924  0.4213916

# proflik/wald.transformed snapshot: MCML

    Code
      fit1[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
      -0.678139870  0.008365397  0.275598584 

---

    Code
      fit1[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.08413263   0.19771678   0.07481774 

---

    Code
      fit1[[method]]$CI
    Output
                       lower     upper pval.lower pval.upper iter.lower iter.upper
      dose         0.0000000 0.4419211 0.96622836 0.05007639         NA         14
      dose_squared 0.1314272 0.3809827 0.04983462 0.04945301          7          7

---

    Code
      fit2[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
      -0.678139870  0.008365397  0.275598584 

---

    Code
      fit2[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.08413263   0.19771678   0.07481774 

---

    Code
      fit2[[method]]$CI
    Output
                           lower         upper
      (Intercept)  -8.430368e-01 -5.132430e-01
      dose          6.371978e-23  1.098244e+18
      dose_squared  1.609797e-01  4.673852e-01

# percentile/hpd snapshot: FMA

    Code
      fit1[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.67942038   0.01295819   0.27406256 

---

    Code
      fit1[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.08348567   0.19530387   0.07436304 

---

    Code
      fit1[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.8447126 -0.5159928
      dose         -0.3644905  0.4016076
      dose_squared  0.1266253  0.4193271

---

    Code
      fit2[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.67942038   0.01295819   0.27406256 

---

    Code
      fit2[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.08348567   0.19530387   0.07436304 

---

    Code
      fit2[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.8477902 -0.5195303
      dose         -0.3662872  0.3992981
      dose_squared  0.1237895  0.4162422

# percentile/hpd snapshot: BMA

    Code
      fit1[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
        -0.7233563    0.1343364    0.2481227 

---

    Code
      fit1[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.10837398   0.24208244   0.06946021 

---

    Code
      fit1[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.9958492 -0.5341520
      dose         -0.2427882  0.6850701
      dose_squared  0.1285954  0.3998694

---

    Code
      fit2[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
        -0.7233563    0.1343364    0.2481227 

---

    Code
      fit2[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.10837398   0.24208244   0.06946021 

---

    Code
      fit2[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.9129171 -0.5102140
      dose         -0.2420483  0.6938121
      dose_squared  0.1054921  0.3744100

# Poisson snapshot: RC

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.88451188   0.62049067  -0.04009246 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
       0.040752584  0.039808460  0.007371927 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      (Intercept)  -0.96438547 -0.80463828
      dose          0.54246753  0.69851382
      dose_squared -0.05454117 -0.02564375

# Poisson snapshot: ERC

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.87990056   0.61124972  -0.03787445 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
       0.040455983  0.038738306  0.007016268 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      (Intercept)  -0.95919283 -0.80060829
      dose          0.53532404  0.68717540
      dose_squared -0.05162609 -0.02412282

# Poisson snapshot: MCML

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.85434971   0.59374567  -0.03682549 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
       0.038539489  0.034503349  0.005806256 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      (Intercept)  -0.92988573 -0.77881370
      dose          0.52612035  0.66137099
      dose_squared -0.04820554 -0.02544543

# Poisson snapshot: FMA

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.85774451   0.59961048  -0.03827769 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
       0.040022104  0.039106177  0.007296265 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      (Intercept)  -0.93763984 -0.78033841
      dose          0.52798149  0.68502270
      dose_squared -0.05628895 -0.02592227

# Poisson snapshot: BMA

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.81604201   0.54829635  -0.03046823 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.08482277   0.10231669   0.01742583 

---

    Code
      fit[[method]]$CI
    Output
                        lower        upper
      (Intercept)  -0.9362879 -0.628417092
      dose          0.3271861  0.703246734
      dose_squared -0.0575862  0.005196177

# clogit snapshot: RC

    Code
      fit[[method]]$coefficients
    Output
              dose dose_squared 
        0.63693308  -0.04695187 

---

    Code
      fit[[method]]$sd
    Output
              dose dose_squared 
        0.09942130   0.02204514 

---

    Code
      fit[[method]]$CI
    Output
                         lower        upper
      dose          0.44207092  0.831795239
      dose_squared -0.09015955 -0.003744198

# clogit snapshot: ERC

    Code
      fit[[method]]$coefficients
    Output
              dose dose_squared 
        0.21121675   0.05299386 

---

    Code
      fit[[method]]$sd
    Output
              dose dose_squared 
       0.035301590  0.008925011 

---

    Code
      fit[[method]]$CI
    Output
                        lower      upper
      dose         0.14202690 0.28040659
      dose_squared 0.03550116 0.07048656

# clogit snapshot: MCML

    Code
      fit[[method]]$coefficients
    Output
              dose dose_squared 
         0.6680322   -0.0567844 

---

    Code
      fit[[method]]$sd
    Output
              dose dose_squared 
        0.09813158   0.02085082 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      dose          0.47569785  0.86036658
      dose_squared -0.09765124 -0.01591755

# clogit snapshot: FMA

    Code
      fit[[method]]$coefficients
    Output
              dose dose_squared 
        0.66550565  -0.05621879 

---

    Code
      fit[[method]]$sd
    Output
              dose dose_squared 
        0.10248403   0.02214376 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      dose          0.46040666  0.86355768
      dose_squared -0.09928857 -0.01171218

# clogit snapshot: BMA

    Code
      fit[[method]]$coefficients
    Output
              dose dose_squared 
        0.62957289  -0.04912569 

---

    Code
      fit[[method]]$sd
    Output
              dose dose_squared 
        0.13710969   0.03042399 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      dose          0.36396710 0.858500724
      dose_squared -0.09865736 0.006771486

# snapshot: binomial_ERR_deg1

    Code
      fit$RC$coefficients
    Output
      (Intercept)          X1        dose 
       -1.1213845   0.4374782   0.8291523 

---

    Code
      fit$RC$sd
    Output
      (Intercept)          X1        dose 
       0.08614808  0.07611521  0.14183554 

---

    Code
      fit$RC$CI
    Output
                       lower      upper
      (Intercept) -1.2902316 -0.9525373
      X1           0.2882951  0.5866612
      dose         0.5511598  1.1071449

# snapshot: binomial_ERR_deg2

    Code
      fit$RC$coefficients
    Output
       (Intercept)           X1         dose dose_squared 
       -0.93045997   0.44257815   0.03556992   0.28382844 

---

    Code
      fit$RC$sd
    Output
       (Intercept)           X1         dose dose_squared 
        0.09586705   0.07655272   0.20887416   0.07941098 

---

    Code
      fit$RC$CI
    Output
                        lower      upper
      (Intercept)  -1.1183559 -0.7425640
      X1            0.2925376  0.5926187
      dose         -0.3738159  0.4449557
      dose_squared  0.1281858  0.4394711

# snapshot: binomial_EXP_deg1

    Code
      fit$RC$coefficients
    Output
      (Intercept)          X1        dose 
       -1.0375175   0.4427830   0.4415451 

---

    Code
      fit$RC$sd
    Output
      (Intercept)          X1        dose 
       0.06993859  0.07650426  0.04080850 

---

    Code
      fit$RC$CI
    Output
                       lower      upper
      (Intercept) -1.1745946 -0.9004404
      X1           0.2928374  0.5927286
      dose         0.3615619  0.5215283

# snapshot: binomial_EXP_deg2

    Code
      fit$RC$coefficients
    Output
       (Intercept)           X1         dose dose_squared 
       -1.00098240   0.44242848   0.36324157   0.02233372 

---

    Code
      fit$RC$sd
    Output
       (Intercept)           X1         dose dose_squared 
        0.08265048   0.07650136   0.10369933   0.02750163 

---

    Code
      fit$RC$CI
    Output
                         lower       upper
      (Intercept)  -1.16297436 -0.83899045
      X1            0.29248856  0.59236839
      dose          0.15999461  0.56648852
      dose_squared -0.03156849  0.07623593

# snapshot: binomial_LINEXP_deg2

    Code
      fit$RC$coefficients
    Output
           (Intercept)               X1      dose_linear dose_exponential 
            -0.9892616        0.4424266        0.3103827        0.3522828 

---

    Code
      fit$RC$sd
    Output
           (Intercept)               X1      dose_linear dose_exponential 
            0.08398917       0.07651524       0.11523109       0.10942094 

---

    Code
      fit$RC$CI
    Output
                             lower      upper
      (Intercept)      -1.15387738 -0.8246459
      X1                0.29245950  0.5923937
      dose_linear       0.08453388  0.5362315
      dose_exponential  0.13782170  0.5667439

# binomial snapshot: ERC

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.74210887   0.29376353   0.04457866 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.07050452   0.10378057   0.02710860 

---

    Code
      fit[[method]]$CI
    Output
                          lower       upper
      (Intercept)  -0.880295186 -0.60392256
      dose          0.090357360  0.49716971
      dose_squared -0.008553222  0.09771054

# binomial snapshot: MCML

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.73270107   0.29100060   0.04072212 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.06982367   0.10339774   0.02762789 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      (Intercept)  -0.86955295 -0.59584918
      dose          0.08834476  0.49365645
      dose_squared -0.01342755  0.09487179

# binomial snapshot: FMA

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.73380690   0.29374272   0.03988671 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.07030784   0.10481320   0.02811012 

---

    Code
      fit[[method]]$CI
    Output
                         lower       upper
      (Intercept)  -0.87199177 -0.59600947
      dose          0.09001549  0.50118564
      dose_squared -0.01635524  0.09461926

# binomial snapshot: BMA

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared 
       -0.73017344   0.28713368   0.04210378 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared 
        0.06152599   0.09014187   0.02427582 

---

    Code
      fit[[method]]$CI
    Output
                          lower       upper
      (Intercept)  -0.857291851 -0.61861015
      dose          0.107033729  0.43912119
      dose_squared  0.004265502  0.08866455

# gaussian snapshot: RC

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared        sigma 
        -0.8596448    0.4123961    0.1923798    1.1007480 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared        sigma 
        0.03544423   0.04824560   0.01132056   0.01421045 

---

    Code
      fit[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.9291143 -0.7901754
      dose          0.3178365  0.5069558
      dose_squared  0.1701919  0.2145677
      sigma         1.0728961  1.1286000

# gaussian snapshot: ERC

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared        sigma 
        -0.8575908    0.4073724    0.1937510    1.0998647 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared        sigma 
        0.03542220   0.04824927   0.01135011   0.01418433 

---

    Code
      fit[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.9270170 -0.7881645
      dose          0.3128056  0.5019393
      dose_squared  0.1715052  0.2159968
      sigma         1.0720639  1.1276655

# gaussian snapshot: MCML

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared        sigma 
        -0.8632862    0.4879678    0.1482892    1.1510453 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared        sigma 
        0.03670882   0.04979688   0.01130102   0.01485994 

---

    Code
      fit[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.9352342 -0.7913382
      dose          0.3903677  0.5855679
      dose_squared  0.1261396  0.1704388
      sigma         1.1219203  1.1801702

# gaussian snapshot: FMA

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared        sigma 
        -0.8637789    0.4907493    0.1469475    1.1511773 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared        sigma 
        0.03702730   0.05193649   0.01291181   0.01488525 

---

    Code
      fit[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.9365178 -0.7910689
      dose          0.3902604  0.5940566
      dose_squared  0.1183053  0.1707397
      sigma         1.1219463  1.1803065

# gaussian snapshot: BMA

    Code
      fit[[method]]$coefficients
    Output
       (Intercept)         dose dose_squared        sigma 
        -0.8607539    0.4838584    0.1478250    1.1509975 

---

    Code
      fit[[method]]$sd
    Output
       (Intercept)         dose dose_squared        sigma 
        0.03752324   0.04523186   0.01112508   0.01398801 

---

    Code
      fit[[method]]$CI
    Output
                        lower      upper
      (Intercept)  -0.9371247 -0.7902548
      dose          0.3954455  0.5666915
      dose_squared  0.1226427  0.1660678
      sigma         1.1228915  1.1756466

\name{ameras}
\alias{ameras}
\title{Analyze multiple exposure realizations}

\description{
Fit regression models accounting for exposure uncertainty using multiple Monte Carlo exposure realizations. Six outcome model families are supported. The first is the Gaussian family for continuous outcomes, \deqn{Y_i \sim N(\mu_i, \sigma^2),} with \eqn{\mu_i = \alpha_0 + \bm X_i^T \bm \alpha +\beta_1 D_i+\beta_2 D_i^2 + \bm M_i^T \bm \beta_{m1}D_i + \bm M_i^T \bm \beta_{m2}D_i^2}. Here \eqn{\bm X_i} are covariates, \eqn{D_i} is the exposure with measurement error, and \eqn{\bm M_i} are binary effect modifiers. The quadratic exposure terms and effect modification are optional.

For non-Gaussian families, three relative risk models for the main exposure are supported, the usual exponential \eqn{RR_i=\exp(\beta_1 D_i+\beta_2 D_i^2+ \bm M_i^T \bm \beta_{m1}D_i + \bm M_i^T \bm \beta_{m2} D_i^2)} and the linear excess relative risk (ERR) model \eqn{RR_i= 1+\beta_1 D_i+\beta_2 D_i^2 + \bm M_i^T \bm \beta_{m1}D_i + \bm M_i^T \bm \beta_{m2}D_i^2}, where the quadratic and effect modification terms are optional. Finally, the linear-exponential relative risk model \eqn{RR_i= 1+(\beta_1 + \bm{M}_i^T \bm{\beta}_{m1}) D_i \exp\{(\beta_2+  \bm{M}_i^T \bm{\beta}_{m2})D_i\}} is supported.

The second supported family is logistic regression for binary outcomes, with probabilities \deqn{p_i/(1-p_i)=RR_i\exp(\alpha_0+\bm X_i^T \bm \alpha).}

Third is Poisson regression for counts, \deqn{Y_i \sim \text{Poisson}(\mu_i),} where \eqn{\mu_i=RR_i \exp(\alpha_0 +\bm X_i^T \bm  \alpha)\times \text{offset}_i} with optional offset.

Fourth is proportional hazards regression for time-to-event data, with hazard function \deqn{h(t) = h_0(t)RR_i\exp(\bm X_i^T \bm \alpha),} with \eqn{h_0} the baseline hazard.

Fifth is multinomial logistic regression for a categorical outcome with \eqn{Z>2} outcome categories, with the last category as the referent category (i.e., \eqn{\alpha_{0,Z}=\bm \alpha_{Z}=\beta_{1,Z}=\beta_{2,Z}=\bm \beta_{m1,Z} = \bm \beta_{m2,Z}=0}): \deqn{P(Y_i=z)=RR_i\exp(\alpha_{0,z}+\bm X_i^T \bm \alpha_{z})/\{1+\sum_{s=1}^{Z-1} RR_i\exp(\alpha_{0,s}+\bm X_i^T \bm \alpha_{s})\}}

Sixth is conditional logistic regression for matched case control data, for which \deqn{P\left(Y_i = 1, Y_k = 0 \forall k \neq i \bigg| \sum_{i \in \mathcal{R}} Y_i = 1\right) = RR_i\exp(\bm X_i^T \bm \alpha)/\{\sum_{k \in \mathcal{R}}RR_k\exp(\bm X_k^T \bm \alpha)\},} where \eqn{\mathcal{R}} is the matched set corresponding to individual \eqn{i}.

Methods include regression calibration (Carroll et al. 2006 \doi{10.1201/9781420010138}), extended regression calibration (Little et al. 2023 \doi{10.1038/s41598-023-42283-y}), Monte Carlo maximum likelihood (Stayner et al. 2007 \doi{10.1667/RR0677.1}), frequentist model averaging (Kwon et al. 2023 \doi{10.1371/journal.pone.0290498}), and Bayesian model averaging (Kwon et al. 2016 \doi{10.1002/sim.6635}).

}

\usage{
ameras(formula=NULL, data, family="gaussian", methods="RC", 
  Y=NULL, dosevars=NULL, doseRRmod=NULL, deg=NULL,
  M=NULL, X=NULL, offset=NULL, entry=NULL, exit=NULL,
  setnr=NULL,
  CI=NULL, params.profCI=NULL,
  maxit.profCI=NULL, tol.profCI=NULL,
  transform=NULL,
  transform.jacobian=NULL, inpar=NULL, loglim=1e-30, MFMA=100000, 
  prophaz.numints.BMA=10, ERRprior.BMA="doubleexponential", nburnin.BMA=5000, 
  niter.BMA=20000, nchains.BMA=2, thin.BMA=10, included.replicates.BMA=NULL, 
  optim.method="Nelder-Mead", control=NULL, keep.data=TRUE, ... )
}
\arguments{
\item{formula}{an object of class \code{"formula"} containing the model specification. See Details.}
\item{data}{input data frame.}
\item{family}{outcome model family: \code{"gaussian"}, \code{"binomial"}, \code{"poisson"}, \code{"prophaz"}, \code{"multinomial"} or \code{"clogit"} (default \code{"gaussian"}).}
\item{methods}{character vector of one or multiple methods to apply. Options: \code{"RC"}, \code{"ERC"}, \code{"MCML"}, \code{"FMA"}, \code{"BMA"} (default \code{"RC"}).}
\item{Y}{\strong{Deprecated}. Use the formula interface instead. Name or column index of the outcome variable for linear, binomial, Poisson, multinomial and conditional logistic models, or event indicator variable for the proportional hazards model.}
\item{dosevars}{\strong{Deprecated}. Use the formula interface instead. Names or column indices of exposure replicate vectors.}
\item{doseRRmod}{\strong{Deprecated}. Use the formula interface instead. The functional form of the dose-response relationship; options are exponential RR (\code{"EXP"}), linear ERR (\code{"ERR"}), or linear-exponential RR (\code{"LINEXP"}) (default \code{"ERR"}).}
\item{deg}{\strong{Deprecated}. Use the formula interface instead. For \code{doseRRmod="ERR"} and \code{doseRRmod="EXP"}, whether to fit a linear (\code{deg=1}) or linear-quadratic (\code{deg=2}) dose-response model (default linear).}
\item{M}{\strong{Deprecated}. Use the formula interface instead. Names or column indices of binary effect modifying variables (optional).}
\item{X}{\strong{Deprecated}. Use the formula interface instead. Names or column indices of other covariates (optional).}
\item{offset}{\strong{Deprecated}. Use the formula interface instead. Name or column index of offset variable for Poisson regression (optional).}
\item{entry}{\strong{Deprecated}. Use the formula interface instead. Name or column index of left truncation time variable for proportional hazards regression (optional).}
\item{exit}{\strong{Deprecated}. Use the formula interface instead. Name or column index of exit time variable, required when \code{family=prophaz}.}
\item{setnr}{\strong{Deprecated}. Use the formula interface instead. Name or column index of integer-valued matched set variable, required when \code{family="clogit"}.}
\item{CI}{\strong{Deprecated}. Use confint() to compute confidence intervals instead. Method for calculation of 95\% confidence or credible intervals (see Details).
For RC, ERC, and MCML, options are \code{"wald.orig"}, \code{"wald.transformed"}, \code{"proflik"} (default \code{"proflik"}).
For FMA and BMA, options are \code{"percentile"} and \code{"hpd"} (default \code{"percentile"}).
If \code{methods} contains at least one of RC, ERC, and MCML and at least one of FMA and BMA, \code{CI} must be length 2 and specify one method for RC, ERC, and MCML, and one for FMA and BMA (see Details).}
\item{params.profCI}{\strong{Deprecated}. Use confint() to compute confidence intervals instead. When \code{CI="proflik"}, whether to obtain profile-likelihood CIs for all parameters (\code{"all"}) or only dose-related parameters (\code{"dose"}, default).}
\item{maxit.profCI}{\strong{Deprecated}. Use confint() to compute confidence intervals instead. Maximum iterations for determining profile-likelihood CIs; passed to \code{uniroot} (default 20).}
\item{tol.profCI}{\strong{Deprecated}. Use confint() to compute confidence intervals instead. Tolerance for determining profile-likelihood CIs; passed to \code{uniroot} (default \code{1e-2}).}
\item{transform}{function for internal parameter transformation (see Details).}
\item{transform.jacobian}{Jacobian of the transformation function (see Details).}
\item{inpar}{vector of initial values for log-likelihood optimization (optional).}
\item{loglim}{parameter used in likelihood computations to avoid taking the log of very small or negative numbers via \code{log(max(x, loglim))} (default \code{1e-30}).}
\item{MFMA}{number of samples for \code{"FMA"} to compute estimates and CIs (default 100,000).}
\item{prophaz.numints.BMA}{for \code{methods="BMA"} with \code{family="prophaz"}, the number of subintervals with constant baseline hazard (default 10). Cut points are determined based on quantiles of the event time distribution among cases.}
\item{ERRprior.BMA}{prior for dose-related parameters when \code{doseRRmod="ERR"} or \code{"LINEXP"} and \code{methods="BMA"}.
Options: \code{"truncated_normal"}, \code{"truncated_horseshoe"}, \code{"truncated_doubleexponential"}, \code{"normal"}, \code{"horseshoe"}, \code{"doubleexponential"}, see Details (default \code{"doubleexponential"}).}
\item{nburnin.BMA}{number of MCMC burn-in iterations for BMA (default 1,000).}
\item{niter.BMA}{number of MCMC iterations per chain for BMA (default 5,000).}
\item{nchains.BMA}{number of MCMC chains for BMA (default 2).}
\item{thin.BMA}{thinning rate for BMA (default 10).}
\item{included.replicates.BMA}{indices of exposure replicates used in BMA (defaults to all replicates).}
\item{optim.method}{method used for optimization by \code{optim}. Options are \code{"Nelder-Mead"} and \code{"BFGS"}. When using Nelder-Mead, a second optimization with BFGS is run to ensure an optimal fit.}
\item{control}{control list passed to \code{optim} (default \code{list(reltol=1e-10)}).}
\item{keep.data}{whether to attach data to the output object (default \code{TRUE}). When the data object is large, \code{keep.data} can be set to \code{FALSE} to preserve memory. The attached data is used to compute profile likelihood confidence intervals, but can also be supplied seperately when \code{keep.data=FALSE}. See \code{\link[=confint.amerasfit]{confint}}.}
\item{\dots}{other arguments, passed to functions such as \code{transform}.}
}
\value{
The output is an object of class \code{amerasfit}. General components are \code{call} (the function call to \code{ameras}), \code{formula} (the formula object specifying the model), \code{num.rows} (the number of rows in \code{data}), \code{num.replicates} (the number of dose replicates provided), \code{transform} (the used transformation function, if applicable), \code{transform.jacobian} (the used Jacobian function for the transformation, if applicable), \code{other.args} (any other arguments passed to \dots), \code{model} (a list containing the specified model components parsed from the formula), \code{CI.computed} (logical, whether confidence intervals have been attached by \code{\link[=confint.amerasfit]{confint}}), and \code{data} (either the data frame used for model fitting when \code{keep.data=TRUE} or \code{NULL} otherwise).

For each method supplied to \code{methods}, the output contains a list with components:
  \item{coefficients}{named vector of model coefficients.}
  \item{sd}{named vector of standard deviations.}
  \item{runtime}{string with the runtime in seconds.}
  
For RC, ERC, and MCML  the following additional output is included:
  \item{vcov}{covariance matrix for the full parameter vector.}
  \item{optim}{a list object with results returned by optim. Components are \code{par} (raw parameters before applying a transformation if applicable), \code{hessian} (Hessian matrix for \code{par}), \code{convergence} (convergence code with 0 indicating convergence and 1 indicating that the maximal number of iterations was reached), and \code{counts} (the number of likelihood function evaluations used during optimization).}
  \item{loglik}{log-likelihood value at the optimum.}

For RC and ERC, the output additionally contains:
  \item{ERC}{logical, whether the output is for ERC (\code{ERC=TRUE}) or RC (\code{ERC=FALSE}).}
  
For BMA  the output additionally contains:
  \item{samples}{MCMC posterior samples, as obtained from \code{nimble}. This is a list object with \code{nchains.BMA} components, each a named matrix with the samples from one chain in its rows, with columns corresponding to model parameters.} 
  \item{Rhat}{data frame with two columns, \code{Rhat} and \code{n.eff}. The first  column contains the Gelman-Rubin statistics \eqn{\hat R \geq 1} that can be used to assess convergence of MCMC chains. A value of 1 indicates good convergence and values \eqn{>1.05} indicate poor convergence. The effective sample size \code{n.eff} is a measure of how many independent samples the auto-correlated MCMC samples correspond to. A low effective sample size indicates high correlations and/or poor mixing.}
  \item{included.replicates}{indices of replicate exposures that were included to obtain the results.}
  \item{prophaz.timepoints}{for \code{family="prophaz"}, time points defining the intervals on which the estimated baseline hazards is constant; these are \code{prophaz.numints.BMA + 1} time points covering the interval \code{(min(entry), max(exit))}, based on quantiles among observed event times. See Details.}

Finally, for FMA the output additionally contains:
  \item{samples}{the samples generated from the normal distributions associated with each dose replicate.}
  \item{included.samples}{the total number of samples included.}
  \item{included.replicates}{indices of replicate exposures that were included to obtain results. Fits without a valid variance estimate (i.e., non-invertible Hessian or inverse that is not positive definite) or that reach the maximal number of iterations without convergence are filtered out and not used to obtain results. }

The class \code{amerasfit} supports the methods \code{\link[=print.amerasfit]{print}}, \code{\link[=coef.amerasfit]{coef}}, \code{\link[=confint.amerasfit]{confint}}, \code{\link[=summary.amerasfit]{summary}}, and \code{\link{traceplot}}.
}

\details{
Models are specified through formulas of the form \code{Y~dose(dose_expression, model="ERR", deg=1, modifier=M1+M2)+X1+X2}. Here \code{dose_expression} specifies the dose realization columns and is parsed by \code{eval_select} from the \pkg{tidyselect} package. Useful examples are \code{D1:D1000} if the doses are in a sequence of columns with sequential names such as \code{D1}-\code{D1000}, and \code{all_of(dosevars)} where \code{dosevars} is a vector with the names of all dose columns. Further, \code{model} specifies, for non-Gaussian families, whether to use the exponential dose-response model (\code{model="EXP"}), the linear-exponential model (\code{model="LINEXP"}) or the linear ERR model (\code{model="ERR"}). Next, \code{deg} is used to specify whether a quadratic dose term should (\code{deg=2}) or should not (\code{deg=1}) be estimated for the exponential or linear ERR dose-response model. The \code{modifier} term is optional and used to specify binary effect modification variables. Note that interactions in the modifier term are not allowed, e.g. \code{M1*M2}. When \code{deg}, \code{modifier}, and \code{model} are not supplied, the defaults are \code{deg=1}, no effect modifiers, and \code{model="ERR"}. Finally, \code{X1} and \code{X2} above represent optional additional covariates, which can include factor variables and interactions such as \code{X1*X2}. The matched set variable \code{setnr} required for conditional logistic regression is specified on the right-hand side of the formula through a term \code{strata(setnr)}, and an optional offset variable \code{offset} for Poisson regression similarly through a term \code{offset(offset)}. For proportional hazards regression, the left-hand side of the formula should have the form \code{Surv(exit, status)} or \code{Surv(entry, exit, status)}.

A transformation can be used to reparametrize parameters internally (i.e., such that the likelihoods are evaluated at \code{transform(parameters)}, where \code{parameters} are unconstrained), and should be specified when fitting linear excess relative risk and linear-exponential models to ensure nonnegative odds/risk/hazard. The included function \code{\link{transform1}} applies an exponential transformation to the desired parameters, see \code{?transform1}. When supplying a function to \code{transform}, this should be a function of the full parameter vector, returning a full (transformed) parameter vector. In particular, the full parameter vector contains parameters in the following order: \eqn{\alpha_0, \bm \alpha, \beta_1, \beta_2, \bm \beta_{m1}, \bm \beta_{m2}, \sigma}, where \eqn{\bm \alpha}, \eqn{\bm\beta_{m1}} and \eqn{\bm \beta_{m2}} can be vectors, with lengths matching \eqn{\bm X} and \eqn{\bm M}, respectively. \eqn{\sigma} is only included for the linear model (Gaussian family), and no intercept is included for the proportional hazards and conditional logistic models. For the multinomial model, the full parameter vector is the concatenation of \eqn{Z-1} parameter vectors in the order as given above, where \eqn{Z} is the number of outcome categories, with the last category chosen as the referent category. See \code{vignette("transformations", package="ameras")} for an example of how to specify a custom transformation function. 

When no transformation is specified and the linear ERR model is used, \code{transform1} is used for ERR parameters \eqn{\beta_1} and \eqn{\beta_2} by default, with lower limits \eqn{-1/max(D)} for \eqn{\beta_1} in the linear dose-response and \eqn{(0,-1/max(D^2))} for \eqn{(\beta_1,\beta_2)} in the linear-quadratic dose-response, respectively. For the linear-exponential model, a lower limit of 0 is used for \eqn{\beta_1}, and no transformation is used for \eqn{\beta_2}. If effect modifiers \code{M} are specified, no transformation is used for those parameters. When negative RRs are obtained during optimization, an error will be generated and a different transformation or bounds should be used. All output is returned in the original parametrization. The Jacobian of the transformation (\code{transform.jacobian}) is required when using a transformation. For \code{\link{transform1}}, the Jacobian is given by \code{\link{transform1.jacobian}}. No transformations are used in BMA, and FMA is applied on the parameters using the parametrization as given in above with variances obtained using the delta method with the provided Jacobian function.  

For BMA, a prior distribution for exposure-response parameters can be chosen when using linear or linear-exponential exposure-response model. The options are normal, horshoe, and double exponential priors, and the same priors truncated at 0 to yield positive values. In particular:

\itemize{
  \item Normal: \eqn{\beta_j \sim N(0,1000)} for all exposure-response parameters \eqn{\beta_j}
  \item Horseshoe (shrinkage prior): \eqn{\tau \sim \text{Cauchy}(0,1)^+; \lambda_j \sim \text{Cauchy}(0,1)^+; \beta_j \sim N(0, \tau^2 \lambda_j^2)}. Here \eqn{\tau} is shared across all parameters
  \item Double exponential (shrinkage prior): \eqn{\lambda_j \sim \text{Cauchy}(0,1)^+; \beta_j \sim \text{DoubleExponential}(0,\lambda_j)}
}
For all other parameters, and when using the exponential exposure-response model or the Gaussian outcome family, the prior is \eqn{N(0, 1000)}. For the parameter \eqn{\sigma} in the Gaussian family, this prior is truncated at 0. 

Because the proportional hazards model is not available in \code{nimble}, \code{ameras} uses a piecewise constant baseline hazard for Bayesian model averaging. The interval \code{min(entry), max(exit))} is divided into \code{prophaz.numints.BMA} subintervals with cutpoints obtained as quantiles of the distribution of event times among cases, and a baseline hazard parameter is estimated for each subinterval.
}

\seealso{
  \code{\link[=confint.amerasfit]{confint}} for computing confidence intervals,
  \code{\link[=summary.amerasfit]{summary}} for a summary of the fitted model
  including confidence intervals if computed, \code{\link[=coef.amerasfit]{coef}} 
  for extracting coefficients.
}

\examples{
\donttest{
  data(data, package="ameras")
  ameras(Y.gaussian~dose(V1:V10, modifier=M1+M2)+X1+X2, data=data, family="gaussian") 
}
}

\references{
  Roberti, S., Kwon D., Wheeler W., Pfeiffer R. (in preparation).
  ameras: An R Package to Analyze Multiple Exposure Realizations in Association Studies
}

\name{confint.amerasfit}
\alias{confint.amerasfit}
\title{Confidence intervals for an amerasfit object}

\description{
  Computes confidence intervals for the parameters of a fitted
  \code{amerasfit} object. This is a separate step from model fitting,
  i.e., \code{\link{ameras}} fits the model and \code{confint} computes 
  intervals and attaches them to the fitted object.
}

\usage{
\method{confint}{amerasfit}(object, parm="dose", level=0.95, 
        type=c("proflik","percentile"), maxit.profCI=20, 
        tol.profCI=1e-2, data=NULL, \dots)
}

\arguments{
  \item{object}{
    A fitted model object of class \code{amerasfit}, as returned by
    \code{\link{ameras}}.
  }
  \item{parm}{
    Either \code{"dose"} to compute intervals for dose-related parameters
    only, \code{"all"} for all parameters, or a character vector of specific
    parameter names. Only used when \code{type = "proflik"} since Wald
    intervals are cheap to compute for all parameters simultaneously.
    Defaults to \code{"dose"} since profile likelihood computation can be
    extensive.
  }
  \item{level}{
    The confidence level (default \code{0.95}).
  }
  \item{type}{
    The type(s) of confidence intervals to determine. For
    RC, ERC, and MCML, this can be one of:
    \describe{
      \item{\code{"wald.orig"}}{
        Wald intervals on the original parameter scale using the delta
        method variance-covariance matrix.
      }
      \item{\code{"wald.transformed"}}{
        Wald intervals computed on the transformed (reparametrized) scale
        and then back-transformed. Only available when a transformation
        was used during fitting.
      }
      \item{\code{"proflik"}}{
        Profile likelihood intervals based on the chi-squared
        approximation. More accurate than Wald intervals but
        computationally intensive, especially for large datasets or
        complex models.
      }
    }
    For FMA and BMA, confidence intervals are based on the generated samples and
    possible confidence interval types are:
    \describe{
      \item{\code{"percentile"}}{
        Equal-tailed percentile intervals.
      }
      \item{\code{"hpd"}}{
        Highest posterior density intervals via
        \code{\link[coda]{HPDinterval}} from the \pkg{coda} package.
      }
    }
    If \code{object} contains results for at least one of RC, ERC, and MCML and at least one of FMA and BMA, \code{type} must be length 2 and specify one method for RC, ERC, and MCML, and one for FMA and BMA.
  }
  \item{data}{
    The original data frame used for fitting. Only required when
    \code{type = "proflik"} and the model was fitted with
    \code{keep.data = FALSE}
  }
  \item{maxit.profCI}{
    Maximum number of iterations for the root-finding algorithm used to
    locate profile likelihood interval bounds. Only used when
    \code{type = "proflik"}. Defaults to \code{20}.
  }
  \item{tol.profCI}{
    Tolerance for the root-finding algorithm. Only used when
    \code{type = "proflik"}. Defaults to \code{1e-2}. Reduce for more
    precise bounds at the cost of additional computation.
  }
  \item{\dots}{
    Additional arguments, currently unused.
  }
}

\value{
  The original \code{amerasfit} object with a \code{CI} element added to
  each fitted method result. For RC, ERC, and MCML the \code{CI} element
  is a data frame with columns:
  \describe{
    \item{\code{lower}}{Lower confidence bound.}
    \item{\code{upper}}{Upper confidence bound.}
  }
  When \code{type = "proflik"}, four additional columns are included:
  \describe{
    \item{\code{pval.lower}}{
      P-value at the lower bound, should be close to \eqn{1 -}
      \code{level}.
    }
    \item{\code{pval.upper}}{
      P-value at the upper bound, should be close to \eqn{1 -}
      \code{level}.
    }
    \item{\code{iter.lower}}{
      Number of iterations used by the root-finding algorithm for the
      lower bound.
    }
    \item{\code{iter.upper}}{
      Number of iterations used by the root-finding algorithm for the
      upper bound.
    }
  }
  For FMA and BMA the \code{CI} element is a data frame with columns
  \code{lower} and \code{upper}.
}

\details{

  For (extended) regression calibration and Monte Carlo maximum likelihood, 
  Wald and profile likelihood intervals can be obtained. When a parameter 
  transformation \eqn{\bm\theta = h(\bm\eta)} is used, \code{type="wald.transformed"} 
  yields the CI at significance level \eqn{\alpha} of \eqn{h(\bm\eta \pm z_{1-\alpha/2} \bm V)} 
  where \eqn{z_{1-\alpha/2}} is the \eqn{1-\alpha/2}-quantile of the standard normal distribution and \eqn{\bm V} is the vector 
  of standard deviations estimated using the inverse Hessian matrix, 
  and \code{type="wald.orig"} uses the delta method to obtain the CI 
  \eqn{h(\bm\eta)\pm z_{1-\alpha/2} \bm V_*} where \eqn{\bm V_*} is the vector of 
  standard deviations estimated using \eqn{J H^{-1} J^T} with \eqn{J} the 
  Jacobian of the transformation and \eqn{H} is the Hessian. When no 
  transformation is used, \code{type="wald.orig"} should be used. 
  The third option is \code{proflik}, which uses the profile likelihood to 
  compute confidence bounds. 
  
  For FMA and BMA, the options for 
  confidence/credible intervals are \code{type="percentile"} which uses 
  percentiles, and \code{type="hpd"} which computes highest posterior density 
  intervals using \code{HPDinterval} from the \code{coda} package, both using 
  the FMA samples or Bayesian posterior samples.
  
  Profile likelihood intervals (\code{type="proflik"}) require
  re-evaluating the likelihood repeatedly and can be time-consuming.
  The \code{parm} argument can be used to restrict computation to dose
  parameters only (the default) when intervals for the other parameters are
  not of interest.

  When the model was fitted with \code{keep.data=FALSE} and
  \code{type="proflik"} is used for \code{confint}, the original data must be
  supplied via the \code{data} argument. Wald intervals do not require the
  data and can always be computed from the stored Hessian and parameter
  estimates alone.
}

\seealso{
  \code{\link{ameras}} for model fitting,
  \code{\link[=summary.amerasfit]{summary}} for a summary of the fitted model
  including confidence intervals if computed,
  \code{\link[stats]{confint}} for the generic function.
}

\examples{
data("data", package = "ameras")

## Fit the model
fit <- ameras(Y.binomial~dose(V1:V10, model="ERR"), data = data, family = "binomial",
               methods = "RC")

## Wald intervals (fast)
fit <- confint(fit, type = "wald.orig")
summary(fit)

## Profile likelihood intervals for dose parameters only (slower)
\donttest{
fit <- confint(fit, type = "proflik", parm = "dose")
summary(fit)
}

## With keep.data = FALSE, supply data explicitly for proflik
\donttest{
fit2 <- ameras(Y.binomial~dose(V1:V10, model="ERR"), data = data, family = "binomial",
               methods = "RC", keep.data = FALSE)
fit2 <- confint(fit2, type = "proflik", data = data)
}

## FMA and BMA with percentile intervals
\donttest{
fit3 <- ameras(Y.binomial~dose(V1:V10, model="ERR"), data = data, family = "binomial", 
               methods = c("FMA", "BMA"))
fit3 <- confint(fit3, type = "percentile")
summary(fit3)
}
}

\name{traceplot}
\alias{traceplot}
\alias{traceplot.amerasfit}
\title{Traceplots for MCMC samples}
\description{
Produce MCMC traceplots for \code{amerasfit} objects.
}
\usage{
traceplot(object, ...)

\method{traceplot}{amerasfit}(object, iter = 5000, Rhat = TRUE, n.eff = TRUE, pdf = FALSE, ...)
}
\arguments{
  \item{object}{a \code{amerasfit} object containing BMA output to be plotted}
  \item{iter}{number of iterations to include in the traceplot (defaults to last 5000)}
  \item{Rhat}{logical; whether to include R-hat diagnostics in the plot (default TRUE)}
  \item{n.eff}{logical; whether to include effective sample size in the plot (default TRUE)}
  \item{pdf}{logical; whether to save the output as a PDF (default FALSE)}
  \item{...}{additional arguments passed to \code{MCMCtrace}}
}
\details{
Wrapper for \code{MCMCvis::MCMCtrace} to produce MCMC diagnostic plots. See \code{?MCMCtrace} for more plotting options that can be provided through \code{...}.
}
\value{
Traceplots and posterior density plots.
}
\seealso{
\code{\link[MCMCvis]{MCMCtrace}}
}
\examples{
\donttest{
   data(data, package="ameras")
   fit <- ameras(Y.gaussian~dose(V1:V10), data, methods="BMA")
   traceplot(fit)
}
}

\name{summary.amerasfit}
\alias{summary.amerasfit}
\alias{print.summary.amerasfit}
\title{Summarize an amerasfit object}

\description{
  Produces a summary of a fitted \code{amerasfit} object, including
  parameter estimates, standard errors, and confidence intervals if
  computed via \code{\link[=confint.amerasfit]{confint}}.
}

\usage{
\method{summary}{amerasfit}(object, \dots)

\method{print}{summary.amerasfit}(x, digits = max(3, getOption("digits") - 3), \dots)
}

\arguments{
  \item{object}{
    A fitted model object of class \code{amerasfit}, as returned by
    \code{\link{ameras}}.
  }
  \item{x}{
    An object of class \code{summary.amerasfit}, as returned by
    \code{summary.amerasfit}.
  }
  \item{digits}{
    The number of significant digits to use. Defaults to
    \code{max(3, getOption("digits") - 3)}.
  }
  \item{\dots}{
    Additional arguments, currently unused.
  }
}

\value{
  \code{summary.amerasfit} returns an object of class
  \code{summary.amerasfit}, which is a list containing the following
  elements:
  \describe{
    \item{\code{call}}{
      The matched call from the original \code{\link{ameras}} invocation.
    }
    \item{\code{summary_table}}{
      A data frame with one row per parameter per method, containing
      columns:
      \describe{
        \item{\code{Method}}{The estimation method (RC, ERC, MCML, FMA, or BMA).}
        \item{\code{Term}}{The parameter name.}
        \item{\code{Estimate}}{The parameter estimate.}
        \item{\code{SE}}{The standard error.}
        \item{\code{CI.lower}}{
          The lower confidence bound, if confidence intervals have been
          computed via \code{\link[=confint.amerasfit]{confint}}.
        }
        \item{\code{CI.upper}}{
          The upper confidence bound, if confidence intervals have been
          computed via \code{\link[=confint.amerasfit]{confint}}.
        }
        \item{\code{pval.lower}}{
          p-value associated with the lower bound of the profile likelihood 
          CI, the assess validity of the obtained bound. Only included if
          profile likelihood confidence intervals were computed via
          \code{\link[=confint.amerasfit]{confint}}.
        }
        \item{\code{pval.upper}}{
          p-value associated with the upper bound of the profile likelihood 
          CI, the assess validity of the obtained bound. Only included if
          profile likelihood confidence intervals were computed via
          \code{\link[=confint.amerasfit]{confint}}.
        }
        \item{\code{Rhat}}{
          The Gelman-Rubin convergence diagnostic, included only when
          BMA results are present. Values above 1.05 indicate potential convergence problems.
        }
        \item{\code{n.eff}}{
          The effective sample size, included only when BMA results are present.
        }
      }
    }
    \item{\code{runtime_table}}{
      A data frame with columns \code{Method} and \code{Runtime},
      reporting the computation time in seconds for each method.
    }
    \item{\code{total_runtime_seconds}}{
      The total computation time in seconds across all methods.
    }
    \item{\code{CI.computed}}{
      Logical. \code{TRUE} if confidence intervals have been computed
      via \code{\link[=confint.amerasfit]{confint}}, \code{FALSE}
      otherwise.
    }
  }
}

\details{
  \code{summary.amerasfit} collects results from all estimation methods
  present in the fitted object into a single summary table. Columns for
  confidence intervals are only printed if they have been computed by
  \code{\link[=confint.amerasfit]{confint}}. When BMA results are present 
  in the fitted object, the summary table includes columns \code{Rhat} and 
  \code{n.eff}, with \code{NA} values for all other methods.
}

\seealso{
  \code{\link{ameras}} for model fitting,
  \code{\link[=confint.amerasfit]{confint}} for computing confidence
  intervals,
  \code{\link[=print.amerasfit]{print}} for a shorter printed summary,
  \code{\link[=coef.amerasfit]{coef}} for extracting coefficients.
}

\examples{
data("data", package = "ameras")

## Fit the model
fit <- ameras(Y.binomial~dose(V1:V10, model="ERR"), data = data, family = "binomial",
              methods = "RC")

## Summary without confidence intervals
summary(fit)

## Summary with confidence intervals
fit <- confint(fit, method = "wald.orig")
summary(fit)

## Access the summary table directly
s <- summary(fit)
s$summary_table

## Multiple methods
\dontrun{
fit2 <- ameras(Y.binomial~dose(V1:V10, model="ERR"), data = data, family = "binomial",
              methods = c("RC", "ERC", "MCML"))
fit2 <- confint(fit2, method = "wald.orig")
summary(fit2)
}
}

\name{data}
\alias{data}

\title{
  Example data
}
\description{
 Data includes outcomes of all six supported types in the appropriately named columns. For proportional hazards regression, the observed exit time is \code{time} and event status is \code{event}. For conditional logistic regression, the matched set variable is \code{setnr}. The data has 10 exposure replicates in columns \code{V1}-\code{V10}.
}
%\details{}
\examples{

 data(data, package="ameras")

 # Display a few rows of the data
 data[1:5, ]

}
\keyword{data}
%\seealso{  }

\name{transform1.jacobian}
\alias{transform1.jacobian}
\title{Jacobian of the exponential parameter transformation}

\description{
Computes the Jacobian matrix of \link{transform1}. Note that lower limits do not need to be specified as the Jacobian is independent of those
}

\usage{
transform1.jacobian(params, index.t=1:length(params), ... )
}
\arguments{
  \item{params}{input parameter vector (before transformation) to evaluate the Jacobian at}
  \item{index.t}{indices of parameters to be transformed (default all)}
  \item{...}{not used}
}
\value{
Jacobian matrix.
}



\examples{
params <- c(.1, .5, 1)
transform1.jacobian(params)
}

\name{coef.amerasfit}
\alias{coef.amerasfit}
\title{Estimated coefficients for an amerasfit object}

\description{
  Returns a data frame with all the parameters of a fitted
  \code{amerasfit} object. The resulting object has a column
  for every method supplied to `methods` when calling `ameras`,
  with rows corresponding to parameters.
}

\usage{
\method{coef}{amerasfit}(object, \dots)
}

\arguments{
  \item{object}{
    A fitted model object of class \code{amerasfit}, as returned by
    \code{\link{ameras}}.
  }
  \item{\dots}{
    Additional arguments, currently unused.
  }
}

\value{
  Data frame with estimated model parameters. Column names correspond to the
  `methods` used in the `ameras` call, and row names correspond to parameter
  names.
}



\seealso{
  \code{\link{ameras}} for model fitting,
  \code{\link[=summary.amerasfit]{summary}} for a summary of the fitted model
  including confidence intervals if computed.
}

\examples{
\donttest{
data("data", package = "ameras")
dosevars <- paste0("V", 1:10)

## Fit the model
fit <- ameras(Y.binomial~dose(V1:V10, model="ERR"), data = data, family = "binomial", 
methods = c("RC","ERC"))
## Full matrix
coef(fit)

## Vector with RC parameters
coef(fit)$RC
}
}

\title{Inverse of exponential parameter transformation}
\alias{transform1.inv}
\name{transform1.inv}

\description{
Inverse of \code{transform1} for the purpose of deriving initial values.
}

\usage{
transform1.inv(params, index.t=1:length(params), lowlimit=rep(0,length(index.t)), ... )
}
\arguments{
  \item{params}{full input parameter vector}
  \item{index.t}{indices of parameters to be transformed (default all)}
  \item{lowlimit}{lower limits to be applied (default zero), where the k-th component of \code{lowlimit} is applied to the k-th index in \code{index.t} }
  \item{...}{not used}
}
\value{
Transformed parameter vector.
}



\examples{
params <- c(.1, .5, 1) # Desired initial values on original scale
transform1.inv(params, lowlimit=c(0, -1, 1)) # Initial values to use on transformed scale
}

\name{ameras-package}
\alias{ameras-package}
\docType{package}
\title{Analyze multiple exposure realizations in association studies}
\description{Analyze association studies with multiple realizations of a noisy or uncertain exposure. These can be obtained from e.g. a two-dimensional Monte Carlo dosimetry system (Simon et al 2015 <doi:10.1667/RR13729.1>) to characterize exposure uncertainty.
    Methods include regression calibration (Carroll et al. 2006 \doi{10.1201/9781420010138}), extended 
    regression calibration (Little et al. 2023 \doi{10.1038/s41598-023-42283-y}), Monte Carlo maximum 
    likelihood (Stayner et al. 2007 \doi{10.1667/RR0677.1}), frequentist model averaging (Kwon et al. 2023 \doi{10.1371/journal.pone.0290498}), 
    and Bayesian model averaging (Kwon et al. 2016 \doi{10.1002/sim.6635}). Supported model families are
    Gaussian, binomial, multinomial, Poisson, proportional hazards, and conditional logistic.}
\details{
  The function used to fit models is \code{\link{ameras}}. To attach confidence/credible intervals, use the method \code{\link[=confint.amerasfit]{confint}}. To visualize the exposure uncertainty in the dose realizations, use \code{\link{ecdfplot}}.
}
\author{Sander Roberti <sander.roberti@nih.gov>, William Wheeler <WheelerB@imsweb.com>, Ruth Pfeiffer <pfeiffer@mail.nih.gov>, and Deukwoo Kwon <DKwon@uams.edu}

\keyword{measurement error}
\keyword{exposure uncertainty}
\keyword{exposure replicates}
\keyword{regression calibration} 
\keyword{Monte Carlo Maximum Likelihood}
\keyword{Bayesian model averaging} 
\keyword{frequentist model averaging}

\references{
  Roberti, S., Kwon D., Wheeler W., Pfeiffer R. (in preparation).
  ameras: An R Package to Analyze Multiple Exposure Realizations in Association Studies
}

\name{print.amerasfit}
\alias{print.amerasfit}
\title{Simple summary for an amerasfit object}

\description{
  Prints a simple summary of a fitted \code{amerasfit} object.
}

\usage{
\method{print}{amerasfit}(x, digits = max(3, getOption("digits") - 3), \dots)
}

\arguments{
  \item{x}{
    A fitted model object of class \code{amerasfit}, as returned by
    \code{\link{ameras}}.
  }
  \item{digits}{
    Number of significant digits to be printed. Default is `max(3, getOption("digits") - 3)`
  }
  \item{\dots}{
    Additional arguments, currently unused.
  }
}

\value{
  Prints the `ameras` call, number of rows and dose replicates in the data, runtime,
  and model coefficients.
}



\seealso{
  \code{\link{ameras}} for model fitting,
  \code{\link[=summary.amerasfit]{summary}} for a more detailed summary of the fitted models
  including confidence intervals if computed.
}

\examples{
\donttest{
data("data", package = "ameras")

## Fit the model
fit <- ameras(Y.binomial~dose(V1:V10, model="ERR"), data = data, family = "binomial", 
              methods = c("RC","ERC"))

## Default print
fit
print(fit)

## More digits
print(fit, digits=5)
}
}

\name{transform1}
\alias{transform1}
\title{Exponential parameter transformation}

\description{
Applies exponential transformation \eqn{f(\theta_i)=\exp(\theta_i)+L_i} to one or multiple components of parameter vector \eqn{\bm \theta}, where \eqn{L_i} are lower limits that can be different for each component
}

\usage{
transform1(params, index.t=1:length(params), lowlimit=rep(0,length(index.t)), 
  boundcheck=FALSE, boundtol=1e-3, ... )
}
\arguments{
  \item{params}{full input parameter vector}
  \item{index.t}{indices of parameters to be transformed (default all)}
  \item{lowlimit}{lower limits to be applied (default zero), where the k-th component of \code{lowlimit} is applied to the k-th index in \code{index.t} }
  \item{boundcheck}{whether to produce a warning when any of the transformed parameters are within \code{boundtol} of \code{lowlimit}}
  \item{boundtol}{tolerance for producing a warning for reaching the boundary}
  \item{...}{not used}
}
\value{
Transformed parameter vector.
}



\examples{
params <- c(.1, .5, 1)
transform1(params, lowlimit=c(0, -1, 1))
}

\name{ecdfplot}
\alias{ecdfplot}
\title{Visualize multiple dose realizations}
\description{
Create a descriptive figure to visualize the distribution of dose and its uncertainty.
}
\usage{
ecdfplot(data, dosevars, xlab="Dose", 
  ylab="Cumulative distribution", log.xaxis=TRUE)
}
\arguments{
  \item{data}{data frame containing columns with dose vectors}
  \item{dosevars}{names or column indices of dose vectors.}
  \item{xlab}{label for the x-axis, default \code{"Dose"}}
  \item{ylab}{label for the y-axis, default \code{"Cumulative distribution"}}
  \item{log.xaxis}{logical, whether to use a log-scale for the dose axis (default \code{TRUE}). When \code{log.xaxis=TRUE}, any zeros in the doses are excluded while plotting.}
}
\details{
In the left panel, the empirical cumulative distribution function (ECDF) is
plotted for each dose realization. In other words, each curve shows one distribution
of dose across individuals. The spread within individual curves reflects the dose range
across individuals, while the spread between curves reflects between-realization variation
on the cohort level.

In the right panel, ECDFs are plotted for each individual,
showing distributions within individuals. A wide spread within individual curves
is indicative of large within-individual variation, while the spread between
curves reflects between-individual variation.

When using a log-scale for the x-axis, any zero dose values are excluded before plotting.
}

\examples{
\dontrun{
   data(data, package="ameras")
   ecdfplot(data,  dosevars=paste0("V", 1:10))
}
}
